/*
 * Created on 2004-11-13
 */
package crn;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * The Case Retrieval Net.
 * @author Piotrek
 */
public class CRN 
{	
	/**
	 * A list of all ies in the CRN.
	 * @author Piotrek
	 */
	private ArrayList ies = new ArrayList();
	
	/**
	 * A list of all cases in the CRN.
	 * @author Piotrek
	 */
	private ArrayList cases = new ArrayList();
	
	/**
	 * The weights of the IEs. Name of the IE is a key. The value is a weight of the IE (Float).
	 * @author Piotrek
	 */
	private Hashtable iesWeights = new Hashtable();
	
	/**
	 * The IE specific comparison manager.
	 * @author Piotrek
	 */
	private IESpecificComparisonManager specificComparisonManager = new IESpecificComparisonManager();
	
	/**
	 * Adds the given case to the CRN.
	 * @author Piotrek
	 * @param newCase The case to be added. The added case is not the one given, but its close copy.
	 * @param defaultWeightForNewIEs A default weight for the new IEs. Used only if there is not yet defined weight for the type of the given IE.
	 */
	public void AddCase(Case newCase, float defaultWeightForNewIEs)
	{		
		if (newCase == null)
			throw new NullPointerException("Case must not be null.");
				
		// checking if there is no such a case in this CRN
		ArrayList similiarCases = FindSimilarCases(newCase, 1.0f);
				
		if (similiarCases.size() == 1 && ((Case)similiarCases.get(0)).IsAccepted() == newCase.IsAccepted())
			return; // we don't need to double the data
		
		if (similiarCases.size() == 2 && ((Case)similiarCases.get(0)).IsAccepted() == newCase.IsAccepted())
			return; // we don't need to double the data
		
		if (similiarCases.size() == 2 && ((Case)similiarCases.get(1)).IsAccepted() == newCase.IsAccepted())
			return; // we don't need to double the data
	
		// creating a case to be added to this CRN
		String newName = "Case " + Integer.toString(cases.size() + 1);
			
		Case myNewCase = new Case(newName, newCase.IsAccepted(), this);
				
		// adding the IEs
		Iterator iterator = newCase.GetIEReferencesIterator();
				
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			AddIE(ie, defaultWeightForNewIEs, myNewCase);
		}
						
		cases.add(myNewCase);
	}
	
	/**
	 * Removes the given case from the CRN.
	 * @author Piotrek
	 * @param travelCase A case to be removed.
	 */
	public void RemoveCase(Case travelCase) 
	{
		if (travelCase == null)
			throw new NullPointerException("Case must not be null in CRN.RemoveCase().");
		
		if (!cases.contains(travelCase))
			throw new IllegalArgumentException("There is no such a case in the CRN in CRN.RemoveCase().");
			
		ArrayList iesToRemove = new ArrayList();
		
		Iterator iterator = ies.iterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			if (ie.GetCaseReferencesCount() == 1)
			{
				if (((Case)(ie.GetCaseReferencesInterator().next())).equals(travelCase))
					iesToRemove.add(ie);
			}
			else
			{
				if (ie.ContainsCase(travelCase)) 
					ie.RemoveCaseReference(travelCase);
			}
		}
		 
		cases.remove(travelCase);
		
		iterator = iesToRemove.iterator();
		
		while (iterator.hasNext())
		{
			ies.remove(iterator.next());
		}
	}
	
	/**
	 * Finds all the similiar cases in the CRN.
	 * @author Piotrek
	 * @param travelCase A case to find similiar cases to.
	 * @param similarityTreshold The level at which two cases are consider similiar (0 <= treshold <= 1).
	 * @return A list of similiat cases. If there are no similiar cases, the empty list is returned.
	 */
	public ArrayList FindSimilarCases(Case travelCase, float similarityTreshold) 
	{	
		if (travelCase == null)
			throw new NullPointerException("Case must not be null in CRN.FindSimilarCases().");
		
		if (similarityTreshold < 0.0f || similarityTreshold > 1.0f)
			throw new IllegalArgumentException("Similarity treshold must be in range [0;1] in CRN.FindSimilarCases().");
		
		ArrayList activatedCases = GetActivatedCases(travelCase, similarityTreshold);
		
		ArrayList similiarCases = new ArrayList();
		
		Iterator iterator = activatedCases.iterator();
		
		while (iterator.hasNext())
		{
			Case candidateCase = (Case)iterator.next();
			
			CompareCases(candidateCase, travelCase);
			
			if (candidateCase.GetSimilarity() >= similarityTreshold)
				similiarCases.add(candidateCase);
		}
		
		return similiarCases;
	}
	
	/**
	 * Finds the most similiar case in the CRN.
	 * @author Piotrek
	 * @param travelCase A case to find the most similiar case for.
	 * @return The most similar case. If there are more cases with the same similarity the more recent one is returned. If there are no cases in the history the null is returned.
	 */
	public Case FindTheMostSimilarCase(Case travelCase) 
	{	
		if (travelCase == null)
			throw new NullPointerException("Case must not be null in CRN.FindTheMostSimilarCase().");
				
		ArrayList activatedCases = GetActivatedCases(travelCase, 0);
		
		Case theMostSimilarCase = null;
		
		Iterator iterator = activatedCases.iterator();
		
		while (iterator.hasNext())
		{
			Case candidateCase = (Case)iterator.next();
			
			CompareCases(candidateCase, travelCase);
			
			if (theMostSimilarCase == null || candidateCase.GetSimilarity() > theMostSimilarCase.GetSimilarity())
				theMostSimilarCase = candidateCase;
			else
			{
				if (theMostSimilarCase != null && candidateCase.GetSimilarity() == theMostSimilarCase.GetSimilarity())
				{
					String theMostSimilarCaseName = theMostSimilarCase.GetName().substring(5);
					String candidateCaseName = candidateCase.GetName().substring(5);
					
					int theMostSimilarCaseNumber = Integer.parseInt(theMostSimilarCaseName);
					int candidateCaseNumber = Integer.parseInt(candidateCaseName);
					
					if (candidateCaseNumber > theMostSimilarCaseNumber) // it is more recent
						theMostSimilarCase = candidateCase;
				}
			}
		}
		
		return theMostSimilarCase;
	}
	
	/**
	 * Gets the weight of the given IE.
	 * @author Piotrek
	 * @param ieName The name of IE to look for.
	 * @return The weight of the requestes IE.
	 * @throws IllegalArgumentException If there is no such an IE in the CRN. 
	 */
	public float GetIEWeight(String ieName) throws IllegalArgumentException
	{
		if (ieName == null)
			throw new NullPointerException("IE name must not be null in CRN.GetIEWeight().");
		
		Object value = iesWeights.get(ieName);
		
		if (value == null)
			throw new IllegalArgumentException("There is no such an IE weight.");
				
		return ((Float)value).floatValue();
	}
	
	/**
	 * Sets the weight of the given IE.
	 * @author Piotrek
	 * @param ieName The name of IE to set for.
	 * @param ieWeight The new weight for teh given type of IE.
	 * @throws IllegalArgumentException If there is no such an IE in the CRN. 
	 */
	public void SetIEWeight(String ieName, float ieWeight) throws IllegalArgumentException
	{				
		iesWeights.put(ieName, new Float(ieWeight));
	}
	
	/**
	 * Compares the two given cases. Sets the similarity to the first one of them (muCase).
	 * @author Piotrek
	 * @param myCase The fisrt case to be compared. Must belong to this CRN. This one will have the similiairty set.
	 * @param otherCase The second case to be compared. Do not have to belong to this CRN (usually does not). This one will not have the similiairty set.
	 * @throws IllegalArgumentException Thrown if the myCase case does not belong to this CRN.
	 */
	private void CompareCases(Case myCase, Case otherCase) throws IllegalArgumentException
	{
		if (myCase == null)
			throw new NullPointerException("MyCase must not be null!");
		
		if (otherCase == null)
			throw new NullPointerException("OtherCase must not be null!");
			
		if (myCase.GetCRN() != this)
			throw new IllegalArgumentException("MyCase does not belong to this CRN.");
		
		// gathering names of IEs in both cases
		
		ArrayList ieNames = new ArrayList();
		
		Iterator iterator = myCase.GetIEReferencesIterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			if (!ieNames.contains(ie.name))
				ieNames.add(ie.name);
		}
		
		iterator = otherCase.GetIEReferencesIterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			if (!ieNames.contains(ie.name))
				ieNames.add(ie.name);
		}
		
		// comparing the cases

		float sum = 0.0f;
		float weightsSum = 0.0f;
		
		iterator = ieNames.iterator();
		
		while (iterator.hasNext())
		{
			String ieName = (String)iterator.next();
			
			ArrayList myIEs = myCase.GetIEs(ieName);
			ArrayList otherIEs = otherCase.GetIEs(ieName);
			
			float similarity = CompareIEsSets(myIEs, otherIEs);
			
			// get the weight
			
			float weight;
			
			Float weightFloat = (Float)iesWeights.get(ieName);
			
			if (weightFloat != null)
				weight = weightFloat.floatValue();
			else
				weight = IEImportance.NEUTRAL;
			
			// update values
			
			sum += similarity * weight;
			
			weightsSum += weight;
		}		
		
		myCase.SetSimilarity(sum / weightsSum);
	}
	
	/**
	 * Compares two sets of IEs (all of which are of the same type). 
	 * @author Piotrek
	 * @param ies1 The first set of IEs to compare.
	 * @param ies2 The second set of IEs to compare.
	 * @return The similarity of those two sets of IEs.
	 */
	private float CompareIEsSets(ArrayList ies1, ArrayList ies2)
	{
		if (ies1.size() == 0)
		{
			if (ies2.size() > 0)
				return 0.5f;
			
			return 0.0f;
		}
			
		if (ies2.size() == 0)
		{
			if (ies1.size() > 0)
				return 0.5f;
		
			return 0.0f;
		}
				
		ArrayList l1;
		ArrayList l2;
		
		if (ies1.size() < ies2.size())
		{
			l1 = ies1;
			l2 = ies2;
		}
		else
		{
			l1 = ies2;
			l2 = ies1;
		}
		
		int[] permutaion;
		
		PermutationGenerator permutaionGenerator = new PermutationGenerator(l2.size());
		
		float maxSimilarity = 0.0f;
		
		while (permutaionGenerator.hasMore())
		{
			permutaion = permutaionGenerator.getNext();
			
			float similarity = 0.0f;
			
			for (int i = 0; i < permutaion.length; ++i)
			{				
				if (i < l1.size())
				{
					IE ie1 = (IE)l1.get(i);
					IE ie2 = (IE)l2.get(i);
					
					boolean fCompared = false;
					
					float value = 0.0f;
					
					if (specificComparisonManager.HasSpecificComparator(ie1.name))
					{
						value = specificComparisonManager.CompareIEs(ie1, ie2);
						
						fCompared = true;					 
					}
					
					if (!fCompared)
					{
						if (ie1.value.compareTo(ie2.value) == 0)
							value = 1.0f;	// equal
						else
							value = 0.0f;	// different values
					}
					
					similarity += value;
				}
				else
					similarity += 0.5;
			}
			
			similarity /= l2.size();
			
			if (similarity > maxSimilarity)
				maxSimilarity = similarity;
		}
		
		return maxSimilarity;
	}
	
	/**
	 * Adds the given IE to the CRN.
	 * @author Piotrek
	 * @param ie The IE to be added.
	 * @param defaultWeight The default weight for this IE (if a new one).
	 * @param travelCase The case this IE is referencing to.
	 */
	private void AddIE(IE ie, float defaultWeight, Case travelCase)
	{
		// checking if there is such an IE in the CRN
				
		if (ies.contains(ie))
		{
			int i = ies.indexOf(ie);
			
			IE ieForUpadte = (IE)ies.get(i);
			
			ieForUpadte.AddCaseReference(travelCase);
			
			travelCase.AddIEReference(ieForUpadte);
		}
		else
		{
			IE newIE = new IE(ie);
			
			newIE.AddCaseReference(travelCase);
			
			ies.add(newIE);
			
			travelCase.AddIEReference(newIE);
		}
		
		// the weight
		
		Object object = iesWeights.get(ie.name);
		
		if (object == null)
		{			
			iesWeights.put(ie.name, new Float(defaultWeight));
		}
	}
		
	/**
	 * Gets the list of activated cases by the given one. Activated are cases that has at least one similiar enough IE (with the similarity greater or equal to the treshold).
	 * @author Piotrek
	 * @param travelCase The activator case.
	 * @param activationTreshold The activation treshold.
	 * @return The list of all activated cases. If there are no activated cases, the empty list is returned.
	 */
	private ArrayList GetActivatedCases(Case travelCase, float activationTreshold)
	{		
		// activating IEs
		
		ArrayList activatedNames = new ArrayList();
		
		Iterator iesIterator = travelCase.GetIEReferencesIterator();
		
		while (iesIterator.hasNext())
		{
			IE ie = (IE)iesIterator.next();
			
			if (iesWeights.containsKey(ie.name)) // it means we've got such a type of IE
				activatedNames.add(ie.name);
		}
		
		// gathering the activated cases
		
		ArrayList activatedCases = new ArrayList();
		
		iesIterator = ies.iterator();
		
		while (iesIterator.hasNext())
		{
			IE ie = (IE)iesIterator.next();
			
			if (activatedNames.contains(ie.name))			
			{
				ArrayList travelCaseIEs = travelCase.GetIEs(ie.name);
				
				float maxSimilarity = 0.0f;
				
				Iterator iterator = travelCaseIEs.iterator();
				
				while (iterator.hasNext())
				{
					IE travelCaseIE = (IE)iterator.next();
					
					float similarity = 0.0f;
					
					boolean fCompared = false;
					
					if (specificComparisonManager.HasSpecificComparator(ie.name))
					{
						similarity = specificComparisonManager.CompareIEs(ie, travelCaseIE);
						
						fCompared = true;
					}
					
					if (!fCompared)
					{
						if (ie.value.compareTo(travelCaseIE.value) == 0)
							similarity = 1.0f;	// equal
						else
							similarity = 0.0f;	// different values
					}
					
					if (similarity > maxSimilarity)
						maxSimilarity = similarity;
				}
				
				if (maxSimilarity >= activationTreshold)
				{
					Iterator casesIterator = ie.GetCaseReferencesInterator();
					
					while (casesIterator.hasNext())
					{
						Case activatedCase = (Case)casesIterator.next();
						
						if (!activatedCases.contains(activatedCase))
							activatedCases.add(activatedCase);
					}
				}
			}
		}
		
		return activatedCases;
	}
	
	/**
	 * Adds a specific comparator for the IEs with the given name.
	 * @author Piotrek
	 * @param ieName The name of the IEs the given comaparator should be applied to.
	 * @param ieComparator The comparator for the IEs with the given name.
	 * @see crn.IESpecificComparisonManager#AddSpecificComparator(java.lang.String ieName, crn.IEComparator ieComparator)
	 */
	public void AddSpecificComparator(String ieName, IEComparator ieComparator)
	{		
		specificComparisonManager.AddSpecificComparator(ieName, ieComparator);
	}
	
	/**
	 * Removes a specific comparator for the IEs with the give name.
	 * @author Piotrek
	 * @param ieName A name of the IEs for which the specific comparator is to be removed.
	 * @see crn.IESpecificComparisonManager#RemoveSpecificComparator(java.lang.String ieName)
	 */
	public void RemoveSpecificComparator(String ieName)
	{
		if (ieName == null)
			throw new NullPointerException("IE name must nmot be null in CRN.RemoveSpecificComparator().");
		
		specificComparisonManager.RemoveSpecificComparator(ieName);
	}
	
	/**
	 * Stores the CRN to the given file.
	 * @author Piotrek
	 * @param fileName The name of file to store the CRN to.
	 * @throws IOException Thrown in case of IO problems.
	 */
	public void StoreToFile(String fileName) throws IOException
	{
		FileWriter fileWriter = new FileWriter(fileName);
		
		fileWriter.write("<?xml version=\"1.0\"?>\r\n");
		
		fileWriter.write("<CRN>\r\n");
		
		// Cases
		fileWriter.write("   <Cases>\r\n");
		
		Iterator iterator = cases.iterator();
		
		while (iterator.hasNext())
		{
			Case travelCase = (Case)iterator.next();
			
			fileWriter.write("      <Case name=\"");
			fileWriter.write(travelCase.GetName());
			fileWriter.write("\" isAccepted=\"");
			fileWriter.write(Boolean.toString(travelCase.IsAccepted()));
			fileWriter.write("\"/>\r\n");
		}
		
		fileWriter.write("   </Cases>\r\n");
			
		// IEs
		fileWriter.write("   <IEs>\r\n");
		
		iterator = ies.iterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			fileWriter.write("      <IE name=\"");
			fileWriter.write(ie.name);
			fileWriter.write("\" value=\"");
			fileWriter.write(ie.value);
			fileWriter.write("\">\r\n");
			
			Iterator references = ie.GetCaseReferencesInterator();
			
			while (references.hasNext())
			{
				Case travelCase = (Case)references.next();
				
				fileWriter.write("         <CaseReference name=\"");
				fileWriter.write(travelCase.GetName());
				fileWriter.write("\"/>\r\n");
			}
			
			fileWriter.write("      </IE>\r\n");
		}
		
		fileWriter.write("   </IEs>\r\n");
		
		// weights
		fileWriter.write("   <Weights>\r\n");
		
		iterator = iesWeights.keySet().iterator();
		
		while (iterator.hasNext())
		{
			String ieName = (String)iterator.next();
			Float value = (Float)iesWeights.get(ieName);
			
			fileWriter.write("      <Weight ieName=\"");
			fileWriter.write(ieName);
			fileWriter.write("\" value=\"");
			fileWriter.write(value.toString());
			fileWriter.write("\"/>\r\n");
		}
		
		fileWriter.write("   </Weights>\r\n");
		
		fileWriter.write("</CRN>\r\n");
		
		fileWriter.flush();
	}
	
	/**
	 * Gets the iterator for the cases.
	 * @author Piotrek
	 * @return The requested iterator.
	 */
	public Iterator GetCases()
	{
		return cases.iterator();
	}
	
	/**
	 * Removes all the cases from the CRN.
	 * @author Piotrek
	 */
	public void RemoveAllCases()
	{
		cases.clear();
		
		ies.clear();
	}
}
