/*
 * Created on 2004-12-29
 */
package crn;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Parses the XML file containg the description of the CRN.
 * @author Piotrek
 */
public class CRNFileParser extends DefaultHandler
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public CRNFileParser()
	{
		super();
	}
	
	/**
	 * Parses the given file.
	 * @author Piotrek
	 * @param fileName The name of the file to be parsed.
	 * @return The CRN created basing on the file.
	 * @throws Exception Thrown in case of error.
	 * @throws IOException In case of any IO error.
	 * @throws SAXException In case of some parsing error.
	 */
	public CRN Parse(String fileName) throws Exception, IOException, SAXException
	{
		if (fileName == null)
			throw new NullPointerException("File name must not be null.");
			
		ClearData();
				
		SAXParser saxParser;
		
		try
		{
			saxParser = CreateSAXParser();
		}
		
		catch (Throwable xcp)
		{
			throw new Exception("Failed to create a parser.");
		}
		
		saxParser.parse(new File(fileName), this);
				
		return GetCRN();
	}
	
	/**
	 * Clears all the internal data, so the parser is ready to parse a new file.
	 * @author Piotrek
	 *
	 */
	private void ClearData()
	{
		ies.clear();
		weights.clear();
		caseReferences.clear();
		cases.clear();
		
		currentIE = null;
		
		inCases = false;
		inIE = false;
		inIEs = false;
		inWeights = false;
	}
	
	/**
	 * Creates the CRN according to the data read from the file.
	 * @author Piotrek
	 * @return The created CRN.
	 */
	private CRN GetCRN()
	{
		CRN crn = new CRN();
		
		// adding IEs to the cases
		
		Iterator iterator = ies.iterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			ArrayList references = (ArrayList)caseReferences.get(ie);
			
			Iterator referencesIterator = references.iterator();
			
			while (referencesIterator.hasNext())
			{
				String caseName = (String)referencesIterator.next();
				
				Iterator casesIterator = cases.iterator();
				
				Case theCase = null;
				
				while (casesIterator.hasNext())
				{
					Case travelCase = (Case)casesIterator.next();
					
					if (travelCase.GetName().compareTo(caseName) == 0)
					{
						theCase = travelCase;
						
						break;
					}
				}
				
				if (theCase != null)
				{
					theCase.AddIEReference(ie);
					ie.AddCaseReference(theCase);
				}
			}
		}
		
		// weights
		
		iterator = weights.keySet().iterator();
		
		while (iterator.hasNext())
		{
			String ieName = (String)iterator.next();
			
			Float valueFloat = (Float)weights.get(ieName);
			
			crn.SetIEWeight(ieName, valueFloat.floatValue());
		}
		
		// adding cases
		
		iterator = cases.iterator();
		
		while (iterator.hasNext())
		{
			Case travelCase = (Case)iterator.next();
			
			crn.AddCase(travelCase, IEImportance.NEUTRAL);
		}
		
		return crn;
	}
	
	/**
	 * Creates an instance of the SAXParser.
	 * @author Piotrek
	 * @return The requested parser.
	 * @throws ParserConfigurationException If it was imposible to create a parser satisfing the current configuration.
	 * @throws SAXException
	 */
	private static SAXParser CreateSAXParser() throws ParserConfigurationException, SAXException
	{
	 	SAXParserFactory factory = SAXParserFactory.newInstance();
		
	 	SAXParser saxParser = factory.newSAXParser();
	 	
	 	return saxParser;
	}
		
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uriNamespace, String simpleName, String qualifiedName, Attributes attributes) throws SAXException
	{
		String name = simpleName;
		
		if (name.equals("")) 
			name = qualifiedName; // not namespace aware
		
		if (name.equals("Cases"))
		{
			inCases = true;
		}
		
		if (name.equals("Case"))
		{
			if (!inCases)
				throw new SAXParseException("Case tag outside the cases tag.", locator);
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("Case tag has invalid number of attributes.", locator);
			
			if (attributes.getValue("name") == null)
				throw new SAXParseException("Case tag has no name attribute.", locator);
			
			if (attributes.getValue("isAccepted") == null)
				throw new SAXParseException("Case tag has no isAccepted attribute.", locator);
			
			String caseName = attributes.getValue("name");
			boolean isAccpeted = attributes.getValue("isAccepted").equals("true");
			
			cases.add(new Case(caseName, isAccpeted, null));
		}
		
		if (name.equals("IEs"))
		{
			inIEs = true;
		}
		
		if (name.equals("IE"))
		{
			if (!inIEs)
				throw new SAXParseException("IE tag outside the IEs tag.", locator);
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("IE tag has invalid number of attributes.", locator);
			
			if (attributes.getValue("name") == null)
				throw new SAXParseException("IE tag has no name attribute.", locator);
			
			if (attributes.getValue("value") == null)
				throw new SAXParseException("IE tag has no value attribute.", locator);
			
			String ieName = attributes.getValue("name");
			String ieValue = attributes.getValue("value");
			
			currentIE = new IE(ieName, ieValue);
			
			ies.add(currentIE);
			
			inIE = true;
			
			caseReferences.put(currentIE, new ArrayList());
		}
		
		if (name.equals("CaseReference"))
		{
			if (!inIE)
				throw new SAXParseException("CaseReference tag outside the IE tag.", locator);
			
			if (attributes == null || attributes.getLength() != 1)
				throw new SAXParseException("CaseReference tag has invalid number of attributes.", locator);
			
			if (attributes.getValue("name") == null)
				throw new SAXParseException("CaseReference tag has no name attribute.", locator);
						
			String caseName = attributes.getValue("name");
			
			ArrayList references = (ArrayList)caseReferences.get(currentIE);
			
			references.add(caseName);
		}
		
		if (name.equals("Weights"))
		{
			inWeights = true;
		}
		
		if (name.equals("Weight"))
		{
			if (!inWeights)
				throw new SAXParseException("Weight tag outside the Weights tag.", locator);
			
			if (attributes == null || attributes.getLength() != 2)
				throw new SAXParseException("Weight tag has invalid number of attributes.", locator);
			
			if (attributes.getValue("ieName") == null)
				throw new SAXParseException("Weight tag has no ieName attribute.", locator);
			
			if (attributes.getValue("value") == null)
				throw new SAXParseException("Weight tag has no value attribute.", locator);
						
			String ieName = attributes.getValue("ieName");
			float value = IEImportance.NEUTRAL;
			
			try
			{
				value = Float.parseFloat(attributes.getValue("value"));
			}
			
			catch (NumberFormatException xcp)
			{
				String valueString = attributes.getValue("value");
				
				if (valueString.indexOf(".") != -1)
					valueString.replaceAll(".", ",");
				else
					valueString.replaceAll(",", ".");
				
				try
				{
					value = Float.parseFloat(valueString);
				}
				
				catch (NumberFormatException xcp2)
				{
					value = IEImportance.NEUTRAL;
				}
			}
			
			weights.put(ieName, new Float(value));
		}
	}
	
	/**
	 * The hashtable mapping the IEs to the ArrayLists containg the names of referring cases (Strings).
	 */
	private Hashtable caseReferences = new Hashtable();
	
	/**
	 * The hashtable mapping the IEs' names (String) to their values (Float).
	 */
	private Hashtable weights = new Hashtable();
	
	/**
	 * The current IE tag. Needed to add the case referneces to it.
	 * @author Piotrek
	 */
	private IE currentIE = null;
		
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void endElement(String uriNamespace, String qualifiedName, String simpleName)
	{
		String name = simpleName;
		
		if (name.equals("")) 
			name = qualifiedName; // not namespace aware
		
		if (name.equals("Cases"))
		{
			inCases = false;
		}
			
		if (name.equals("IEs"))
		{
			inIEs = false;
		}
		
		if (name.equals("IE"))
		{
			inIE = false;
		}
		
		if (name.equals("Weights"))
		{
			inWeights = false;
		}
	}
	
	/**
	 * A list of parsed cases.
	 * @author Piotrek
	 */
	private ArrayList cases = new ArrayList();
	
	/**
	 * A list of parsed IEs.
	 * @author Piotrek
	 */
	private ArrayList ies = new ArrayList();
	
	/**
	 * Indicates whether the parser is in the Cases section.
	 * @author Piotrek
	 */
	private boolean inCases = false;
	
	/**
	 * Indicates whether the parser is in the IEs section.
	 * @author Piotrek
	 */
	private boolean inIEs = false;
	
	/**
	 * Indicates whether the parser is in the IE section.
	 * @author Piotrek
	 */
	private boolean inIE = false;
	
	/**
	 * Indicates whether the parser is in the Weights section.
	 * @author Piotrek
	 */
	private boolean inWeights = false;
	
	/**
	 * @author Piotrek
	 * @see org.xml.sax.ContentHandler#setDocumentLocator(org.xml.sax.Locator)
	 */
	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
	}
	
	/**
	 * The SAX locator.
	 * @author Piotrek
	 */
	private Locator locator;
}
