/*
 * Created on 2004-11-13
 */
package crn;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * The case.
 * @author Piotrek
 */
public class Case
{
	/**
	 * A name of the case.
	 * @author Piotrek
	 */
	private String name;
	
	/**
	 * Indicates whether this case has been accepted or rejected by the user.
	 * @author Piotrek
	 */
	private boolean fAccepted;
	
	/**
	 * A list of references to all the IEs constituting this case.
	 * @author Piotrek
	 */
	protected ArrayList iesReferences = new ArrayList();
	
	/**
	 * The similarity with the other case. Set by the CRN. Relevant only after calling the CRN.CompareCases().
	 */
	private float similarity = 0.0f;
	
	/**
	 * The owner of this case. Null if there is no owner.
	 */
	private CRN crn = null;
	
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param name The name of this case.
	 * @param fAccepted Indicates whether this case has been accepted or rejected by the user.
	 * @param crn The owner of this case. Set to null if there is no owner.
	 */
	public Case(String name, boolean fAccepted, CRN crn)
	{
		if (name == null)
			throw new NullPointerException("Name must not be null in Case.Case().");
				
		this.name = name;
		this.fAccepted = fAccepted;
		this.crn = crn;
	}
	
	/**
	 * Indicates whether this case has been accepted or rejected by the user.
	 * @author Piotrek
	 * @return Returns a boolean value indicating whether this case has been accepted or rejected by the user.
	 */
	public boolean IsAccepted()
	{
		return fAccepted;
	}
	
	/**
	 * Gets the name of this case.
	 * @author Piotrek
	 * @return Returns the name.
	 */
	public String GetName()
	{
		return name;
	}
	
	/**
	 * @author Piotrek
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object object)
	{
		if (object == null)
			throw new NullPointerException("Other object must not be null.");
		
		if (!Case.class.isInstance(object))
			return false;
		
		Case otherCase = (Case)object;
		
		return (otherCase.name.compareTo(name) == 0 && otherCase.fAccepted == fAccepted);
	}
	
	/**
	 * Tests the equality of two cases. They may belong to two different CRNs.
	 * @author Piotrek
	 * @param otherCase The other case this one is to be compared to.
	 * @return A boolean value indicating whether this and the other cases are equal.
	 */
	public boolean IsEqual(Case otherCase)
	{
		if (otherCase == null)
			throw new NullPointerException("Other case must not be null.");
		
		return equals(otherCase);
	}
	
	/**
	 * Adds a reference to the IE references list of this case.
	 * @author Piotrek
	 * @param ie The IE to be added.
	 */
	public void AddIEReference(IE ie)
	{
		if (iesReferences.contains(ie))
			return;
		
		iesReferences.add(ie);
	}
	
	/**
	 * Gets the IE references iterator.
	 * @author Piotrek
	 * @return The requested iterator.
	 */
	public Iterator GetIEReferencesIterator()
	{
		return iesReferences.iterator();
	}
	
	/**
	 * Gets the similarity.
	 * @author Piotrek
	 * @return Returns the similarity.
	 */
	public float GetSimilarity()
	{
		return similarity;
	}
	
	/**
	 * Sets the similarity.
	 * @author Piotrek
	 * @param similarity The similarity to set.
	 */
	public void SetSimilarity(float similarity)
	{
		if (similarity < 0.0f || similarity > 1.0f)
			throw new IllegalArgumentException("Similarity must be in range of [0;1].");
		
		this.similarity = similarity;
	}
	
	/**
	 * Gets the owner of this case.
	 * @author Piotrek
	 * @return Returns the owner of this case. Null is returned, if there is no owner.
	 */
	public CRN GetCRN()
	{
		return crn;
	}
	
	/**
	 * Chesks whether this case has an IE of the given type.
	 * @author Piotrek
	 * @param ieName The name of the IE type.
	 * @return The boolean value indicating whether this case has an IE of the given type.
	 */
	public boolean HasIEType(String ieName)
	{
		Iterator iterator = iesReferences.iterator();
		
		while (iterator.hasNext())
		{
			if (((IE)iterator.next()).name.compareTo(ieName) == 0)
				return true;
		}
		
		return false;
	}
	
	/**
	 * Gets the IEs referencing to this case of the given type.
	 * @author Piotrek
	 * @param name Name of type.
	 * @return The referencing IEs or an empty list, if there are no IEs of the requested type.
	 */
	public ArrayList GetIEs(String name)
	{
		ArrayList ies = new ArrayList();
		
		Iterator iterator = iesReferences.iterator();
		
		while (iterator.hasNext())
		{
			IE ie = (IE)iterator.next();
			
			if (ie.name.compareTo(name) == 0)
				ies.add(ie);
		}
		
		return ies;
	}
	
	/**
	 * Sets the value indicating whether this case should be accepted or not.
	 * @author Piotrek
	 * @param fAccepted The new value.
	 */
	public void SetAccepted(boolean fAccepted)
	{
		this.fAccepted = fAccepted;
	}
}
