/*
 * Created on 2004-11-15
 */

package crn;

/**
 * @author http://www.merriampark.com/perm.htm
 */

//--------------------------------------
//Systematically generate permutations. 
//--------------------------------------

import java.math.BigInteger;

/**
 * Systematically generate permutations. 
 * @author http://www.merriampark.com/perm.htm
 */
public class PermutationGenerator {

/**
 * @author http://www.merriampark.com/perm.htm
 */
private int[] a;
/**
 * @author http://www.merriampark.com/perm.htm
 */
private BigInteger numLeft;
/**
 * @author http://www.merriampark.com/perm.htm
 */
private BigInteger total;

//-----------------------------------------------------------
// Constructor. WARNING: Don't make n too large.
// Recall that the number of permutations is n!
// which can be very large, even when n is as small as 20 --
// 20! = 2,432,902,008,176,640,000 and
// 21! is too big to fit into a Java long, which is
// why we use BigInteger instead.
//----------------------------------------------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * @param n
 */
public PermutationGenerator (int n) {
 if (n < 1) {
   throw new IllegalArgumentException ("Min 1");
 }
 a = new int[n];
 total = getFactorial (n);
 reset ();
}

//------
// Reset
//------

/**
 * @author http://www.merriampark.com/perm.htm
 */
public void reset () {
 for (int i = 0; i < a.length; i++) {
   a[i] = i;
 }
 numLeft = new BigInteger (total.toString ());
}

//------------------------------------------------
// Return number of permutations not yet generated
//------------------------------------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * 
 * @return Gets the number of left possibilities.
 */
public BigInteger getNumLeft () {
 return numLeft;
}

//------------------------------------
// Return total number of permutations
//------------------------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * 
 * @return Gets the total number of possibilities.
 */
public BigInteger getTotal () {
 return total;
}

//-----------------------------
// Are there more permutations?
//-----------------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * 
 * @return A boolean value indicating if there are some more permutations to be visited. 
 */
public boolean hasMore () {
 return numLeft.compareTo (BigInteger.ZERO) == 1;
}

//------------------
// Compute factorial
//------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * 
 * @param n
 * @return Gets the factorial of the given number.
 */
private static BigInteger getFactorial (int n) {
 BigInteger fact = BigInteger.ONE;
 for (int i = n; i > 1; i--) {
   fact = fact.multiply (new BigInteger (Integer.toString (i)));
 }
 return fact;
}

//--------------------------------------------------------
// Generate next permutation (algorithm from Rosen p. 284)
//--------------------------------------------------------

/**
 * @author http://www.merriampark.com/perm.htm
 * 
 * @return Gets an array being the next permutation.
 */
public int[] getNext () {

 if (numLeft.equals (total)) {
   numLeft = numLeft.subtract (BigInteger.ONE);
   return a;
 }

 int temp;

 // Find largest index j with a[j] < a[j+1]

 int j = a.length - 2;
 while (a[j] > a[j+1]) {
   j--;
 }

 // Find index k such that a[k] is smallest integer
 // greater than a[j] to the right of a[j]

 int k = a.length - 1;
 while (a[j] > a[k]) {
   k--;
 }

 // Interchange a[j] and a[k]

 temp = a[k];
 a[k] = a[j];
 a[j] = temp;

 // Put tail end of permutation after jth position in increasing order

 int r = a.length - 1;
 int s = j + 1;

 while (r > s) {
   temp = a[s];
   a[s] = a[r];
   a[r] = temp;
   r--;
   s++;
 }

 numLeft = numLeft.subtract (BigInteger.ONE);
 return a;

}
}
