/*
 * Created on 2005-05-16
 */
package gui;

import gui.dialogs.MainDialog;
import java.util.ArrayList;
import scheduler.TestOptions;
import scheduler.TravelConstraints;
import scheduler.TravelDivisionsConnectionsFinder;
import scheduler.modules.ChangesJessJudger;
import scheduler.modules.ChangesJudgerStub;
import scheduler.modules.ConnectionsFinderTasksManager;
import scheduler.modulesinterfaces.ChangesJudger;
import world.WorldTravelInfoProvider;

/**
 * A thread used to solve the second step of the scheduling algorithm.
 * @author Piotrek
 */
public final class SecondStepSolverThread extends Thread
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param mainDialog The main dialog to be notified of the results.
	 * @param firstStepSolutions A list of travel divisions found by the first step of the travel scheduling algorithm.
	 * @param travelConstraints The travel constraints to be applied.
	 */
	public SecondStepSolverThread(MainDialog mainDialog, ArrayList firstStepSolutions, TravelConstraints travelConstraints)
	{
		if (mainDialog == null)
			throw new NullPointerException("Main dialog must not be null in SecondStepSolverThread.SecondStepSolverThread().");
		
		if (firstStepSolutions == null)
			throw new NullPointerException("First step solutions must not be null in SecondStepSolverThread.SecondStepSolverThread().");
		
		if (travelConstraints == null)
			throw new NullPointerException("Travel constraints must not be null in SecondStepSolverThread.SecondStepSolverThread().");
		
		this.mainDialog = mainDialog;
		this.firstStepSolutions = firstStepSolutions;
		this.travelConstraints = travelConstraints;
	}
	
	/**
	 * The main dialog to be notified of the results.
	 * @author Piotrek
	 */
	private MainDialog mainDialog = null;
	
	/**
	 * A list of the solutions found by the first step of the travel scheduling algorithm.
	 * @author Piotrek
	 */
	private ArrayList firstStepSolutions = null;
	
	/**
	 * The travel constraints.
	 * @author Piotrek
	 */
	private TravelConstraints travelConstraints = null;
	
	/**
	 * @author Piotrek
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		WorldTravelInfoProvider worldTravelInfoProvider = new WorldTravelInfoProvider();
				
		TravelDivisionsConnectionsFinder connectionsFinder = new TravelDivisionsConnectionsFinder();
		
		ChangesJudger changesJudger = null;
		
		switch (TestOptions.GetChangesDroppingEngine())
		{
			case TestOptions.CHANGES_DROPPING_RULES_ENGINE_JESS:
					
				changesJudger = new ChangesJessJudger();
			
				break;
				
			case TestOptions.CHANGES_DROPPING_RULES_ENGINE_JAVA:
				
				changesJudger = new ChangesJudgerStub();
			
				break;
				
			default:
				
				throw new RuntimeException("Unexpected changes dropping engine.");
		}
				
		changesJudger.SetTravelConstraints(travelConstraints);
		
		ConnectionsFinderTasksManager tasksManager = new ConnectionsFinderTasksManager();
		
		connectionsFinder.SetTravelInfoProvider(worldTravelInfoProvider);
		connectionsFinder.SetChangesJudger(changesJudger);
		connectionsFinder.SetTasksManager(tasksManager);
		
		try
		{
			ArrayList solutions = connectionsFinder.FindConnections(firstStepSolutions, travelConstraints);
				
			mainDialog.OnSecondStepReady(solutions);
		}
		
		catch (OutOfMemoryError xcp)
		{
			mainDialog.OnSchedulingAlgorithmStepError("Out of memory");
		}
		
		catch (Throwable t)
		{
			mainDialog.OnSchedulingAlgorithmStepError(t.getMessage());
		}
	}
}
