/*
 * Created on 2005-04-09
 */
package gui.controls;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.util.ArrayList;
import javax.swing.JPanel;
import javax.swing.JTextField;
import world.Connection;
import world.MeanOfTransportation;
import world.Station;
import world.TravelTime;

/**
 * A list view used to display a list of stations and text fields 
 * for their arrival and departure times.
 * @author Piotrek
 */
public class StationsTimetableListView extends JPanel
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param stations A list of stations to edit time of arrivals and departures for. Ignored if connection is not null.
	 * @param meanOfTransportation A mean of transportation. This is required only to count the default times of travel. Ignored if connection is not null.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @param connection A connection to be loaded. Set it to null of adding a new connection (A default times will be loaded accordingly to distances and mean of transportation.)
	 */
	StationsTimetableListView(ArrayList stations, int meanOfTransportation, Connection connection)
	{
		super();
		
		if (connection != null)
		{
			stations = connection.GetStations();
			
			meanOfTransportation = connection.GetMeanOfTransportation();
			
			this.connection = connection;
		}
		
		if (stations == null)
			throw new NullPointerException("Stations list must not be null in StationsTimetableListView.StationsTimetableListView().");
		
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("A mean of transportation must be a valid value in StationsTimetableListView.StationsTimetableListView().");
		
		setLayout(null);
		
		setPreferredSize(new Dimension(725, stations.size() * ROW_HEIGHT));
		setSize(725, stations.size() * ROW_HEIGHT);
				
		this.stations = stations;
		this.meanOfTransportation = meanOfTransportation;
		
		setBackground(Color.WHITE);
		
		LoadStations();
		
		setVisible(true);
	}
	
	/**
	 * A connection to be loaded. Null if creating a new connections.
	 * @author Piotrek
	 */
	private Connection connection = null;
	
	/**
	 * A mean of transportation for the stations.
	 * @author Piotrek
	 */
	private int meanOfTransportation = -1;
	
	/**
	 * A list of stations to edit time of arrivals and departures for.
	 * @author Piotrek
	 */
	private ArrayList stations = null;
	
	/**
	 * Loads all the stations into a list view.
	 * @author Piotrek
	 */
	private void LoadStations()
	{
		if (connection == null)
			CreateNewTimetable();
		else
			LoadConnection();
	}
	
	/**
	 * Loads the stations and creates a new timetable for them accordingly to the
	 * distances and the mean of transportation.
	 * @author Piotrek
	 */
	private void CreateNewTimetable()
	{
		int count = stations.size();
		
		Station previousStation = null;
		
		int accumulatedTime = 0;
		
		for (int i = 0; i < count; ++i)
		{				
			// arrival text field
			JTextField arrivalTextField = new JTextField();
			
			arrivalTextFields.add(arrivalTextField);
						
			add(arrivalTextField);
			
			arrivalTextField.setBounds(340, i * ROW_HEIGHT + 5, 180, 20);
			
			// departure textfield
			JTextField departureTextField = new JTextField();
			
			departureTextFields.add(departureTextField);
			
			add(departureTextField);
			
			departureTextField.setBounds(540, i * ROW_HEIGHT + 5, 180, 20);
			
			// setting default times
			
			Station currentStation = (Station)stations.get(i);
			
			if (previousStation == null)
			{
				arrivalTextField.setText("8:00");
				departureTextField.setText("+0:00");
			}
			else
			{
				int travelTime = TravelTime.GetTravelTime(previousStation.GetLocation(), currentStation.GetLocation(), meanOfTransportation);
				
				accumulatedTime += travelTime;
				
				arrivalTextField.setText(TravelTime.GetShiftTime(accumulatedTime));
				
				if (i < count - 1)
				{
					accumulatedTime += TravelTime.GetChangeTime(meanOfTransportation);
				
					departureTextField.setText(TravelTime.GetShiftTime(accumulatedTime));
				}
				else // it's the last one
				{
					departureTextField.setText(TravelTime.GetShiftTime(accumulatedTime));
				}
			}
			
			previousStation = currentStation;
		}
	}
	
	/**
	 * Loads the stations and a timetable from the connection.
	 * It is assumed the connection member is set before calling this method.
	 * @author Piotrek
	 */
	private void LoadConnection()
	{
		int count = stations.size();
						
		for (int i = 0; i < count; ++i)
		{				
			// arrival text field
			JTextField arrivalTextField = new JTextField();
			
			arrivalTextFields.add(arrivalTextField);
						
			add(arrivalTextField);
			
			arrivalTextField.setBounds(340, i * ROW_HEIGHT + 5, 180, 20);
			
			// departure textfield
			JTextField departureTextField = new JTextField();
			
			departureTextFields.add(departureTextField);
			
			add(departureTextField);
			
			departureTextField.setBounds(540, i * ROW_HEIGHT + 5, 180, 20);
			
			// setting times
			
			int arrivalTime = ((Integer)connection.GetArrivals().get(i)).intValue();
			int departureTime = ((Integer)connection.GetDepartures().get(i)).intValue();
			
			arrivalTextField.setText(TravelTime.GetTime(arrivalTime));
			departureTextField.setText(TravelTime.GetTime(departureTime));
		}
	}
		
	/**
	 * An ordered list of the arrival text fields.
	 * @author Piotrek
	 */
	protected ArrayList arrivalTextFields = new ArrayList();
	
	/**
	 * Gets an ordered list of arrival text fields.
	 * @author Piotrek
	 * @return An ordered list of arrival text fields.
	 */
	public ArrayList GetArrivalTextFields()
	{
		return arrivalTextFields;
	}
	
	/**
	 * An ordered list of the departure text fields.
	 * @author Piotrek
	 */
	private ArrayList departureTextFields = new ArrayList();
	
	/**
	 * Gets an ordered list of departure text fields.
	 * @author Piotrek
	 * @return An ordered list of arrival text fields.
	 */
	public ArrayList GetDepartureTextFields()
	{
		return departureTextFields;
	}
	
	/**
	 * Sets the arrival time for the station with the given index.
	 * @author Piotrek
	 * @param time A time to be set.
	 * @param index An index of a stations to set the time for.
	 */
	public void SetArrivalTime(String time, int index)
	{
		if (time == null)
			throw new NullPointerException("Time must not be null in StationsTimetableListView.SetArrivalTime().");
		
		if (index < 0 || index >= stations.size())
			throw new IllegalArgumentException("Index in StationsTimetableListView.SetArrivalTime() is out of range.");
		
		((JTextField)arrivalTextFields.get(index)).setText(time);
	}
	
	/**
	 * Sets the departure time for the station with the given index.
	 * @author Piotrek
	 * @param time A time to be set.
	 * @param index An index of a stations to set the time for.
	 */
	public void SetDepartureTime(String time, int index)
	{
		if (time == null)
			throw new NullPointerException("Time must not be null in StationsTimetableListView.SetDepartureTime().");
		
		if (index < 0 || index >= stations.size())
			throw new IllegalArgumentException("Index in StationsTimetableListView.SetDepartureTime() is out of range.");
		
		((JTextField)departureTextFields.get(index)).setText(time);
	}
	
	/**
	 * A height of a single row in the list view.
	 * @author Piotrek
	 */
	private static final int ROW_HEIGHT = 30;
		
	/**
	 * @see java.awt.Component#paint(java.awt.Graphics)
	 * @author Piotrek
	 */
	public void paint(Graphics g)
	{
		super.paint(g);
		
		int count = stations.size();
		
		for (int i = 0; i < count; ++i)
		{
			Station station = (Station)stations.get(i);
			
			g.drawString("#" + (i + 1), 30, i * ROW_HEIGHT + 17);
			g.setColor(Color.BLUE);
			g.drawString(station.name, 100, i * ROW_HEIGHT + 17);
			g.setColor(Color.BLACK);
			
			// frame
			g.drawLine(0, i * ROW_HEIGHT + 29, getWidth(), i * ROW_HEIGHT + 29);
			
			g.drawLine(78, i * ROW_HEIGHT, 78, i * ROW_HEIGHT + 30);
			g.drawLine(328, i * ROW_HEIGHT, 328, i * ROW_HEIGHT + 30);
			g.drawLine(528, i * ROW_HEIGHT, 528, i * ROW_HEIGHT + 30);
			g.drawLine(728, i * ROW_HEIGHT, 728, i * ROW_HEIGHT + 30);
		}
	}
}
