/*
 * Created on 2005-03-29
 */
package gui.controls;

import gui.dialogs.MainDialog;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Iterator;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import scheduler.ConnectionsFinderTask;
import scheduler.TravelPart;
import scheduler.TravelPartConnectionChunk;
import scheduler.TravelPartsList;
import world.Airport;
import world.BusStop;
import world.City;
import world.Connection;
import world.Harbour;
import world.Location;
import world.MeanOfTransportation;
import world.Station;
import world.TrainStation;
import world.World;

/**
 * A view displaying the current state og the associated world.
 * @author Piotrek
 */
public final class WorldView extends JPanel implements MouseListener
{	
	/**
	 * Constructor.
	 * @param parent The parent dialog.
	 * @author Piotrek
	 */
	public WorldView(MainDialog parent)
	{
		setLayout(null);
		setBackground(Color.BLUE);
		setPreferredSize(new Dimension(width, height));
		
		worldImage = Toolkit.getDefaultToolkit().getImage("./Images/WorldImage.jpg");		
		airportImage = Toolkit.getDefaultToolkit().getImage("./Images/Airport.gif");
		trainStationImage = Toolkit.getDefaultToolkit().getImage("./Images/TrainStation.gif");
		busStopImage = Toolkit.getDefaultToolkit().getImage("./Images/BusStop.gif");
		harbourImage = Toolkit.getDefaultToolkit().getImage("./Images/Harbour.gif");
		
		this.parent = parent;
	}
	
	/**
	 * The parent dialog.
	 * @author Piotrek
	 */
	private MainDialog parent = null;
	
	/**
	 * An image of the world (it is displayed in the background of the world).
	 * @author Piotrek
	 */
	private Image worldImage = null;
	
	/**
	 * An image of the airport symbol.
	 * @author Piotrek; 
	 */
	private Image airportImage = null;
	
	/**
	 * An image of the train station symbol.
	 * @author Piotrek
	 */
	private Image trainStationImage = null;
	
	/**
	 * An image of the bus stop symbol.
	 * @author Piotrek
	 */
	private Image busStopImage = null;
	
	/**
	 * An image of the harbour symbol.
	 * @author Piotrek
	 */
	private Image harbourImage = null;
	
	/**
	 * The width of the view. Multipile it by a scale factor to get a real width.
	 * @author Piotrek
	 */
	public static final int width = 1000;
	
	/**
	 * The height of the view. Multipile it by a scale factor to get a real height.
	 * @author Piotrek
	 */
	public static final int height = 1000;
	
	/**
	 * A scale of the display.
	 * @author Piotrek
	 */
	private int scale = 1;
	
	/**
	 * Indicates whether to draw the names of the objects in the view.
	 * @author Piotrek
	 */
	public boolean fShowNames = true;
	
	/**
	 * Indicates whether to draw the connections in the view.
	 * @author Piotrek
	 */
	public boolean fShowConnections = true;
	
	/**
	 * Indicates whether to draw the names of the airports in the view.
	 * @author Piotrek
	 */
	public boolean fShowAirportsNames = true;
	
	/**
	 * Indicates whether to draw the air connections in the view.
	 * @author Piotrek
	 */
	public boolean fShowAirConnections = true;
	
	/**
	 * Indicates whether to draw the names of the bus stops in the view.
	 * @author Piotrek
	 */
	public boolean fShowBusStopsNames = true;
	
	/**
	 * Indicates whether to draw the bus connections in the view.
	 * @author Piotrek
	 */
	public boolean fShowBusConnections = true;
	
	/**
	 * Indicates whether to draw the names of the train stations in the view.
	 * @author Piotrek
	 */
	public boolean fShowTrainStationsNames = true;
	
	/**
	 * Indicates whether to draw the railway connections in the view.
	 * @author Piotrek
	 */
	public boolean fShowRailwayConnections = true;
	
	/**
	 * Indicates whether to draw the names of the harbours in the view.
	 * @author Piotrek
	 */
	public boolean fShowHarboursNames = true;
	
	/**
	 * Indicates whether to draw the sea connections in the view.
	 * @author Piotrek
	 */
	public boolean fShowSeaConnections = true;
	
	/**
	 * Indicates whether to draw the names of the cities in the view.
	 * @author Piotrek
	 */
	public boolean fShowCitiesNames = true;
	
	/**
	 * Draws a line on the given graphics between the two specified points taking into account the fact that the West and East edges are the single edge (roundness of the world).
	 * @author Piotrek
	 * @param g The graphics to draw on.
	 * @param x1 The x coordinate of the first of the points.
	 * @param y1 The y coordinate of the first of the points.
	 * @param x2 The x coordinate of the second of the points.
	 * @param y2 The y coordinate of the second of the points.
	 */
	private void DrawLine(Graphics g, int x1, int y1, int x2, int y2)
	{
		if (g == null)
			throw new NullPointerException("G must not be null in WorldView.DrawLine().");
		
		if (IsDirectLine(x1, y1, x2, y2))
			g.drawLine(x1, y1, x2, y2);
		else
		{
			// drawing the line through the East/West edge			
			if (x1 > x2)
			{
				int temp;
				
				temp = x1;
				
				x1 = x2;
				x2 = temp;
				
				temp = y1;
				
				y1 = y2;
				y2 = temp;
			}
			
			int realWidth = width * scale;
						
			double dx = x1 + realWidth - x2;
			double dy = y2 - y1;
			
			if (dx != 0)
			{
				double tan = dy / dx;
				
				g.drawLine(x1, y1, 0, y1 + (int)(tan * x1));
				g.drawLine(x2, y2, realWidth, y2 - (int)(tan * (realWidth - x2)));
			}
			else
			{
				// this is a vertical line (both points are positioned on the East/West edge)
				g.drawLine(0, y1, 0, y2);
				g.drawLine(realWidth, y1, realWidth, y2);
			}
		}
	}
	
	/**
	 * Draws a string on the given graphics in the middle between the two specified points taking into account the fact that the West and East edges are the single edge (roundness of the world).
	 * @author Piotrek
	 * @param g The graphics to draw on.
	 * @param text The text to be displayed.
	 * @param p1 The first point.
	 * @param p2 The second point.
	 */
	private void DrawString(Graphics g, String text, Point p1, Point p2)
	{
		if (g == null)
			throw new NullPointerException("G must not be null in WorldView.DrawLine().");
		
		if (IsDirectLine(p1.x, p1.y, p2.x, p2.y))
			g.drawString(text, (p1.x + p2.x) / 2, (p1.y + p2.y) / 2);
		else
		{
			int x1 = p1.x;
			int x2 = p2.x;
			
			int y1 = p1.y;
			int y2 = p2.y;
			
			// drawing the string next to the two parts of the line going through the East/West edge			
			if (x1 > x2)
			{
				int temp;
				
				temp = x1;
				
				x1 = x2;
				x2 = temp;
				
				temp = y1;
				
				y1 = y2;
				y2 = temp;
			}
			
			int realWidth = width * scale;
						
			double dx = x1 + realWidth - x2;
			double dy = y2 - y1;
			
			if (dx != 0)
			{
				double tan = dy / dx;
				
				g.drawString(text, (x1) / 2, (2 * y1 + (int)(tan * x1)) / 2);
				g.drawString(text, (x2 + realWidth) / 2, (2 * y2 - (int)(tan * (realWidth - x2))) / 2);
			}
			else
			{
				// this is a vertical line (both points are positioned on the East/West edge)
				g.drawString(text, 0, (y1 + y2) / 2);
				g.drawString(text, realWidth, (y1 + y2) / 2);
			}
		}
	}
	
	/**
	 * Checks whether the line between the two given points is a direct line or maybe it goes through the Easr/West edge.
	 * @author Piotrek
	 * @param x1 The x coordinate of the first of the points.
	 * @param y1 The y coordinate of the first of the points.
	 * @param x2 The x coordinate of the second of the points.
	 * @param y2 The y coordinate of the second of the points.
	 * @return A boolean value indicating whether the line between the two given points is a direct line or maybe it goes through the Easr/West edge.
	 */
	private boolean IsDirectLine(int x1, int y1, int x2, int y2)
	{
		int realWidth = width * scale;
		
		// direct distance
		double directDistance = Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
		
		// East/West common edge distance
		double eastWestDistance = 0;
		
		if (x1 < x2)
			eastWestDistance = Math.sqrt((x1 + realWidth - x2) * (x1 + realWidth - x2) + (y1 - y2) * (y1 - y2));
		else
			eastWestDistance = Math.sqrt((x1 - x2 - realWidth) * (x1 - x2 - realWidth) + (y1 - y2) * (y1 - y2));
		
		return (directDistance <= eastWestDistance);
	}
	
	/**
	 * @see java.awt.Component#paint(java.awt.Graphics)
	 * @author Piotrek
	 */
	public void paint(Graphics g)
	{		
		if (fUpdateScrollers)
		{
			JScrollPane parentScrollPane = (JScrollPane)this.getParent().getParent();
			
			JScrollBar horizontalScrollBar = parentScrollPane.getHorizontalScrollBar();
			JScrollBar verticalScrollBar = parentScrollPane.getVerticalScrollBar();
			
			horizontalScrollBar.setValue(newHorizontalPosition);
			verticalScrollBar.setValue(newVerticalPosition);
			
			fUpdateScrollers = false;
		}
		
		g.drawImage(worldImage, 0, 0, width * scale, height * scale, 0, 0, 1000, 1000, this);
		
		World world = World.GetInstance();
				
		synchronized (world)
		{		
			// firstly drawing rectangles...
							
			// cities
			
			Iterator cities = world.GetCities();
			
			while (cities.hasNext())
			{
				City city = (City)cities.next();
				
				Point point = LocationToViewPoint(city.GetLocation());
				
				int halfEdge = CITY_HALF_SIDE * scale;
				
				g.setColor(Color.RED);
				g.fillRect(point.x - halfEdge, point.y - halfEdge, 2 * halfEdge, 2 * halfEdge);
			}
			
			// airports
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.AIRPLANE))
			{
				Iterator airports = world.GetAirports();
				
				while (airports.hasNext())
				{
					Airport airport = (Airport)airports.next();
					
					Point point = LocationToViewPoint(airport.GetLocation());
					
					int halfEdge = AIRPORT_HALF_SIDE * scale;
					
					g.setColor(Color.DARK_GRAY);
					g.fillRect(point.x - halfEdge, point.y - halfEdge, 2 * halfEdge, 2 * halfEdge);
				}
			}
			
			// train stations
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.TRAIN))
			{
				Iterator trainStations = world.GetTrainStations();
				
				while (trainStations.hasNext())
				{
					TrainStation trainStation = (TrainStation)trainStations.next();
					
					Point point = LocationToViewPoint(trainStation.GetLocation());
					
					int halfEdge = TRAIN_STATION_HALF_SIDE * scale;
					
					g.setColor(Color.DARK_GRAY);
					g.fillRect(point.x - halfEdge, point.y - halfEdge, 2 * halfEdge, 2 * halfEdge);
				}
			}
			
			// bus stops
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.BUS)) 
			{
				Iterator busStops = world.GetBusStops();
				
				while (busStops.hasNext())
				{
					BusStop busStop = (BusStop)busStops.next();
					
					Point point = LocationToViewPoint(busStop.GetLocation());
					
					int halfEdge = BUS_STOP_HALF_SIDE * scale;
					
					g.setColor(Color.DARK_GRAY);
					g.fillRect(point.x - halfEdge, point.y - halfEdge, 2 * halfEdge, 2 * halfEdge);
				}
			}
			
			// harbours
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.SHIP))
			{
				Iterator harbours = world.GetHarbours();
				
				while (harbours.hasNext())
				{
					Harbour harbour = (Harbour)harbours.next();
					
					Point point = LocationToViewPoint(harbour.GetLocation());
					
					int halfEdge = HARBOUR_HALF_SIDE * scale;
					
					g.setColor(Color.DARK_GRAY);
					g.fillRect(point.x - halfEdge, point.y - halfEdge, 2 * halfEdge, 2 * halfEdge);
				}
			}
			
			// secondly the symbols
			
			// airports
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.AIRPLANE))
			{
				Iterator airports = world.GetAirports();
				
				while (airports.hasNext())
				{
					Airport airport = (Airport)airports.next();
					
					Point point = LocationToViewPoint(airport.GetLocation());
					
					g.drawImage(airportImage, point.x - 10, point.y - 10, point.x + 10, point.y + 10, 0, 0, 20, 20, this);
				}
			}
			
			// train stations
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.TRAIN))
			{
				Iterator trainStations = world.GetTrainStations();
				
				while (trainStations.hasNext())
				{
					TrainStation trainStation = (TrainStation)trainStations.next();
					
					Point point = LocationToViewPoint(trainStation.GetLocation());
								
					g.drawImage(trainStationImage, point.x - 10, point.y - 10, point.x + 10, point.y + 10, 0, 0, 20, 20, this);
				}
			}
			
			// bus stops
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.BUS))
			{
				Iterator busStops = world.GetBusStops();
				
				while (busStops.hasNext())
				{
					BusStop busStop = (BusStop)busStops.next();
					
					Point point = LocationToViewPoint(busStop.GetLocation());
					
					g.drawImage(busStopImage, point.x - 10, point.y - 10, point.x + 10, point.y + 10, 0, 0, 20, 20, this);
				}
			}
			
			// harbours
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.SHIP))
			{
				Iterator harbours = world.GetHarbours();
				
				while (harbours.hasNext())
				{
					Harbour harbour = (Harbour)harbours.next();
					
					Point point = LocationToViewPoint(harbour.GetLocation());
					
					g.drawImage(harbourImage, point.x - 10, point.y - 10, point.x + 10, point.y + 10, 0, 0, 20, 20, this);
				}
			}
			
			// thirdly the names and marking
			
			// cities
			
			cities = world.GetCities();
			
			while (cities.hasNext())
			{
				City city = (City)cities.next();
				
				Point point = LocationToViewPoint(city.GetLocation());
				
				int halfEdge = CITY_HALF_SIDE * scale;
				
				if (fShowNames && fShowCitiesNames)
				{
					g.setColor(Color.ORANGE);
					
					g.drawString(city.name, point.x - 20, point.y + halfEdge + 12);
				}
				
				if (city == editedCity)
				{
					g.setColor(Color.ORANGE);
					
					g.drawRect(point.x - halfEdge - 5, point.y - halfEdge - 5, 2 * halfEdge + 9, 2 * halfEdge + 9);
				}
			}
			
			// airports
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.AIRPLANE))
			{
				Iterator airports = world.GetAirports();
				
				while (airports.hasNext())
				{
					Airport airport = (Airport)airports.next();
					
					Point point = LocationToViewPoint(airport.GetLocation());
					
					int halfEdge = AIRPORT_HALF_SIDE * scale;
					
					int delta;
					
					if (halfEdge > 15)
						delta = halfEdge + 5;
					else
						delta = 15;
					
					if (fShowNames && fShowAirportsNames)
					{
						g.setColor(Color.WHITE);
				
						g.drawString(airport.name, point.x - 15, point.y + delta + 7);
					}
					
					if (airport == editedAirport)
					{
						g.setColor(Color.ORANGE);
						
						g.drawRect(point.x - delta - 5, point.y - delta - 5, 2 * delta + 9, 2 * delta + 9);
					}
				}
			}
			
			// train stations
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.TRAIN))
			{
				Iterator trainStations = world.GetTrainStations();
			
				while (trainStations.hasNext())
				{
					TrainStation trainStation = (TrainStation)trainStations.next();
					
					Point point = LocationToViewPoint(trainStation.GetLocation());
					
					int halfEdge = TRAIN_STATION_HALF_SIDE * scale;
					
					int delta;
					
					if (halfEdge > 15)
						delta = halfEdge + 5;
					else
						delta = 15;
					
					if (fShowNames && fShowTrainStationsNames)		
					{
						g.setColor(Color.ORANGE);
					
						g.drawString(trainStation.name, point.x - 15, point.y + delta + 7);
					}
					
					if (trainStation == editedTrainStation)
					{
						g.setColor(Color.ORANGE);
						
						g.drawRect(point.x - delta - 5, point.y - delta - 5, 2 * delta + 9, 2 * delta + 9);
					}
				}
			}
			
			// bus stops
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.BUS))
			{
				Iterator busStops = world.GetBusStops();
				
				while (busStops.hasNext())
				{
					BusStop busStop = (BusStop)busStops.next();
					
					Point point = LocationToViewPoint(busStop.GetLocation());
					
					int halfEdge = BUS_STOP_HALF_SIDE * scale;
					
					int delta;
					
					if (halfEdge > 15)
						delta = halfEdge + 5;
					else
						delta = 15;
					
					if (fShowNames && fShowBusStopsNames)
					{
						g.setColor(Color.CYAN);
						
						g.drawString(busStop.name, point.x - 15, point.y + delta + 7);
					}
					
					if (busStop == editedBusStop)
					{
						g.setColor(Color.ORANGE);
						
						g.drawRect(point.x - delta - 5, point.y - delta - 5, 2 * delta + 9, 2 * delta + 9);
					}
				}
			}
			
			// harbours
			
			if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == MeanOfTransportation.SHIP))
			{
				Iterator harbours = world.GetHarbours();
				
				while (harbours.hasNext())
				{
					Harbour harbour = (Harbour)harbours.next();
					
					Point point = LocationToViewPoint(harbour.GetLocation());
					
					int halfEdge = HARBOUR_HALF_SIDE * scale;
					
					int delta;
					
					if (halfEdge > 15)
						delta = halfEdge + 5;
					else
						delta = 15;
					
					if (fShowNames && fShowHarboursNames)
					{
						g.setColor(Color.MAGENTA);
						
						g.drawString(harbour.name, point.x - 15, point.y + delta + 7);
					}
					
					if (harbour == editedHarbour)
					{
						g.setColor(Color.ORANGE);
						
						g.drawRect(point.x - delta - 5, point.y - delta - 5, 2 * delta + 9, 2 * delta + 9);
					}
				}
			}
			
			// drawing connections
			
			if (fShowConnections)
			{
				Iterator connections = world.GetConnections();
				
				while (connections.hasNext())
				{
					Connection connection = (Connection)connections.next();
					
					if (mode != MODE_CHOOSE_STATION || (mode == MODE_CHOOSE_STATION && stationsChooser.GetStationType() == connection.GetMeanOfTransportation()))
					{	
						if (mode == MODE_CONNECTION_EDITION)
						{
							Connection selectedConnection = connectionEditChooser.GetSelectedConnection();
							
							if (selectedConnection != null && selectedConnection != connection)
								continue;
						}
						
						// settings the colour of the line
						
						switch (connection.GetMeanOfTransportation())
						{
							case MeanOfTransportation.AIRPLANE:
								
								g.setColor(Color.WHITE);
							
								if (!fShowAirConnections)
									continue;
								
								break;
								
							case MeanOfTransportation.BUS:
								
								g.setColor(Color.YELLOW);
							
								if (!fShowBusConnections)
									continue;
							
								break;
								
							case MeanOfTransportation.TRAIN:
								
								g.setColor(Color.PINK);
							
								if (!fShowRailwayConnections)
									continue;
							
								break;
								
							case MeanOfTransportation.SHIP:
								
								g.setColor(Color.MAGENTA);
							
								if (!fShowSeaConnections)
									continue;
							
								break;
						}
						
						// drawing the connection
						
						ArrayList connectionStations = connection.GetStations();
						
						Iterator iterator = connectionStations.iterator();
						
						int index = 1;
						
						Point oldPoint = null;
						
						while (iterator.hasNext())
						{
							Station station = (Station)iterator.next();
							
							Point point = LocationToViewPoint(station.GetLocation());
																	
							if (index > 1)
								DrawLine(g, oldPoint.x, oldPoint.y, point.x, point.y);
							
							++index;
							
							oldPoint = point;
						}
					}
				}
			}
			
			// now the numbers on the connection list and line showing the route 
			// in the creation of a new connection mode
			
			if (mode == MODE_CHOOSE_STATION)
			{			
				ArrayList connectionStations = stationsChooser.GetStations();
				
				Iterator iterator = connectionStations.iterator();
				
				int index = 1;
				
				Point oldPoint = null;
				
				while (iterator.hasNext())
				{
					Station station = (Station)iterator.next();
					
					Point point = LocationToViewPoint(station.GetLocation());
									
					g.setColor(Color.CYAN);
						
					Integer integer = new Integer(index);
					g.drawString(integer.toString(), point.x + 15, point.y);
									
					if (index > 1)
						DrawLine(g, oldPoint.x, oldPoint.y, point.x, point.y);
					
					++index;
					
					oldPoint = point;
				}
			}
			
			// the origin and destination of the test travel
			if (locationsChooser.isVisible() || travelDivisionsChooser.isVisible() || travelSolutionsChooser.isVisible())
			{
				Location origin = locationsChooser.GetOrigin();
				
				if (origin != null)
				{
					Point point = LocationToViewPoint(origin);
					
					g.setColor(Color.YELLOW);
					
					g.drawLine(point.x - 5, point.y - 5, point.x + 5, point.y + 5);
					g.drawLine(point.x - 5, point.y + 5, point.x + 5, point.y - 5);
					
					g.drawString("Origin", point.x - 15, point.y + 20);
				}
				
				Location destination = locationsChooser.GetDestination();
				
				if (destination != null)
				{
					Point point = LocationToViewPoint(destination);
					
					g.setColor(Color.YELLOW);
					
					g.drawLine(point.x - 5, point.y - 5, point.x + 5, point.y + 5);
					g.drawLine(point.x - 5, point.y + 5, point.x + 5, point.y - 5);
					
					g.drawString("Destination", point.x - 30, point.y + 20);
				}
			}
			
			// travel divisions
			if (travelDivisionsChooser.isVisible())
			{
				TravelPartsList division = travelDivisionsChooser.GetSelectedDivision();
				
				if (division != null)
				{
					Iterator iterator = division.GetTravelParts();
					
					int counter = 1;
					
					while (iterator.hasNext())
					{
						TravelPart travelPart = (TravelPart)iterator.next();
						
						Point origin = LocationToViewPoint(travelPart.GetOrigin());
						Point destination = LocationToViewPoint(travelPart.GetDestination());
						
						g.setColor(new Color(50, 255, 50));
						
						DrawLine(g, origin.x, origin.y, destination.x, destination.y);
						
						g.setColor(new Color(255, 255, 255));
						
						String meanOfTransportation = MeanOfTransportation.GetMeanOfTransportationString(travelPart.GetMeanOfTransportation());
						
						DrawString(g, counter + ": " + meanOfTransportation, origin, destination);
						
						++counter;
					}
				}
			}
			
			// travel solutions
			if (travelSolutionsChooser.isVisible())
			{
				ConnectionsFinderTask solution = travelSolutionsChooser.GetSelectedSolution();
				
				if (solution != null)
				{
					Iterator iterator = solution.GetConnectionChunks();
					
					int counter = 1;
					
					while (iterator.hasNext())
					{
						TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
						
						Connection connection = chunk.GetRealConnection();
						
						if (connection == null)
						{
							Point origin = LocationToViewPoint(chunk.GetOrigin());
							Point destination = LocationToViewPoint(chunk.GetDestination());
							
							g.setColor(new Color(50, 255, 50));
							
							DrawLine(g, origin.x, origin.y, destination.x, destination.y);
							
							g.setColor(new Color(255, 255, 255));
							
							String meanOfTransportation = MeanOfTransportation.GetMeanOfTransportationString(chunk.GetMeanOfTransportation());
							
							String details = "";
							
							DateFormat formater = DateFormat.getInstance();
							
							details += " [" + formater.format(chunk.GetStartDate()) + " - " + formater.format(chunk.GetEndDate()) + "]";
							
							DrawString(g, counter + ": " + meanOfTransportation + details, origin, destination);
							
							++counter;
						}
						else
						{							
							int startStationIndex = chunk.GetOriginStationIndex();
							int endStationIndex = chunk.GetDestinationStationIndex();
							
							int indexWithLabel = (startStationIndex + endStationIndex) / 2;
							
							for (int i = startStationIndex; i < endStationIndex; ++i)
							{
								Point origin = LocationToViewPoint(((Station)connection.GetStations().get(i)).GetLocation());
								Point destination = LocationToViewPoint(((Station)connection.GetStations().get(i + 1)).GetLocation());
								
								g.setColor(new Color(50, 255, 50));
								
								DrawLine(g, origin.x, origin.y, destination.x, destination.y);
																
								g.setColor(new Color(255, 255, 255));
								
								String meanOfTransportation = MeanOfTransportation.GetMeanOfTransportationString(chunk.GetMeanOfTransportation());
								
								String details = ", " + connection.GetName();
								
								DateFormat formater = DateFormat.getInstance();
								
								if (i == indexWithLabel)
									details += " [" + formater.format(chunk.GetStartDate()) + " - " + formater.format(chunk.GetEndDate()) + "]";
								
								DrawString(g, counter + ": " + meanOfTransportation + details, origin, destination);									
							}
							
							++counter;
						}
					}
				}
			}
		}
	}
	
	/**
	 * A length in pixels (with scale of 1) of a half of length of a side of a square representing a city.
	 * @author Piotrek
	 */
	private static final int CITY_HALF_SIDE = 8;
	
	/**
	 * A length in pixels (with scale of 1) of a half of length of a side of a square representing an airport.
	 * @author Piotrek
	 */
	private static final int AIRPORT_HALF_SIDE = 2;
	
	/**
	 * A length in pixels (with scale of 1) of a half of length of a side of a square representing a train station.
	 * @author Piotrek
	 */
	private static final int TRAIN_STATION_HALF_SIDE = 1;
	
	/**
	 * A length in pixels (with scale of 1) of a half of length of a side of a square representing a bus stop.
	 * @author Piotrek
	 */
	private static final int BUS_STOP_HALF_SIDE = 1;
	
	/**
	 * A length in pixels (with scale of 1) of a half of length of a side of a square representing a harbour.
	 * @author Piotrek
	 */
	private static final int HARBOUR_HALF_SIDE = 3;
	
	/**
	 * Defines a maximal value for a scale. In practice it defines the maximal zoom. It should be a power of 2.
	 * @author Piotrek
	 */
	private static final int MAX_SCALE = 32;
	
	/**
	 * Zooms in the view of the world. It becomes more clear.
	 * @author Piotrek
	 */
	public void ZoomIn()
	{
		if (scale >= MAX_SCALE)
			return;
		
		// reading the current position of scrollers and calculating the new ones
		JScrollPane parentScrollPane = (JScrollPane)this.getParent().getParent();
		
		JScrollBar horizontalScrollBar = parentScrollPane.getHorizontalScrollBar();
		JScrollBar verticalScrollBar = parentScrollPane.getVerticalScrollBar();
				
		newHorizontalPosition = 2 * horizontalScrollBar.getValue() + parentScrollPane.getWidth() / 2;
		newVerticalPosition = 2 * verticalScrollBar.getValue() + parentScrollPane.getHeight() / 2;
		
		if (newHorizontalPosition >= width * scale * 2 - parentScrollPane.getWidth())
			newHorizontalPosition = width * scale * 2 - parentScrollPane.getWidth();
		
		if (newVerticalPosition >= height * scale * 2 - parentScrollPane.getHeight())
			newVerticalPosition = height * scale * 2 - parentScrollPane.getHeight();
		
		// changing the scale
		scale *= 2;
				
		setPreferredSize(new Dimension(width * scale, height * scale));
		
		revalidate();
		
		fUpdateScrollers = true;
	}
	
	/**
	 * The new position of the horizontal scroller to be set.
	 * @author Piotrek
	 */
	private int newHorizontalPosition = 0;
	
	/**
	 * The new postion of the vertical scroller to be set.
	 * @author Piotrek
	 */
	private int newVerticalPosition = 0;
	
	/**
	 * Indicates whether to update the scrollers positions in the paint method.
	 * @author Piotrek
	 */
	private boolean fUpdateScrollers = false;
	
	/**
	 * Zooms out the view of the world. It becomes less clear.
	 * @author Piotrek
	 */
	public void ZoomOut()
	{
		if (scale == 1)
			return;
		
		// reading the current position of scrollers and calculating the new ones
		JScrollPane parentScrollPane = (JScrollPane)this.getParent().getParent();
		
		JScrollBar horizontalScrollBar = parentScrollPane.getHorizontalScrollBar();
		JScrollBar verticalScrollBar = parentScrollPane.getVerticalScrollBar();
		
		newHorizontalPosition = (horizontalScrollBar.getValue() - parentScrollPane.getWidth() / 2) / 2;
		newVerticalPosition = (verticalScrollBar.getValue() - parentScrollPane.getHeight() / 2) / 2;
		
		if (newHorizontalPosition >= width * scale / 2 - parentScrollPane.getWidth())
			newHorizontalPosition = width * scale / 2 - parentScrollPane.getWidth();
		
		if (newVerticalPosition >= height * scale / 2 - parentScrollPane.getHeight())
			newVerticalPosition = height * scale / 2 - parentScrollPane.getHeight();
		
		// changing the scale
		
		scale /= 2;
				
		setPreferredSize(new Dimension(width * scale, height * scale));
		
		revalidate();
		
		fUpdateScrollers = true;
	}
	
	/**
	 * A mode the view is in.
	 * @author Piotrek
	 */
	private int mode;
	
	/**
	 * Sets the mode of this view. Only changes to valid modes are enforced.
	 * @author Piotrek
	 * @param mode
	 */
	public void SetMode(int mode)
	{
		if (mode == MODE_NORMAL_VIEW)
			this.mode = MODE_NORMAL_VIEW;
		else
		if (mode == MODE_ADD_CITY)
			this.mode = MODE_ADD_CITY;
		else
		if (mode == MODE_EDIT_CITY)
			this.mode = MODE_EDIT_CITY;
		else
		if (mode == MODE_ADD_AIRPORT)
			this.mode = MODE_ADD_AIRPORT;
		else
		if (mode == MODE_EDIT_AIRPORT)
			this.mode = MODE_EDIT_AIRPORT;
		else
		if (mode == MODE_ADD_TRAIN_STATION)
			this.mode = MODE_ADD_TRAIN_STATION;
		else
		if (mode == MODE_EDIT_TRAIN_STATION)
			this.mode = MODE_EDIT_TRAIN_STATION;
		else
		if (mode == MODE_ADD_BUS_STOP)
			this.mode = MODE_ADD_BUS_STOP;
		else
		if (mode == MODE_EDIT_BUS_STOP)
			this.mode = MODE_EDIT_BUS_STOP;
		else
		if (mode == MODE_ADD_HARBOUR)
			this.mode = MODE_ADD_HARBOUR;
		else
		if (mode == MODE_EDIT_HARBOUR)
			this.mode = MODE_EDIT_HARBOUR;
		else
		if (mode == MODE_CHOOSE_STATION)
			this.mode = MODE_CHOOSE_STATION;
		else
		if (mode == MODE_CONNECTION_EDITION)
			this.mode = MODE_CONNECTION_EDITION;
		else
		if (mode == MODE_ORIGIN_DESTINATION_CHOOSING)
			this.mode = MODE_ORIGIN_DESTINATION_CHOOSING;
		else
			throw new IllegalArgumentException("It is not a valid mode in WorldView.SetMode().");
	}
	
	/**
	 * Indicates that a view is in a `normal view' mode. It is a mode for browsing the map without any special actions available.
	 * @author Piotrek
	 */
	public static final int MODE_NORMAL_VIEW = 0;
	
	/**
	 * Indicates that a view is in a `add city' mode. It is a mode for adding a city to the world. 
	 * A view is awaitng for a click and places a city where it has been clicked on. 
	 * @author Piotrek
	 */
	public static final int MODE_ADD_CITY = 1;
	
	/**
	 * Indicates that a view is in a `edit city' mode. It is a mode for editing a city properties. 
	 * A click on a city opens a property pane for this city with two additional buttons on it:
	 * `apply changes' and `remove city'. A city being edited is marked on a view.
	 * @author Piotrek
	 */
	public static final int MODE_EDIT_CITY = 2;
	
	/**
	 * Indicates that a view is in a `add airport' mode. It is a mode for adding an airport to the world. 
	 * A view is awaitng for a click and places an airport where it has been clicked on. 
	 * @author Piotrek
	 */
	public static final int MODE_ADD_AIRPORT = 3;
	
	/**
	 * Indicates that a view is in a `edit airport' mode. It is a mode for editing an airport properties. 
	 * A click on an airport opens a property pane for this airport with two additional buttons on it:
	 * `apply changes' and `remove airport'. An airport being edited is marked on a view.
	 * @author Piotrek
	 */
	public static final int MODE_EDIT_AIRPORT = 4;
	
	/**
	 * Indicates that a view is in a `add train station' mode. It is a mode for adding a train station to the world. 
	 * A view is awaitng for a click and places a train station where it has been clicked on. 
	 * @author Piotrek
	 */
	public static final int MODE_ADD_TRAIN_STATION = 5;
	
	/**
	 * Indicates that a view is in a `edit train station' mode. It is a mode for editing a train station properties. 
	 * A click on a train station opens a property pane for this train station with two additional buttons on it:
	 * `apply changes' and `remove tain station'. A train station being edited is marked on a view.
	 * @author Piotrek
	 */
	public static final int MODE_EDIT_TRAIN_STATION = 6;
	
	/**
	 * Indicates that a view is in a `add bus stop' mode. It is a mode for adding a bus stop to the world. 
	 * A view is awaitng for a click and places a bus stop where it has been clicked on. 
	 * @author Piotrek
	 */
	public static final int MODE_ADD_BUS_STOP = 7;
	
	/**
	 * Indicates that a view is in a `edit bus stop' mode. It is a mode for editing a bus stop properties. 
	 * A click on a bus stop opens a property pane for this bus stop with two additional buttons on it:
	 * `apply changes' and `remove bus stop'. A bus stop being edited is marked on a view.
	 * @author Piotrek
	 */
	public static final int MODE_EDIT_BUS_STOP = 8;
	
	/**
	 * Indicates that a view is in a `add harbour' mode. It is a mode for adding a harbour to the world. 
	 * A view is awaitng for a click and places a harbour where it has been clicked on. 
	 * @author Piotrek
	 */
	public static final int MODE_ADD_HARBOUR = 9;
	
	/**
	 * Indicates that a view is in a `edit harbour' mode. It is a mode for editing a harbour properties. 
	 * A click on a bus stop opens a property pane for this harbour with two additional buttons on it:
	 * `apply changes' and `remove harbour'. A harbour being edited is marked on a view.
	 * @author Piotrek
	 */
	public static final int MODE_EDIT_HARBOUR = 10;
	
	/**
	 * Indicates that a view is in `choose station' mode. It is a mode for choosing stations when creating
	 * a new connection. View checks what the type of station is needed, and if such one is clicked notifies 
	 * the stations chooser control.
	 * @author Piotrek
	 */
	public static final int MODE_CHOOSE_STATION = 11;
	
	/**
	 * Indicates that a view is in `connection edition' mode. It is a mode for editing connections
	 * a view is displaying a currently selected connection only or all if there is none seclected.
	 * @author Piotrek
	 */
	public static final int MODE_CONNECTION_EDITION = 12;
	
	/**
	 * Indicates that a view is in `chossing the origin and destination' mode. It is a mode 
	 * enabling user toi click the two locations (the origin and the destination of the travel)
	 * A view is displaying a currently selected location if there is any and sends the 
	 * notifications to the locations chooser panel.
	 * @author Piotrek
	 */
	public static final int MODE_ORIGIN_DESTINATION_CHOOSING = 13;
	
	/**
	 * A universal stations chooser to be notified about selected station. Null if there is none.
	 * @author Piotrek
	 */
	private StationsChooser stationsChooser = null;
	
	/**
	 * Sets the universal stations chooser to be notified about selected station.
	 * @author Piotrek
	 * @param stationChooser A universal stations chooser to be notified about selected station.
	 */
	public void SetUniversalStationsChooser(StationsChooser stationChooser)
	{
		if (stationChooser == null)
			throw new NullPointerException("Stations chooser must not be null in WorldView.SetUniversalStationsChooser().");
		
		this.stationsChooser = stationChooser;
	}
	
	/**
	 * Translates the point of a view to the location.
	 * @author Piotrek
	 * @param point A point in the view area.
	 * @return The location at the given point.
	 */
	Location ViewPointToLocation(Point point)
	{
		if (point == null)
			throw new NullPointerException("Point must not be null in WorldView.ViewPointToLocation().");
		
		int x = (int)((float)point.x / (float)(scale * width) * 1000.0f);
		int y = (int)((float)point.y / (float)(scale * height) * 1000.0f);
		
		return new Location(x, y);
	}
	
	/**
	 * Translates the location to the point on the view area (in the current scale).
	 * @author Piotrek
	 * @param location
	 * @return The translated point.
	 */
	Point LocationToViewPoint(Location location)
	{
		if (location == null)
			throw new NullPointerException("Location must not be null in WorldView.LocationToViewPoint().");
		
		int x = (int)(location.x / 1000.0f * width * scale + scale / 2.0f);
		int y = (int)(location.y / 1000.0f * height * scale + scale / 2.0f);
		
		return new Point(x, y);
	}
	
	/**
	 * @author Piotrek
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		synchronized (World.GetInstance())
		{
			// city
			
			if (mode == MODE_ADD_CITY)
			{
				Location location = ViewPointToLocation(e.getPoint());
							
				City city = new City(location, "Anonymous City");
				
				parent.GetCityPropertiesPanel().StoreData(city);
					
				World.GetInstance().AddCity(city);
					
				SetMode(MODE_NORMAL_VIEW);
				
				parent.GetCityPropertiesPanel().setVisible(false);
					
				invalidate();
				repaint();
			}
			
			if (mode == MODE_EDIT_CITY)
			{
				Point clickedPoint = e.getPoint();
				
				Iterator iterator = World.GetInstance().GetCities();
				
				editedCity = null;
				
				while (iterator.hasNext())
				{
					City city = (City)iterator.next();
					
					Location location = city.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = CITY_HALF_SIDE * scale;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						editedCity = city;
						
						break;
					}
				}
				
				if (editedCity != null)
				{
					parent.GetCityPropertiesPanel().setBounds(10, 10, 180, 295);
					
					parent.GetCityPropertiesPanel().LoadData(editedCity);
					
					parent.GetCityPropertiesPanel().setVisible(true);
					
					repaint();
				}
			}
			
			// airport
			
			if (mode == MODE_ADD_AIRPORT)
			{
				Location location = ViewPointToLocation(e.getPoint());
							
				Airport airport = new Airport(location, "Anonymous Airport");
				
				parent.GetAirportPropertiesPanel().StoreData(airport);
					
				World.GetInstance().AddAirport(airport);
					
				SetMode(MODE_NORMAL_VIEW);
				
				parent.GetAirportPropertiesPanel().setVisible(false);
					
				invalidate();
				repaint();
			}
			
			if (mode == MODE_EDIT_AIRPORT)
			{
				Point clickedPoint = e.getPoint();
				
				Iterator iterator = World.GetInstance().GetAirports();
				
				editedAirport = null;
				
				while (iterator.hasNext())
				{
					Airport airport = (Airport)iterator.next();
					
					Location location = airport.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = AIRPORT_HALF_SIDE * scale;
					
					if (delta < 10)
						delta = 10;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						editedAirport = airport;
						
						break;
					}
				}
				
				if (editedAirport != null)
				{
					parent.GetAirportPropertiesPanel().setBounds(10, 10, 180, 295);
					
					parent.GetAirportPropertiesPanel().LoadData(editedAirport);
					
					parent.GetAirportPropertiesPanel().setVisible(true);
					
					repaint();
				}
			}
			
			// train station
			
			if (mode == MODE_ADD_TRAIN_STATION)
			{
				Location location = ViewPointToLocation(e.getPoint());
							
				TrainStation trainStation = new TrainStation(location, "Anonymous Train Station");
				
				parent.GetTrainStationPropertiesPanel().StoreData(trainStation);
					
				World.GetInstance().AddTrainStation(trainStation);
					
				SetMode(MODE_NORMAL_VIEW);
				
				parent.GetTrainStationPropertiesPanel().setVisible(false);
					
				invalidate();
				repaint();
			}
			
			if (mode == MODE_EDIT_TRAIN_STATION)
			{
				Point clickedPoint = e.getPoint();
				
				Iterator iterator = World.GetInstance().GetTrainStations();
				
				editedTrainStation = null;
				
				while (iterator.hasNext())
				{
					TrainStation trainStation = (TrainStation)iterator.next();
					
					Location location = trainStation.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = TRAIN_STATION_HALF_SIDE * scale;
					
					if (delta < 10)
						delta = 10;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						editedTrainStation = trainStation;
						
						break;
					}
				}
				
				if (editedTrainStation != null)
				{
					parent.GetTrainStationPropertiesPanel().setBounds(10, 10, 180, 295);
					
					parent.GetTrainStationPropertiesPanel().LoadData(editedTrainStation);
					
					parent.GetTrainStationPropertiesPanel().setVisible(true);
					
					repaint();
				}
			}
			
			// bus stop
			
			if (mode == MODE_ADD_BUS_STOP)
			{
				Location location = ViewPointToLocation(e.getPoint());
							
				BusStop busStop = new BusStop(location, "Anonymous Bus Stop");
				
				parent.GetBusStopPropertiesPanel().StoreData(busStop);
					
				World.GetInstance().AddBusStop(busStop);
					
				SetMode(MODE_NORMAL_VIEW);
				
				parent.GetBusStopPropertiesPanel().setVisible(false);
					
				invalidate();
				repaint();
			}
			
			if (mode == MODE_EDIT_BUS_STOP)
			{
				Point clickedPoint = e.getPoint();
				
				Iterator iterator = World.GetInstance().GetBusStops();
				
				editedBusStop = null;
				
				while (iterator.hasNext())
				{
					BusStop busStop = (BusStop)iterator.next();
					
					Location location = busStop.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = BUS_STOP_HALF_SIDE * scale;
					
					if (delta < 10)
						delta = 10;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						editedBusStop = busStop;
						
						break;
					}
				}
				
				if (editedBusStop != null)
				{
					parent.GetBusStopPropertiesPanel().setBounds(10, 10, 180, 295);
					
					parent.GetBusStopPropertiesPanel().LoadData(editedBusStop);
					
					parent.GetBusStopPropertiesPanel().setVisible(true);
					
					repaint();
				}
			}
			
			// harbour
			
			if (mode == MODE_ADD_HARBOUR)
			{
				Location location = ViewPointToLocation(e.getPoint());
							
				Harbour harbour = new Harbour(location, "Anonymous Harbour");
				
				parent.GetHarbourPropertiesPanel().StoreData(harbour);
					
				World.GetInstance().AddHarbour(harbour);
					
				SetMode(MODE_NORMAL_VIEW);
				
				parent.GetHarbourPropertiesPanel().setVisible(false);
					
				invalidate();
				repaint();
			}
			
			if (mode == MODE_EDIT_HARBOUR)
			{
				Point clickedPoint = e.getPoint();
				
				Iterator iterator = World.GetInstance().GetHarbours();
				
				editedHarbour = null;
				
				while (iterator.hasNext())
				{
					Harbour harbour = (Harbour)iterator.next();
					
					Location location = harbour.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = HARBOUR_HALF_SIDE * scale;
					
					if (delta < 10)
						delta = 10;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						editedHarbour = harbour;
						
						break;
					}
				}
				
				if (editedHarbour != null)
				{
					parent.GetHarbourPropertiesPanel().setBounds(10, 10, 180, 295);
					
					parent.GetHarbourPropertiesPanel().LoadData(editedHarbour);
					
					parent.GetHarbourPropertiesPanel().setVisible(true);
					
					repaint();
				}
			}
			
			// choosing of a station
			
			if (mode == MODE_CHOOSE_STATION)
			{
				Iterator iterator = null;
				
				int sideSize = 0;
				
				switch (stationsChooser.GetStationType())
				{
					case MeanOfTransportation.AIRPLANE:
				
						iterator = World.GetInstance().GetAirports();
					
						sideSize = AIRPORT_HALF_SIDE;
						
						break;
					
					case MeanOfTransportation.TRAIN:
						
						iterator = World.GetInstance().GetTrainStations();
					
						sideSize = TRAIN_STATION_HALF_SIDE;
						
						break;
					
					case MeanOfTransportation.BUS:
						
						iterator = World.GetInstance().GetBusStops();
					
						sideSize = BUS_STOP_HALF_SIDE;
						
						break;
					
					case MeanOfTransportation.SHIP:
						
						iterator = World.GetInstance().GetHarbours();
					
						sideSize = HARBOUR_HALF_SIDE;
						
						break;
				}
				
				Point clickedPoint = e.getPoint();
				
				while (iterator.hasNext())
				{
					Station station = (Station)iterator.next();
					
					Location location = station.GetLocation();
					
					Point point = LocationToViewPoint(location);
					
					int delta = sideSize * scale;
					
					if (delta < 10)
						delta = 10;
					
					if (clickedPoint.x >= point.x - delta && clickedPoint.x <= point.x + delta && clickedPoint.y >= point.y - delta && clickedPoint.y <= point.y + delta)
					{
						stationsChooser.AddStation(station);
						
						invalidate();
						
						return;
					}
				}
			}
			
			if (mode == MODE_ORIGIN_DESTINATION_CHOOSING)
			{
				Location location = ViewPointToLocation(e.getPoint());
				
				locationsChooser.ReceiveLocation(location);
			}
		}
	}
	
	/**
	 * A city being edited. Null if there is none.
	 * @author Piotrek
	 */
	public City editedCity = null;
	
	/**
	 * An airport being edited. Null if there is none.
	 * @author Piotrek
	 */
	public Airport editedAirport = null;
	
	/**
	 * A train station being edited. Null if there is none.
	 * @author Piotrek
	 */
	public TrainStation editedTrainStation = null;
	
	/**
	 * A bus stop being edited. Null if there is none.
	 * @author Piotrek
	 */
	public BusStop editedBusStop = null;
	
	/**
	 * A harbour being edited. Null if there is none.
	 * @author Piotrek
	 */
	public Harbour editedHarbour = null;
	
	/**
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 * @author Piotrek
	 */
	public void mouseEntered(MouseEvent e)
	{
		// ignoring
	}
	
	/**
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 * @author Piotrek
	 */
	public void mouseExited(MouseEvent e)
	{
		// ignoring
	}
	
	/**
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 * @author Piotrek
	 */
	public void mousePressed(MouseEvent e)
	{
		// ignoring
	}
	
	/**
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 * @author Piotrek
	 */
	public void mouseReleased(MouseEvent arg0)
	{
		// ignoring
	}
	
	/**
	 * A chooser of a connection to be edited.
	 * Set by the ConnectionsEditChooser.
	 * @author Piotrek
	 */
	private ConnectionEditChooser connectionEditChooser = null;

	/**
	 * Sets a chooser of a connection to be edited.
	 * @author Piotrek
	 * @param connectionEditChooser A chooser of a connection to be edited to be set.
	 */
	public void SetConnectionEditChooser(ConnectionEditChooser connectionEditChooser)
	{
		if (connectionEditChooser == null)
			throw new NullPointerException("Chooser of a connection to be edited must not be null in WorldView.SetConnectionEditChooser().");
		
		this.connectionEditChooser = connectionEditChooser;
	}
	
	/**
	 * A locations chooser to receive the origin/destination location notifications.
	 * @author Piotrek
	 */
	private LocationsChooser locationsChooser = null;
	
	/**
	 * Sets the locations chooser.
	 * @author Piotrek
	 * @param locationsChooser A locations chooser to be set.
	 */
	public void SetLocationsChooser(LocationsChooser locationsChooser)
	{
		if (locationsChooser == null)
			throw new NullPointerException("The locations chooser must not be null in WorldView.SetLocationsChooser().");
		
		this.locationsChooser = locationsChooser;
	}
	
	/**
	 * A list used to choose one of the travel divisions.
	 * @author Piotrek
	 */
	private TravelDivisionsChooser travelDivisionsChooser = null;
	
	/**
	 * Sets a list used to choose one of the travel divisions.
	 * @author Piotrek
	 * @param travelDivisionsChooser A list to be set.
	 */
	public void SetTravelDivisionChooser(TravelDivisionsChooser travelDivisionsChooser)
	{
		this.travelDivisionsChooser = travelDivisionsChooser;
	}
	
	/**
	 * A list used to choose one of the travel solutions.
	 * @author Piotrek
	 */
	private TravelSolutionsChooser travelSolutionsChooser = null;
	
	/**
	 * Sets a list used to choose one of the travel solutions.
	 * @author Piotrek
	 * @param travelSolutionsChooser A list to be set.
	 */
	public void SetTravelSolutionsChooser(TravelSolutionsChooser travelSolutionsChooser)
	{
		this.travelSolutionsChooser = travelSolutionsChooser;
	}
	
	/**
	 * Cancels the marking of the objects being edited.
	 * @author Piotrek
	 */
	public void CancelEditionMarkings()
	{
		editedAirport = null;
		editedBusStop = null;
		editedCity = null;
		editedHarbour = null;
		editedTrainStation = null;
	}
}
