/*
 * Created on 2005-04-08
 */
package gui.dialogs;

import gui.TravelSchedulerWorldEditor;
import gui.controls.StationsTimetableEditor;
import java.awt.Color;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Iterator;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import world.Connection;
import world.MeanOfTransportation;
import world.TravelTime;
import world.World;

/**
 * A dialog used to define a timetable for a connection. A list of stations and their order
 * is already known at the time of creating a timetable.
 * @author Piotrek
 */
public final class ConnectionTimetableCreator extends JDialog implements ActionListener
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param title A title of the dialog window.
	 * @param connectionName A name the new connection is to be given. Ignored if connection is not null.
	 * @param stations A list of stations.
	 * @param meanOfTransportation A mean of transportation. This is required only to count the default times of travel. Ignored if connection is not null.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @param connection A connection to be edited. Set it to null if creating a new connection.
	 */
	public ConnectionTimetableCreator(String title, String connectionName, ArrayList stations, int meanOfTransportation, Connection connection)
	{
		super();
				
		if (title == null)
			throw new NullPointerException("Title must not be null in ConnectionTimetableCreator.ConnectionTimetableCreator().");
		
		if (connectionName == null)
			throw new NullPointerException("Connection name must not be null in ConnectionTimetableCreator.ConnectionTimetableCreator().");
		
		if (connection != null)
		{
			connectionName = connection.GetName();
			
			stations = connection.GetStations();
			
			meanOfTransportation = connection.GetMeanOfTransportation();
		}
				
		//if (connectionName.compareTo("") == 0)
		///	throw new IllegalArgumentException("Connection name must not be empty in ConnectionTimetableCreator.ConnectionTimetableCreator().");
		
		
		if (stations == null)
			throw new NullPointerException("Stations list must not be null in ConnectionTimetableCreator.ConnectionTimetableCreator().");
		
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("Mean of transportation must be a valid one in ConnectionTimetableCreator.ConnectionTimetableCreator().");
		
		setModal(true);
		setTitle(title);
		setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		setSize(800, 700);
		setResizable(false);
		setBackground(Color.YELLOW);
		
		// panel	
		panel = getContentPane();
		
		panel.setLayout(null);
		
		setContentPane(panel);
		
		// ok and cancel buttons
		
		AddOKCancel();
		
		// timetable editor
		timetableEditor = new StationsTimetableEditor(stations, meanOfTransportation, connection);
		
		timetableEditor.setBounds(10, 10, 770, 510);
		
		add(timetableEditor);
		
		// repetition options
		AddRepetitionOptions(connection == null);
		
		stationsCount = stations.size();
		
		this.connectionName = connectionName;
		
		this.stations = stations;
		
		this.meanOfTransportation = meanOfTransportation;
		
		if (connection != null)
		{
			int dayOfWeekMask = connection.GetDayOfWeekMask();
			
			mondayCheckbox.setSelected((dayOfWeekMask & Connection.MONDAY) > 0);
			tuesdayCheckbox.setSelected((dayOfWeekMask & Connection.TUESDAY) > 0);
			wednesdayCheckbox.setSelected((dayOfWeekMask & Connection.WEDNESDAY) > 0);
			thursdayCheckbox.setSelected((dayOfWeekMask & Connection.THURSDAY) > 0);
			fridayCheckbox.setSelected((dayOfWeekMask & Connection.FRIDAY) > 0);
			saturdayCheckbox.setSelected((dayOfWeekMask & Connection.SATURDAY) > 0);
			sundayCheckbox.setSelected((dayOfWeekMask & Connection.SUNDAY) > 0);
		}
	}
	
	/**
	 * A name the new connection is to be given.
	 * @author Piotrek
	 */
	private String connectionName = null;
	
	/**
	 * A mean of transportation for a new connection.
	 * @author Piotrek
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 */
	private int meanOfTransportation;
	
	/**
	 * A count of stations in the timetable.
	 * @author Piotrek
	 */
	private int stationsCount = -1;
	
	/**
	 * A list of stations.
	 * @author Piotrek
	 */
	private ArrayList stations = null;
	
	/**
	 * An editor for times of arrivals and departures per stations.
	 * @author Piotrek
	 */
	private StationsTimetableEditor timetableEditor = null;
	
	/**
	 * An OK button.
	 * @author Piotrek
	 */
	private JButton okButton = null;
	
	/**
	 * A cancel button.
	 * @author Piotrek
	 */
	private JButton cancelButton = null;
	
	/**
	 * The OK dialog result.
	 * @author Piotrek
	 */
	public static final int OK = 1;
	
	/**
	 * The cancel dialog result.
	 * @author Piotrek
	 */
	public static final int CANCEL = 2;
	
	/**
	 * The dialog result.
	 * @author Piotrek
	 */
	private int dialogResult = CANCEL;
	
	/**
	 * Gets the dialog result.
	 * @author Piotrek
	 * @return The dialog result.
	 */
	public int GetDialogResult()
	{
		return dialogResult;
	}
	
	/**
	 * A main panel of this dialog.
	 * @author Piotrek
	 */
	private Container panel = null;
		
	/**
	 * @author Piotrek
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		// ok
		if  (e.getSource() == okButton)
		{
			OnOKButton();
			
			return;
		}
		
		// cancel
		if (e.getSource() == cancelButton)
		{
			dialogResult = CANCEL;
			
			setVisible(false);
			
			return;
		}
		
		// `each day' checkbox
		if (e.getSource() == eachDayCheckbox)
		{
			OnEachDayCheckbox();
			
			return;
		}
		
		// repetition check box
		if (e.getSource() == repetitionsCheckBox)
		{
			OnRepetitionCheckbox();
			
			return;
		}
		
		// times per day preview button
		if (e.getSource() == timesPerDayPreviewButton)
		{
			OnPreviewButton();
			
			return;
		}
	}
	
	/**
	 * A handler for the each day checkbox. It enables/disables all the other day checboxes.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnEachDayCheckbox()
	{
		boolean fSelect = !eachDayCheckbox.isSelected();
		
		mondayCheckbox.setEnabled(fSelect);
		tuesdayCheckbox.setEnabled(fSelect);
		wednesdayCheckbox.setEnabled(fSelect);
		thursdayCheckbox.setEnabled(fSelect);
		fridayCheckbox.setEnabled(fSelect);
		saturdayCheckbox.setEnabled(fSelect);
		sundayCheckbox.setEnabled(fSelect);
	}
	
	/**
	 * A handler for the repetiotion checkbox. It enables/disables other repetition options.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnRepetitionCheckbox()
	{
		boolean fSelect = repetitionsCheckBox.isSelected();
		
		everyXTimeSpanTextField.setEnabled(fSelect);
		timesPerDayPreviewButton.setEnabled(fSelect);
		repetitionsLowerLimit.setEnabled(fSelect);
		repetitionsUpperLimit.setEnabled(fSelect);
	}
	
	/**
	 * A handler for the preview button. It shows a dialog with the times of departures per day.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnPreviewButton()
	{
		// validating
		ArrayList errorsList = new ArrayList(5);
		
		if (!ValidateSettings(errorsList, false))
		{
			TravelSchedulerWorldEditor.ShowErrorsList("There are some errors in the entered settings:", errorsList);
							
			return;
		}
		
		// showing the preview
		try
		{
			int everyXMinutes = TravelTime.ParseDuration(everyXTimeSpanTextField.getText());
								
			int startTime = timetableEditor.GetDepartureTime(0);
		
			int lowerLimit = TravelTime.ParseTime(repetitionsLowerLimit.getText());
			int upperLimit = TravelTime.ParseTime(repetitionsUpperLimit.getText());
								
			TimesPerDayPreviewDialog previewDialog = new TimesPerDayPreviewDialog(startTime, everyXMinutes, lowerLimit, upperLimit);
		
			previewDialog.setVisible(true);
		}
		
		catch (ParseException xcp)
		{
			// there cannpt be any error as the data is validated
			// added to satisfy the java exceptions handling
		}
	}
	
	/**
	 * A handler for the OK button. It adds a new connetions.
	 * @author Piotrek
	 * @see #actionPerformed(ActionEvent)
	 */
	private void OnOKButton()
	{
		// validation
		ArrayList errorsList = new ArrayList();
		
		if (!ValidateSettings(errorsList, true))
		{
			TravelSchedulerWorldEditor.ShowErrorsList("There are some problems with the data you have enetered:", errorsList);
			
			return;
		}
		
		// adding connections
		
		int startTime = timetableEditor.GetArrivalTime(0);
		
		ArrayList startTimes = null;
		
		if (repetitionsCheckBox.isSelected())
		{
			// showing the preview
			try
			{
				int everyXMinutes = TravelTime.ParseDuration(everyXTimeSpanTextField.getText());
									
				int lowerLimit = TravelTime.ParseTime(repetitionsLowerLimit.getText());
				int upperLimit = TravelTime.ParseTime(repetitionsUpperLimit.getText());
				
				startTimes = TravelTime.GetTimesPerDay(startTime, everyXMinutes, lowerLimit, upperLimit);
			}
			
			catch (ParseException xcp)
			{
				// there cannpt be any error as the data is validated
				// added to satisfy the java exceptions handling
			}
		}
		else
		{
			startTimes = new ArrayList();
			
			startTimes.add(new Integer(startTime));
		}
		
		// build a list of shifts
		
		ArrayList shiftsList = GetShiftsList();
		
		// building a day of week mask
		char dayOfWeekMask = GetDayOfWeekMask();
				
		// adding a connection for each start time
		
		boolean fFirstConnection = true;
		
		Iterator iterator = startTimes.iterator();
		
		while (iterator.hasNext())
		{
			// creating a new connection
			int start = ((Integer)iterator.next()).intValue();
			
			ArrayList stations = (ArrayList)this.stations.clone();
			ArrayList arrivals = new ArrayList(stationsCount);
			ArrayList departures = new ArrayList(stationsCount);
			
			// the first station needs to be handled separately
			arrivals.add(new Integer(start));
			
			int departure = ((Integer)shiftsList.get(0)).intValue();
				
			departures.add(new Integer(start + departure));
			
			// the rest of stations
			
			for (int i = 1; i < shiftsList.size(); i += 2)
			{
				int arrival = ((Integer)shiftsList.get(i)).intValue();
				
				arrivals.add(new Integer(start + arrival));
				
				departure = ((Integer)shiftsList.get(i + 1)).intValue();
				
				departures.add(new Integer(start + departure));
			}
			
			// updating the name (the first of them has already a name assigned.)
			if (!fFirstConnection)
				connectionName = World.GetInstance().GetNextConnectionName(meanOfTransportation);
			
			// creating a new instance of connection
			Connection connection = new Connection(connectionName, meanOfTransportation, stations, arrivals, departures, dayOfWeekMask);
			
			// and finally adding it to the world
			World.GetInstance().AddConnection(connection);
			
			// this is used only when editing and then only one connection might be added
			// as the per day repetition options and return connection option are disabled
			addedConnection = connection;
			
			fFirstConnection = false;
		}
		
		// return connection
		if (returnConnectionsCheckBox.isSelected())
		{
			shiftsList = GetInvertedShiftsList();
			
			iterator = startTimes.iterator();
			
			while (iterator.hasNext())
			{
				// creating a new connection
				int start = ((Integer)iterator.next()).intValue();
				
				ArrayList stations = GetInvertedStationsList();
				ArrayList arrivals = new ArrayList(stationsCount);
				ArrayList departures = new ArrayList(stationsCount);
				
				// the first station needs to be handled separately
				arrivals.add(new Integer(start));
				
				int departure = ((Integer)shiftsList.get(0)).intValue();
					
				departures.add(new Integer(start + departure));
				
				// the rest of stations
				
				for (int i = 1; i < shiftsList.size(); i += 2)
				{
					int arrival = ((Integer)shiftsList.get(i)).intValue();
					
					arrivals.add(new Integer(start + arrival));
					
					departure = ((Integer)shiftsList.get(i + 1)).intValue();
					
					departures.add(new Integer(start + departure));
				}
				
				// updating the name of the connection
				connectionName = World.GetInstance().GetNextConnectionName(meanOfTransportation);
				
				// creating a new instance of connection
				Connection connection = new Connection(connectionName, meanOfTransportation, stations, arrivals, departures, dayOfWeekMask);
				
				// and finally adding it to the world
				synchronized (World.GetInstance())
				{
					World.GetInstance().AddConnection(connection);
				}
			}
		}
		
		dialogResult = OK;
		
		setVisible(false);
	}
	
	/**
	 * A connection added to the world if the OK button has been clicked. Null if cancel has been chosen.
	 * @author Piotrek
	 */
	private Connection addedConnection = null;
	
	/**
	 * Gets the connection added to the world if the OK button has been clicked. Null if cancel has been chosen.
	 * @author Piotrek
	 * @return A connection added to the world or null if there was none added.
	 */
	public Connection GetAddedConnection()
	{
		return addedConnection;
	}
	
	/**
	 * Gets a mask of day of week set according to the settings on the dialog.
	 * @author Piotrek
	 * @see Connection#MONDAY
	 * @see Connection#TUESDAY
	 * @see Connection#WEDNESDAY
	 * @see Connection#THURSDAY
	 * @see Connection#FRIDAY
	 * @see Connection#SATURDAY
	 * @see Connection#SUNDAY
	 * @return A mask of day of week.
	 */
	private char GetDayOfWeekMask()
	{
		char dayOfWeekMask = 0;
		
		if (eachDayCheckbox.isSelected())
			dayOfWeekMask = Connection.MONDAY | Connection.TUESDAY | Connection.WEDNESDAY | Connection.THURSDAY | Connection.FRIDAY | Connection.SATURDAY | Connection.SUNDAY;
		else
		{
			if (mondayCheckbox.isSelected())
				dayOfWeekMask |= Connection.MONDAY;
			
			if (tuesdayCheckbox.isSelected())
				dayOfWeekMask |= Connection.TUESDAY;
			
			if (wednesdayCheckbox.isSelected())
				dayOfWeekMask |= Connection.WEDNESDAY;
			
			if (thursdayCheckbox.isSelected())
				dayOfWeekMask |= Connection.THURSDAY;
			
			if (fridayCheckbox.isSelected())
				dayOfWeekMask |= Connection.FRIDAY;
			
			if (saturdayCheckbox.isSelected())
				dayOfWeekMask |= Connection.SATURDAY;
			
			if (sundayCheckbox.isSelected())
				dayOfWeekMask |= Connection.SUNDAY;
		}
		
		return dayOfWeekMask;
	}
	
	/**
	 * Gets an ordered list of shifts of time for a currently defined connection. 
	 * The list starts with the departure time shift for the first station.
	 * For all the others it contain shifts for both arrival and departure.
	 * A list contains Integers representing the shifts in minutes.
	 * It is assumed the settings has been validated before a call to this method.
	 * @author Piotrek
	 * @return A list of shifts in minutes.
	 */
	private ArrayList GetShiftsList()
	{
		ArrayList shifts = new ArrayList(2 * stationsCount - 1);
		
		int startTime = timetableEditor.GetArrivalTime(0);
				
		shifts.add(new Integer(timetableEditor.GetDepartureTime(0) - startTime));
		
		for (int i = 1; i < stationsCount; ++i)
		{			
			shifts.add(new Integer(timetableEditor.GetArrivalTime(i) - startTime));
			shifts.add(new Integer(timetableEditor.GetDepartureTime(i) - startTime));
		}
		
		return shifts;
	}
	
	/**
	 * Gets an ordered list of shifts of time for an inverted list of stations of 
	 * a currently defined connection. 
	 * The list starts with the departure time shift for the first station.
	 * For all the others it contain shifts for both arrival and departure.
	 * A list contains Integers representing the shifts in minutes.
	 * It is assumed the settings has been validated before a call to this method.
	 * @author Piotrek
	 * @return A list of shifts in minutes.
	 */
	private ArrayList GetInvertedShiftsList()
	{
		ArrayList shifts = new ArrayList(2 * stationsCount - 1);
		
		int startTime = timetableEditor.GetDepartureTime(stationsCount -1);
				
		shifts.add(new Integer(startTime - timetableEditor.GetArrivalTime(stationsCount - 1)));
		
		for (int i = stationsCount - 2; i >= 0; --i)
		{
			shifts.add(new Integer(startTime - timetableEditor.GetDepartureTime(i)));
			shifts.add(new Integer(startTime - timetableEditor.GetArrivalTime(i)));
		}
		
		return shifts;
	}
	
	/**
	 * Gets an inverted list of stations.
	 * @author Piotrek
	 * @return An inverted list of stations.
	 */
	private ArrayList GetInvertedStationsList()
	{
		ArrayList invertedStations = new ArrayList(stationsCount);
		
		Iterator iterator = stations.iterator();
		
		while (iterator.hasNext())
			invertedStations.add(0, iterator.next());
		
		return invertedStations;
	}
	
	/**
	 * Adds the OK and Cancel buttons.
	 * @author Piotrek
	 */
	private void AddOKCancel()
	{
		// ok
		
		okButton = new JButton("OK");
		
		okButton.setBounds(10, 635, 80, 25);
		okButton.setBackground(Color.ORANGE);
		
		okButton.addActionListener(this);
		
		add(okButton);
		
		// cancel
		
		cancelButton = new JButton("Cancel");
		
		cancelButton.setBounds(110, 635, 80, 25);
		cancelButton.setBackground(Color.ORANGE);
		
		cancelButton.addActionListener(this);
		
		add(cancelButton);
	}
	
	/**
	 * Adds the options to cuztomise the repetition of this connection.
	 * @author Piotrek
	 * @param fAddNonEditionOptions Indicates whether to show the per day repetitons options and the return connection option. They are not available in the edition mode.
	 */
	private void AddRepetitionOptions(boolean fAddNonEditionOptions)
	{
		// day of week
		
		JLabel label = new JLabel("Day of Week", SwingConstants.CENTER);
		
		label.setOpaque(true);
		label.setBackground(Color.MAGENTA);
		label.setBounds(10, 530, 320, 20);
		
		add(label);
		
		// each day checkbox
		
		eachDayCheckbox = new JCheckBox("Each day");
		
		eachDayCheckbox.setBounds(10, 560, 100, 20);
		eachDayCheckbox.addActionListener(this);
		
		add(eachDayCheckbox);
		
		// Monday
		
		mondayCheckbox = new JCheckBox("Monday");
		
		mondayCheckbox.setBounds(10, 580, 100, 20);
		mondayCheckbox.addActionListener(this);
		
		add(mondayCheckbox);
		
		// Tuesday
		
		tuesdayCheckbox = new JCheckBox("Tuesday");
		
		tuesdayCheckbox.setBounds(10, 600, 100, 20);
		tuesdayCheckbox.addActionListener(this);
		
		add(tuesdayCheckbox);
		
		// Wednesday
		
		wednesdayCheckbox = new JCheckBox("Wednesday");
		
		wednesdayCheckbox.setBounds(120, 560, 100, 20);
		wednesdayCheckbox.addActionListener(this);
		
		add(wednesdayCheckbox);
		
		// Thursday
		
		thursdayCheckbox = new JCheckBox("Thursday");
		
		thursdayCheckbox.setBounds(120, 580, 100, 20);
		thursdayCheckbox.addActionListener(this);
		
		add(thursdayCheckbox);
		
		// Friday
		
		fridayCheckbox = new JCheckBox("Friday");
		
		fridayCheckbox.setBounds(120, 600, 100, 20);
		fridayCheckbox.addActionListener(this);
		
		add(fridayCheckbox);
		
		// Saturday
		
		saturdayCheckbox = new JCheckBox("Saturday");
		
		saturdayCheckbox.setBounds(230, 560, 100, 20);
		saturdayCheckbox.addActionListener(this);
		
		add(saturdayCheckbox);
		
		// Sunday
		
		sundayCheckbox = new JCheckBox("Sunday");
		
		sundayCheckbox.setBounds(230, 580, 100, 20);
		sundayCheckbox.addActionListener(this);
		
		add(sundayCheckbox);
				
		// every x time span
		
		label = new JLabel("Per Day Repetitions", SwingConstants.CENTER);
		
		label.setOpaque(true);
		label.setBackground(Color.MAGENTA);
		label.setBounds(350, 530, 220, 20);
		label.setVisible(fAddNonEditionOptions);
		
		add(label);
		
		// `use repetitions' check box
		repetitionsCheckBox = new JCheckBox("Use repetitions");
		
		repetitionsCheckBox.setSelected(false);
		repetitionsCheckBox.setBounds(350, 560, 120, 20);
		repetitionsCheckBox.addActionListener(this);
		repetitionsCheckBox.setVisible(fAddNonEditionOptions);
		
		add(repetitionsCheckBox);
		
		// label
		label = new JLabel("Every x time span:");
		
		label.setBounds(350, 590, 150, 20);
		label.setVisible(fAddNonEditionOptions);
		
		add(label);
		
		// text field
		everyXTimeSpanTextField = new JTextField();
		
		everyXTimeSpanTextField.setBounds(350, 610, 100, 20);
		everyXTimeSpanTextField.setEnabled(false);
		everyXTimeSpanTextField.setVisible(fAddNonEditionOptions);
		
		add(everyXTimeSpanTextField);
		
		// preview button
		timesPerDayPreviewButton = new JButton("Preview");
		
		timesPerDayPreviewButton.setBackground(Color.ORANGE);
		timesPerDayPreviewButton.setBounds(350, 640, 100, 20);
		timesPerDayPreviewButton.addActionListener(this);
		timesPerDayPreviewButton.setEnabled(false);
		timesPerDayPreviewButton.setVisible(fAddNonEditionOptions);
		
		add(timesPerDayPreviewButton);
		
		// repertitions lower limit
		
		label = new JLabel("Lower limit:");
		
		label.setBounds(470, 550, 100, 20);
		label.setVisible(fAddNonEditionOptions);
		
		add(label);
		
		repetitionsLowerLimit = new JTextField("0:00");
		
		repetitionsLowerLimit.setBounds(470, 570, 100, 20);
		repetitionsLowerLimit.setEnabled(false);
		repetitionsLowerLimit.setVisible(fAddNonEditionOptions);
		
		add(repetitionsLowerLimit);
		
		// repetitions upper limit
		
		label = new JLabel("Upper limit:");
		
		label.setBounds(470, 590, 100, 20);
		label.setVisible(fAddNonEditionOptions);
		
		add(label);
		
		repetitionsUpperLimit = new JTextField("23:59");
		
		repetitionsUpperLimit.setBounds(470, 610, 100, 20);
		repetitionsUpperLimit.setEnabled(false);
		repetitionsUpperLimit.setVisible(fAddNonEditionOptions);
		
		add(repetitionsUpperLimit);
		
		// return connection
		
		label = new JLabel("Return Connections", SwingConstants.CENTER);
		
		label.setOpaque(true);
		label.setBackground(Color.MAGENTA);
		label.setBounds(590, 530, 190, 20);
		label.setVisible(fAddNonEditionOptions);
		
		add(label);
		
		// `add return connection(s)' check box
		returnConnectionsCheckBox = new JCheckBox("Add return connection(s)");
		
		returnConnectionsCheckBox.setSelected(false);
		returnConnectionsCheckBox.setBounds(590, 560, 190, 20);
		returnConnectionsCheckBox.addActionListener(this);
		returnConnectionsCheckBox.setVisible(fAddNonEditionOptions);
		
		add(returnConnectionsCheckBox);
	}
	
	/**
	 * A text field for a lower limit of repetitions.
	 * @author Piotrek
	 */
	private JTextField repetitionsLowerLimit = null;
	
	/**
	 * A text field for an upper limit of repetitions.
	 * @author Piotrek
	 */
	private JTextField repetitionsUpperLimit = null;
	
	/**
	 * A `each day' check box.
	 * @author Piotrek
	 */
	private JCheckBox eachDayCheckbox = null;
	
	/**
	 * A `Monday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox mondayCheckbox = null;
	
	/**
	 * A `Tuesday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox tuesdayCheckbox = null;
	
	/**
	 * A `Wednesday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox wednesdayCheckbox = null;
	
	/**
	 * A `Thursday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox thursdayCheckbox = null;
	
	/**
	 * A `Friday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox fridayCheckbox = null;
	
	/**
	 * A `Saturday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox saturdayCheckbox = null;
	
	/**
	 * A `Sunday' checkbox.
	 * @author Piotrek
	 */
	private JCheckBox sundayCheckbox = null;
	
	/**
	 * A `Use repetitions' check box.
	 * @author Piotrek
	 */
	private JCheckBox repetitionsCheckBox = null;
	
	/**
	 * A text field to receive the every x time span repetition per day.
	 * @author Piotrek
	 */
	private JTextField everyXTimeSpanTextField = null;
	
	/**
	 * A button used to preview the times that the `every x minutes' value defines.
	 * @author Piotrek
	 */
	private JButton timesPerDayPreviewButton = null;
	
	/**
	 * A 'return connection(s) check box.
	 * @author Piotrek
	 */
	private JCheckBox returnConnectionsCheckBox = null;
	
	/**
	 * Validates all the entered settings on this dialog.
	 * @author Piotrek
	 * @param errorsList A list of strings, the error messages will be added to.
	 * @param fValidateDayOfWeek Indicates whether to check the settings for day of week. Preview button does not neet it.
	 * @return A bool value indicating whether the entered settings are OK.
	 */
	public boolean ValidateSettings(ArrayList errorsList, boolean fValidateDayOfWeek)
	{
		if (errorsList == null)
			throw new NullPointerException("Errors list must not be null in ConnectionTimetableCreator.ValidateSettings().");
		
		boolean fOK;
		
		// timetable options
		fOK = timetableEditor.ValidateTimetable(errorsList);
		
		// the repetition options
		
		if (repetitionsCheckBox.isSelected())
		{
			if (everyXTimeSpanTextField.getText().compareTo("") == 0)
			{
				errorsList.add("The `every x time span' text field is empty! You need to provide some value.");
				
				fOK = false;
			}
					
			try
			{
				TravelTime.ParseDuration(everyXTimeSpanTextField.getText());
			}
			
			catch (ParseException xcp)
			{
				errorsList.add("The `every x time span' text field contains an invalid value: " + xcp.getMessage());
				
				fOK = false;
			}
								
			int lowerLimit = -1;
			int upperLimit = -1;
			
			try
			{
				lowerLimit = TravelTime.ParseTime(repetitionsLowerLimit.getText());
			}
			
			catch (Exception xcp)
			{
				errorsList.add("The `lower limit' text field contains an invalid time value: " + xcp.getMessage());
				
				fOK = false;
			}
		
			try
			{
				upperLimit = TravelTime.ParseTime(repetitionsUpperLimit.getText());
			}
			
			catch (Exception xcp)
			{
				errorsList.add("The `upper limit' text field contains an invalid time value: " + xcp.getMessage());
				
				fOK = false;
			}		
			
			if (lowerLimit != -1 && upperLimit != -1 && lowerLimit >= upperLimit)
			{
				errorsList.add("The lower limit is not before the upper limit.");
				
				fOK = false;
			}
		}
		
		// day of week 
		
		if (fValidateDayOfWeek)
		{
			if (!eachDayCheckbox.isSelected())
			{
				if (!mondayCheckbox.isSelected() && !tuesdayCheckbox.isSelected() 
					&& !wednesdayCheckbox.isSelected() && !thursdayCheckbox.isSelected() 
					&& !fridayCheckbox.isSelected() && !saturdayCheckbox.isSelected() 
					&& !sundayCheckbox.isSelected())
				{
					errorsList.add("You must specify a day of week. Choose each day option or select only the desired ones.");
					
					fOK = false;
				}
			}
		}
		
		return fOK;
	}
}