/*
 * Created on 2005-05-20
 */
package gui.dialogs;

import java.awt.Color;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import scheduler.TravelConstraints;
import scheduler.TravelPartsGenerator;

/**
 * A dialog used to query the user for the travel details (time limits, changes count and so on).
 * @author Piotrek
 */
public final class TravelDetailsQueryDialog extends JDialog implements ActionListener
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public TravelDetailsQueryDialog()
	{
		setTitle("Enter Travel Details");
		
		setModal(true);
		setVisible(false);
		setResizable(false);
		
		setSize(810, 530);
		
		setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		
		// panel	
		panel = getContentPane();
		
		panel.setLayout(null);
		panel.setBackground(Color.YELLOW);
		
		setContentPane(panel);
		
		// ok and cancel buttons
		AddOKCancel();
		
		// times panel
		AddTimesPanel();
		
		// allowed means of transportation
		AddMeansOfTransportationPanel();
		
		// max changes count
		AddMaxChangesPanel();
		
		// min & max awaiting times
		AddMinMaxChangeTimesPanel();
		
		// times before flight & cruise
		AddTimesBeforeFlightCruise();
	}
	
	/**
	 * An OK button.
	 * @author Piotrek
	 */
	private JButton okButton = null;
	
	/**
	 * A cancel button.
	 * @author Piotrek
	 */
	private JButton cancelButton = null;
	
	/**
	 * The OK dialog result.
	 * @author Piotrek
	 */
	public static final int OK = 1;
	
	/**
	 * The cancel dialog result.
	 * @author Piotrek
	 */
	public static final int CANCEL = 2;
	
	/**
	 * The dialog result.
	 * @author Piotrek
	 */
	private int dialogResult = CANCEL;
	
	/**
	 * Gets the dialog result.
	 * @author Piotrek
	 * @return The dialog result.
	 */
	public int GetDialogResult()
	{
		return dialogResult;
	}
	
	/**
	 * A main panel of this dialog.
	 * @author Piotrek
	 */
	private Container panel = null;
	
	/**
	 * Adds the OK and Cancel buttons.
	 * @author Piotrek
	 */
	private void AddOKCancel()
	{
		// ok
		
		okButton = new JButton("OK");
		
		okButton.setBounds(10, 460, 80, 25);
		okButton.setBackground(Color.GREEN);
		
		okButton.addActionListener(this);
		
		add(okButton);
		
		// cancel
		
		cancelButton = new JButton("Cancel");
		
		cancelButton.setBounds(110, 460, 80, 25);
		cancelButton.setBackground(Color.GREEN);
		
		cancelButton.addActionListener(this);
		
		add(cancelButton);
	}
	
	/**
	 * @author Piotrek
	 * @see ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		// OK buttom
		if (e.getSource() == okButton)
		{
			OnOKButton();
		}
		
		// Cancel button
		if (e.getSource() == cancelButton)
		{
			dialogResult = CANCEL;
			
			setVisible(false);
						
			return;
		}
		
		// time limit mode combo box
		if (e.getSource() == timeLimitsMode)
		{
			if (timeLimitsMode.getSelectedIndex() == 0)
			{
				departureTimeTextField.setEnabled(true);
				arrivalTimeTextField.setEnabled(false);
			}
			else
			{
				departureTimeTextField.setEnabled(false);
				arrivalTimeTextField.setEnabled(true);
			}
		}
	}
	
	/**
	 * Launched when the user clicks the OK button. It validates the user input.
	 * @author Piotrek
	 */
	private void OnOKButton()
	{
		boolean fOK = true;
		
		String errors = "There are some problems:\n";
		
		if (timeLimitsMode.getSelectedIndex() == 0)
		{
			try
			{
				Date departureDate = DateFormat.getInstance().parse(departureTimeTextField.getText());
				
				if (departureDate.compareTo(DateFormat.getInstance().getCalendar().getTime()) <= 0)
				{
					fOK = false;
					
					errors += "The departure date must be later then now.";
				}
			}
			
			catch (ParseException xcp)
			{
				fOK = false;
				
				errors += " - The departure date is empty or invalid.\n";
			}
		}
		else
		{			
			try
			{
				Date arrivalDate = DateFormat.getInstance().parse(arrivalTimeTextField.getText());
				
				if (arrivalDate.compareTo(DateFormat.getInstance().getCalendar().getTime()) <= 0)
				{
					fOK = false;
					
					errors += "The arrival date must be later then now.";
				}
			}
			
			catch (ParseException xcp)
			{
				fOK = false;
				
				errors += " - The arrival date is empty or invalid.\n";
			}
		}
		
		if (!taxiCheckBox.isSelected() && !trainCheckBox.isSelected() && !airplaneCheckBox.isSelected() && !busCheckBox.isSelected() && !shipCheckBox.isSelected())
		{
			errors += " - You must allow at least one mean of transportation except the walking.\n";
			
			fOK = false;
		}
		
		try
		{
			int maxChangeCount = Integer.parseInt(maxChangesTextField.getText());
			
			if (maxChangeCount < 0)
				throw new Exception("Too small.");
			
			int maxPartsCount = TravelPartsGenerator.GetMaxPartsCount();
			
			if (maxChangeCount > maxPartsCount)
			{
				fOK = false;
				
				errors += " - The max changes count must not be greater then " + maxPartsCount;
			}
		}
		
		catch (Exception xcp)
		{
			errors += " - The max changes count must be a non-negative integer.\n";
			
			fOK = false;
		}
		
		// min and max change times
		int minChangeTime = -1981;
		int maxChangeTime = -1981;
		
		try
		{
			minChangeTime = Integer.parseInt(minChangeTimeTextField.getText());
		}
		
		catch (Exception xcp)
		{
			fOK = false;
			
			errors += " - The min change time is not a valid integer.\n";
		}
		
		try
		{
			maxChangeTime = Integer.parseInt(maxChangeTimeTextField.getText());
		}
		
		catch (Exception xcp)
		{
			fOK = false;
			
			errors += " - The max change time is not a valid integer.\n";
		}
		
		if (minChangeTime != -1981 && maxChangeTime != -1981)
		{
			if (minChangeTime < 0)
			{
				fOK = false;
				
				errors += " - The min change time must be a positive integer.\n";
			}
			
			if (maxChangeTime < 0)
			{
				fOK = false;
				
				errors += " - The max change time must be a positive integer.\n";
			}
			
			if (minChangeTime >= 0 && maxChangeTime >= 0)
			{
				if (minChangeTime > maxChangeTime)
				{
					fOK = false;
					
					errors += " - The min change time must not be greater then the max change time.\n";
				}
			}
		}
		
		// times before flight or cruise
		int timeBeforeFlight = -1981;
		int timeBeforeCruise = -1981;
		
		try
		{
			timeBeforeFlight = Integer.parseInt(timeBeforeFlightTextField.getText());
		}
		
		catch (Exception xcp)
		{
			fOK = false;
			
			errors += " - The time before a flight is not a valid integer.\n";
		}
		
		try
		{
			timeBeforeCruise = Integer.parseInt(timeBeforeCruiseTextField.getText());
		}
		
		catch (Exception xcp)
		{
			fOK = false;
			
			errors += " - The time before a cruise is not a valid integer.\n";
		}
		
		if (timeBeforeFlight != -1981)
		{
			if (timeBeforeFlight < 0)
			{
				fOK = false;
				
				errors += " - The time before a flight must be a positive integer.\n";
			}
		}
		
		if (timeBeforeCruise != -1981)
		{
			if (timeBeforeCruise < 0)
			{
				fOK = false;
				
				errors += " - The time before a cruise must be a positive integer.\n";
			}
		}
		
		if (fOK)
		{
			fValidated = true;
			
			dialogResult = OK;
		
			setVisible(false);
		}
		else
		{
			JOptionPane.showMessageDialog(null, errors, "Error", JOptionPane.ERROR_MESSAGE);
		}
	}
		
	/**
	 * Adds a panel with the times options.
	 * @author Piotrek
	 */
	private void AddTimesPanel()
	{
		JPanel timePanel = new JPanel(null);
		
		timePanel.setBounds(10, 10, 385, 295);
		timePanel.setBackground(Color.BLUE);
		
		add(timePanel);
		
		// title label
		JLabel titleLabel = new JLabel("Time Options", SwingConstants.CENTER);
		
		titleLabel.setBounds(10, 10, 365, 25);
		titleLabel.setBackground(Color.ORANGE);
		titleLabel.setForeground(Color.BLACK);
		titleLabel.setOpaque(true);
		
		timePanel.add(titleLabel);
		
		// time limits mode label
		JLabel timeLimitsLabel = new JLabel("Time limits mode:");
		
		timeLimitsLabel.setForeground(Color.WHITE);
		timeLimitsLabel.setBounds(10, 60, 200, 25);
		
		timePanel.add(timeLimitsLabel);
		
		// time limits mode combo box
		timeLimitsMode = new JComboBox();
		
		timeLimitsMode.addItem("Departure Time");
		timeLimitsMode.addItem("Arrival Time");
		
		timeLimitsMode.setSelectedIndex(0);
		
		timeLimitsMode.setBounds(10, 100, 365, 25);
		
		timeLimitsMode.addActionListener(this);
		
		timePanel.add(timeLimitsMode);
		
		// departure time label
		JLabel departureTimeLabel = new JLabel("Departure time:");
		
		departureTimeLabel.setForeground(Color.WHITE);
		departureTimeLabel.setBounds(10, 140, 200, 35);
		
		timePanel.add(departureTimeLabel);
				
		// departure time text field
		departureTimeTextField = new JTextField();
		
		departureTimeTextField.setBounds(10, 180, 365, 25);
		
		timePanel.add(departureTimeTextField);
		
		Calendar calendar = Calendar.getInstance();
		
		if (calendar.getActualMaximum(Calendar.DAY_OF_MONTH) - calendar.get(Calendar.DAY_OF_MONTH) < 1)
		{
			if (calendar.get(Calendar.MONTH) == Calendar.DECEMBER)
				calendar.roll(Calendar.YEAR, 1);
			
			calendar.roll(Calendar.MONTH, 1);
		}
		
		calendar.roll(Calendar.DAY_OF_MONTH, 1);
		
		departureTimeTextField.setText(DateFormat.getInstance().format(calendar.getTime()));
				
		// arrival time label
		JLabel arrivalTimeLabel = new JLabel("Arrival time:");
		
		arrivalTimeLabel.setForeground(Color.WHITE);
		arrivalTimeLabel.setBounds(10, 220, 200, 35);
		
		timePanel.add(arrivalTimeLabel);
		
		// arrival time text field
		arrivalTimeTextField = new JTextField();
		
		arrivalTimeTextField.setBounds(10, 260, 365, 25);
		arrivalTimeTextField.setEnabled(false);
		
		timePanel.add(arrivalTimeTextField);
		
		calendar = Calendar.getInstance();
		
		if (calendar.getActualMaximum(Calendar.DAY_OF_MONTH) - calendar.get(Calendar.DAY_OF_MONTH) < 14)
		{
			if (calendar.get(Calendar.MONTH) == Calendar.DECEMBER)
				calendar.roll(Calendar.YEAR, 1);
			
			calendar.roll(Calendar.MONTH, 1);
		}
		
		calendar.roll(Calendar.DAY_OF_MONTH, 14);
		
		arrivalTimeTextField.setText(DateFormat.getInstance().format(calendar.getTime()));
	}
	
	/**
	 * A combo box used to define the time limits mode.
	 * @author Piotrek
	 */
	private JComboBox timeLimitsMode = null;
	
	/**
	 * A departure time text field.
	 * @author Piotrek
	 */
	private JTextField departureTimeTextField = null;
	
	/**
	 * An arrival time text field.
	 * @author Piotrek
	 */
	private JTextField arrivalTimeTextField = null;
	
	/**
	 * Adds a panel with the means of transportation to be chosen as allowed.
	 * @author Piotrek
	 */
	private void AddMeansOfTransportationPanel()
	{
		JPanel motPanel = new JPanel(null);
		
		motPanel.setBackground(Color.BLUE);
		motPanel.setBounds(405, 10, 385, 135);
		
		add(motPanel);
		
		// title label
		JLabel titleLabel = new JLabel("Allowed Means Of Transportation", SwingConstants.CENTER);
		
		titleLabel.setBounds(10, 10, 365, 25);
		titleLabel.setBackground(Color.ORANGE);
		titleLabel.setForeground(Color.BLACK);
		titleLabel.setOpaque(true);
		
		motPanel.add(titleLabel);
		
		// check boxes
		
		// walk
		JCheckBox walkCheckBox = new JCheckBox("Walk");
		
		walkCheckBox.setForeground(Color.WHITE);
		walkCheckBox.setBounds(10, 60, 100, 25);
		walkCheckBox.setOpaque(false);
		
		// walk is always allowed
		walkCheckBox.setSelected(true);
		walkCheckBox.setEnabled(false);
		
		motPanel.add(walkCheckBox);
		
		// airplane
		
		airplaneCheckBox = new JCheckBox("Airplane");
		
		airplaneCheckBox.setForeground(Color.WHITE);
		airplaneCheckBox.setBounds(10, 100, 100, 25);
		airplaneCheckBox.setOpaque(false);
		
		airplaneCheckBox.setSelected(true);
				
		motPanel.add(airplaneCheckBox);
		
		// taxi
		
		taxiCheckBox = new JCheckBox("Taxi");
		
		taxiCheckBox.setForeground(Color.WHITE);
		taxiCheckBox.setBounds(130, 60, 100, 25);
		taxiCheckBox.setOpaque(false);
		
		taxiCheckBox.setSelected(true);
				
		motPanel.add(taxiCheckBox);
		
		// ship
		
		shipCheckBox = new JCheckBox("Ship");
		
		shipCheckBox.setForeground(Color.WHITE);
		shipCheckBox.setBounds(130, 100, 100, 25);
		shipCheckBox.setOpaque(false);
		
		shipCheckBox.setSelected(true);
				
		motPanel.add(shipCheckBox);
		
		// bus
		
		busCheckBox = new JCheckBox("Bus");
		
		busCheckBox.setForeground(Color.WHITE);
		busCheckBox.setBounds(240, 60, 100, 25);
		busCheckBox.setOpaque(false);
		
		busCheckBox.setSelected(true);
				
		motPanel.add(busCheckBox);
		
		// train
		
		trainCheckBox = new JCheckBox("Train");
		
		trainCheckBox.setForeground(Color.WHITE);
		trainCheckBox.setBounds(240, 100, 100, 25);
		trainCheckBox.setOpaque(false);
		
		trainCheckBox.setSelected(true);
				
		motPanel.add(trainCheckBox);
	}
	
	/**
	 * A check box used to choose the airplane as an allowed mean of transportation.
	 * @author Piotrek
	 */
	private JCheckBox airplaneCheckBox = null;
	
	/**
	 * A check box used to choose the taxi as an allowed mean of transportation.
	 * @author Piotrek
	 */
	private JCheckBox taxiCheckBox = null;
	
	/**
	 * A check box used to choose the ship as an allowed mean of transportation.
	 * @author Piotrek
	 */
	private JCheckBox shipCheckBox = null;
	
	/**
	 * A check box used to choose the bus as an allowed mean of transportation.
	 * @author Piotrek
	 */
	private JCheckBox busCheckBox = null;
	
	/**
	 * A check box used to choose the train as an allowed mean of transportation.
	 * @author Piotrek
	 */
	private JCheckBox trainCheckBox = null;
	
	/**
	 * Adds a panel with the option of max changes count.
	 * @author Piotrek
	 */
	private void AddMaxChangesPanel()
	{
		JPanel maxChangesPanel = new JPanel(null);
		
		maxChangesPanel.setBackground(Color.BLUE);
		maxChangesPanel.setBounds(405, 155, 385, 135);
		
		add(maxChangesPanel);
		
		// title label
		JLabel titleLabel = new JLabel("Max Changes Count", SwingConstants.CENTER);
		
		titleLabel.setBounds(10, 10, 365, 25);
		titleLabel.setBackground(Color.ORANGE);
		titleLabel.setForeground(Color.BLACK);
		titleLabel.setOpaque(true);
		
		maxChangesPanel.add(titleLabel);
		
		// max changes label
		
		JLabel maxChangesLabel = new JLabel("Max changes count:");
		
		maxChangesLabel.setForeground(Color.WHITE);
		maxChangesLabel.setBounds(10, 60, 365, 25);
		
		maxChangesPanel.add(maxChangesLabel);
		
		// max changes text field
		maxChangesTextField = new JTextField();
		
		maxChangesTextField.setBounds(10, 100, 365, 25);
		
		maxChangesTextField.setText("10");
		
		maxChangesPanel.add(maxChangesTextField);
	}
	
	/**
	 * A text field containing the max changes count value.
	 * @author Piotrek
	 */
	private JTextField maxChangesTextField = null;
	
	/**
	 * Gets the travel constraints entered in the dialog.
	 * @author Piotrek
	 * @return The constrains or the null value if the dialog has not been launched and accepted by the user.
	 */
	public TravelConstraints GetTravelConstraints()
	{
		if (!fValidated)
			return null;
		
		TravelConstraints travelConstraints = new TravelConstraints();
		
		if (timeLimitsMode.getSelectedIndex() == 0)
		{
			travelConstraints.fIsLimitDeparture = true;
			
			try
			{
				travelConstraints.dateLimit = DateFormat.getInstance().parse(departureTimeTextField.getText());
			}
			
			catch (ParseException xcp)
			{
				// impossible since the data has been validated
			}
		}
		else
		{
			travelConstraints.fIsLimitDeparture = false;
			
			try
			{
				travelConstraints.dateLimit = DateFormat.getInstance().parse(arrivalTimeTextField.getText());
			}
			
			catch (ParseException xcp)
			{
				// impossible since the data has been validated
			}
		}
		
		travelConstraints.maxChangesCount = Integer.parseInt(maxChangesTextField.getText());
		
		travelConstraints.isAirplaneAllowed = airplaneCheckBox.isSelected();
		travelConstraints.isBusAllowed = busCheckBox.isSelected();
		travelConstraints.isShipAllowed = shipCheckBox.isSelected();
		travelConstraints.isTaxiAllowed = taxiCheckBox.isSelected();
		travelConstraints.isTrainAllowed = trainCheckBox.isSelected();
		
		// min change time
		try
		{
			travelConstraints.minChangeTime = Integer.parseInt(minChangeTimeTextField.getText());
		}
		
		catch (Exception xcp)
		{
			// impossible since the data has been validated
		}
		
		// max change time
		try
		{
			travelConstraints.maxChangeTime = Integer.parseInt(maxChangeTimeTextField.getText());
		}
		
		catch (Exception xcp)
		{
			// impossible since the data has been validated
		}
		
		// time before flight
		try
		{
			travelConstraints.timeBeforeFlight = Integer.parseInt(timeBeforeFlightTextField.getText());
		}
		
		catch (Exception xcp)
		{
			// impossible since the data has been validated
		}
		
		// time before cruise
		try
		{
			travelConstraints.timeBeforeCruise = Integer.parseInt(timeBeforeCruiseTextField.getText());
		}
		
		catch (Exception xcp)
		{
			// impossible since the data has been validated
		}
		
		return travelConstraints;
	}
	
	/**
	 * Indicates whether the user input has been validated.
	 * @author Piotrek
	 */
	private boolean fValidated = false;
	
	/**
	 * Adds a panel with the min and max values for the change times.
	 * @author Piotrek
	 */
	private void AddMinMaxChangeTimesPanel()
	{
		JPanel changeTimesPanel = new JPanel(null);
		
		changeTimesPanel.setBackground(Color.BLUE);
		changeTimesPanel.setBounds(405, 300, 385, 135);
		
		add(changeTimesPanel);
		
		// title label
		JLabel titleLabel = new JLabel("Min & Max Change Times", SwingConstants.CENTER);
		
		titleLabel.setBounds(10, 10, 365, 25);
		titleLabel.setBackground(Color.ORANGE);
		titleLabel.setForeground(Color.BLACK);
		titleLabel.setOpaque(true);
		
		changeTimesPanel.add(titleLabel);
		
		// min label
		
		JLabel minChangeTimeLabel = new JLabel("Min change time (in minutes):");
		
		minChangeTimeLabel.setForeground(Color.WHITE);
		minChangeTimeLabel.setBounds(10, 60, 175, 25);
		
		changeTimesPanel.add(minChangeTimeLabel);
		
		// min change time text field
		minChangeTimeTextField = new JTextField();
		
		minChangeTimeTextField.setBounds(10, 100, 175, 25);
		
		minChangeTimeTextField.setText("5");
		
		changeTimesPanel.add(minChangeTimeTextField);
		
		// max label
		
		JLabel maxChangeTimeLabel = new JLabel("Max change time (in minutes):");
		
		maxChangeTimeLabel.setForeground(Color.WHITE);
		maxChangeTimeLabel.setBounds(200, 60, 175, 25);
		
		changeTimesPanel.add(maxChangeTimeLabel);
		
		// min awaiting time text field
		maxChangeTimeTextField = new JTextField();
		
		maxChangeTimeTextField.setBounds(200, 100, 175, 25);
		
		maxChangeTimeTextField.setText("120");
		
		changeTimesPanel.add(maxChangeTimeTextField);
	}
	
	/**
	 * Adds a panel with the times one need to be before the flight & cruise.
	 * @author Piotrek
	 */
	private void AddTimesBeforeFlightCruise()
	{
		JPanel timesPanel = new JPanel(null);
		
		timesPanel.setBackground(Color.BLUE);
		timesPanel.setBounds(10, 315, 385, 135);
		
		add(timesPanel);
		
		// title label
		JLabel titleLabel = new JLabel("Times Before Departures", SwingConstants.CENTER);
		
		titleLabel.setBounds(10, 10, 365, 25);
		titleLabel.setBackground(Color.ORANGE);
		titleLabel.setForeground(Color.BLACK);
		titleLabel.setOpaque(true);
		
		timesPanel.add(titleLabel);
		
		// flight label
		
		JLabel timeBeforeFlightLabel = new JLabel("Before flight (in minutes):");
		
		timeBeforeFlightLabel.setForeground(Color.WHITE);
		timeBeforeFlightLabel.setBounds(10, 60, 175, 25);
		
		timesPanel.add(timeBeforeFlightLabel);
		
		TravelConstraints constraints = new TravelConstraints();
		
		// time before flight text field
		timeBeforeFlightTextField = new JTextField();
		
		timeBeforeFlightTextField.setBounds(10, 100, 175, 25);
		
		timeBeforeFlightTextField.setText(Integer.toString(constraints.timeBeforeFlight));
		
		timesPanel.add(timeBeforeFlightTextField);
		
		// max label
		
		JLabel timeBeforeCruiseLabel = new JLabel("Before cruise (in minutes):");
		
		timeBeforeCruiseLabel.setForeground(Color.WHITE);
		timeBeforeCruiseLabel.setBounds(200, 60, 175, 25);
		
		timesPanel.add(timeBeforeCruiseLabel);
		
		// min awaiting time text field
		timeBeforeCruiseTextField = new JTextField();
		
		timeBeforeCruiseTextField.setBounds(200, 100, 175, 25);
		
		timeBeforeCruiseTextField.setText(Integer.toString(constraints.timeBeforeCruise));
		
		timesPanel.add(timeBeforeCruiseTextField);
	}
	
	/**
	 * A text field with the min time of awaiting when the change takes place.
	 * @author Piotrek
	 */
	private JTextField minChangeTimeTextField = null;
	
	/**
	 * A text field with the max time of awaiting when the change takes place.
	 * @author Piotrek
	 */
	private JTextField maxChangeTimeTextField = null;
	
	/**
	 * A text field with the time one need to be at the airport before the flight.
	 * @author Piotrek
	 */
	private JTextField timeBeforeFlightTextField = null;
	
	/**
	 * A text field with the time one need to be at the harbour before the cruise.
	 * @author Piotrek
	 */
	private JTextField timeBeforeCruiseTextField = null;
}