/*
 * Created on 2005-08-18
 */
package scheduler;

/**
 * A class containing information on what mechanisms should be turned off while tests.
 * It also contains test counters.
 * @author Piotrek
 */
public final class TestOptions
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	private TestOptions()
	{
		// this is only to prevent creation of instances of this class
	}
	
	/**
	 * A number of queries for stations in range.
	 * @author Piotrek
	 */
	private static long stationsInRangeCounter;
	
	/**
	 * A number of queries for connections for stations.
	 * @author Piotrek
	 */
	private static long connectionsForStationsCounter;
	
	/**
	 * Means of transportation choosing rules engine to be used.
	 * @author Piotrek
	 */
	private static int meansOfTransporationChoosingEngine;
	
	/**
	 * Changes dropping rules engine to be used.
	 * @author Piotrek
	 */
	private static int changesDroppingEngine;
	
	/**
	 * Means of transportation choosing rules engine using JAVA hard coded rules.
	 * @author Piotrek
	 */
	public static final int MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JAVA = 1;
	
	/**
	 * Means of transportation choosing rules engine using JESS.
	 * @author Piotrek
	 */
	public static final int MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS = 2;
	
	/**
	 * Changes dropping rules engine using JAVA hard coded rules.
	 * @author Piotrek
	 */
	public static final int CHANGES_DROPPING_RULES_ENGINE_JAVA = 1;
	
	/**
	 * Changes dropping rules engine using JESS.
	 * @author Piotrek
	 */
	public static final int CHANGES_DROPPING_RULES_ENGINE_JESS = 2;
	
	/**
	 * Resets all counters and other options to default values (full functionality).
	 * @author Piotrek
	 */
	public static void Reset()
	{
		stationsInRangeCounter = 0;
		
		connectionsForStationsCounter = 0;
		
		fUseRulesDroppingChanges = true;
		
		fLimitAlternativeChangesConnections = true;
		
		fIgnoreContinentIslandInformation = false;
		
		meansOfTransporationChoosingEngine = MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS;
		
		changesDroppingEngine = CHANGES_DROPPING_RULES_ENGINE_JESS;
	}
	
	/**
	 * Resets all counters.
	 * @author Piotrek
	 */
	public static void ResetCounters()
	{
		stationsInRangeCounter = 0;
		
		connectionsForStationsCounter = 0;
	}
	
	/**
	 * Gets an engine to be used to choose means of transportation.
	 * @author Piotrek
	 * @return An engine to be used to choose means of transportation.
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JAVA
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS
	 */
	public static int GetMeansOfTransportationChoosingEngine()
	{
		return meansOfTransporationChoosingEngine;
	}
	
	/**
	 * Sets the engine to be used to choose means of transportation.
	 * @author Piotrek
	 * @param engine An engine to be used.
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JAVA
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS
	 */
	public static void SetMeansOfTransportationChoosingEngine(int engine)
	{
		if (engine != MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JAVA && engine != MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS)
			throw new IllegalArgumentException("Wrong engine in TestOptions.SetMeansOfTransportationChoosingEngine().");
		
		meansOfTransporationChoosingEngine = engine;
	}
	
	/**
	 * Gets an engine to be used to drop changes.
	 * @author Piotrek
	 * @return An engine to be used to drop changes.
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JAVA
	 * @see #MEANS_OF_TRANSPORTATION_CHOOSING_RULES_ENGINE_JESS
	 */
	public static int GetChangesDroppingEngine()
	{
		return changesDroppingEngine;
	}
	
	/**
	 * Sets the engine to be used to drop changes.
	 * @author Piotrek
	 * @param engine An engine to be used.
	 * @see #CHANGES_DROPPING_RULES_ENGINE_JAVA
	 * @see #CHANGES_DROPPING_RULES_ENGINE_JESS
	 */
	public static void SetChangesDroppingEngine(int engine)
	{
		if (engine != CHANGES_DROPPING_RULES_ENGINE_JAVA && engine != CHANGES_DROPPING_RULES_ENGINE_JESS)
			throw new IllegalArgumentException("Wrong engine in TestOptions.SetChangesDroppingEngine().");
		
		changesDroppingEngine = engine;
	}
	
	/**
	 * Gets a number of queries for stations in range.
	 * @author Piotrek
	 * @return A number of queries for stations in range.
	 */
	public static long GetStationsInRangeQueriesCount()
	{
		return stationsInRangeCounter;
	}
	
	/**
	 * Gets a number of queries for connections for stations.
	 * @author Piotrek
	 * @return A number of queries for connections for stations.
	 */
	public static long GetConnectionsForStationsQueriesCount()
	{
		return connectionsForStationsCounter;
	}
	
	/**
	 * Increases a number of queries for stations in range by one.
	 * @author Piotrek
	 */
	public static void IncreaseStationsInRangeQueriesCount()
	{
		++stationsInRangeCounter;
	}
		
	/**
	 * Increases a number of queries for connections for stations by one.
	 * @author Piotrek
	 */
	public static void IncreaseConnectionsForStationsQueriesCount()
	{
		++connectionsForStationsCounter;
	}
	
	/**
	 * Indicates whether to use rules to choose the appropriate means of transportation.
	 * @author Piotrek
	 */
	private static boolean fUseRulesChoosingMeansOfTransportation = true;
	
	/**
	 * Indicates whether to use rules to choose the appropriate means of transportation.
	 * @author Piotrek
	 * @return A boolean value indicating whether to use rules to choose the appropriate means of transportation.
	 */
	public static boolean ShouldUseRulesChoosingMeansOfTransportation()
	{
		return fUseRulesChoosingMeansOfTransportation;
	}
	
	/**
	 * Sets a value indicating whether to use rules to choose the appropriate means of transportation.
	 * @author Piotrek
	 * @param fNewValue A new value of this parameter.
	 */
	public static void SetUseRulesChoosingMeansOfTransportation(boolean fNewValue)
	{
		fUseRulesChoosingMeansOfTransportation = fNewValue;
	}
	
	/**
	 * Indicates whether to use rules dropping some of the possible changes.
	 * @author Piotrek
	 */
	private static boolean fUseRulesDroppingChanges = true;
	
	/**
	 * Indicates whether to use rules dropping some of the possible changes.
	 * @author Piotrek
	 * @return A boolean value indicating whether to use rules dropping some of the possible changes.
	 */
	public static boolean ShouldUseRulesDroppingChanges()
	{
		return fUseRulesDroppingChanges;
	}
	
	/**
	 * Sets a value indicating whether to use rules dropping some of the possible changes.
	 * @author Piotrek
	 * @param fNewValue A new value of this parameter.
	 */
	public static void SetUseRulesDroppingChanges(boolean fNewValue)
	{
		fUseRulesDroppingChanges = fNewValue;
	}
	
	/**
	 * Indicates whether to limit the number of starting connections or alternative changes.
	 * @author Piotrek
	 */
	private static boolean fLimitAlternativeChangesConnections = true;
	
	/**
	 * Indicates whether to limit the number of starting connections or alternative changes.
	 * @author Piotrek
	 * @return A boolean value indicating whether to limit the number of starting connections or alternative changes.
	 */
	public static boolean ShouldLimitAlternativeChangesConnections()
	{
		return fLimitAlternativeChangesConnections;
	}
	
	/**
	 * Sets a value indicating whether to limit the number of starting connections or alternative changes.
	 * @author Piotrek
	 * @param fNewValue A new value of this parameter.
	 */
	public static void SetLimitAlternativeChangesConnections(boolean fNewValue)
	{
		fLimitAlternativeChangesConnections = fNewValue;
	}
	
	/**
	 * Indicates whether to ignore the continent and island names information while planning.
	 * @author Piotrek
	 */
	private static boolean fIgnoreContinentIslandInformation = false;
	
	/**
	 * Indicates whether to ignore the continent and island names information while planning.
	 * @author Piotrek
	 * @return A boolean value indicating whether to ignore the continent and island names information while planning.
	 */
	public static boolean ShouldIgnoreContinentIslandInformation()
	{
		return fIgnoreContinentIslandInformation;
	}
	
	/**
	 * Sets a value indicating whether to ignore the continent and island names information while planning.
	 * @author Piotrek
	 * @param fNewValue A new value of this parameter.
	 */
	public static void SetIgnoreContinentIslandInformation(boolean fNewValue)
	{
		fIgnoreContinentIslandInformation = fNewValue;
	}
}
