/*
 * Created on 2005-05-16
 */
package scheduler;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import scheduler.modules.ConnectionsFinderTasksManager;
import scheduler.modulesinterfaces.ChangesJudger;
import scheduler.modulesinterfaces.ConnectionsTravelInfoProvider;
import scheduler.modulesinterfaces.MeansOfTransportationProvider;
import world.Connection;
import world.Location;
import world.MeanOfTransportation;
import world.Station;
import world.TravelTime;

/**
 * A class used to find connections for the given set of travel parts.
 * @author Piotrek
 */
public final class TravelDivisionsConnectionsFinder
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public TravelDivisionsConnectionsFinder()
	{
		// does nothing
	}
	
	/**
	 * A provider of the travel information on the world.
	 * @author Piotrek
	 */
	private ConnectionsTravelInfoProvider infoProvider = null;
	
	/**
	 * Sets the provider of the travel information on the world.
	 * @author Piotrek
	 * @param infoProvider A provider to be set.
	 */
	public void SetTravelInfoProvider(ConnectionsTravelInfoProvider infoProvider)
	{
		if (infoProvider == null)
			throw new NullPointerException("The travel info provider must not be null in TravelDivisionsConnectionsFinder.SetTraverlInfoProvider().");
		
		this.infoProvider = infoProvider;
	}
	
	/**
	 * Finds connections for each of the given travel divisions.
	 * @author Piotrek
	 * @param travelDivisionsList A list of travel divisions to receive the results. It must be empty.
	 * @param travelConstraints The travel constraints to be applied.
	 * @return A list of connections for each of the travel divisions.
	 */
	public ArrayList FindConnections(ArrayList travelDivisionsList, TravelConstraints travelConstraints)
	{
		if (travelDivisionsList == null)
			throw new NullPointerException("Travel divisions list must not be null in TravelDivisionsConnectionsFinder.FindConnections().");
						
		if (travelConstraints == null)
			throw new NullPointerException("Travel constraints must not be null in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (infoProvider == null)
			throw new RuntimeException("Set the travel info provider first in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (changesJudger == null)
			throw new RuntimeException("Set the changes judger first in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (tasksManager == null)
			throw new RuntimeException("Set the tasks manager first in TravelDivisionsConnectionsFinder.FindConnections().");
			
		this.travelConstraint = travelConstraints;
		
		ArrayList resultsList = new ArrayList();
		
		Date departureLimit = null;
		Date arrivalLimit = null;
				
		if (travelConstraints.fIsLimitDeparture)
		{
			departureLimit = travelConstraints.dateLimit;
			arrivalLimit = null; // no limit
			
			fRevert = false;
		}
		else
		{
			departureLimit = travelConstraints.dateLimit; // it's OK since we will revert everything later
			arrivalLimit = new Date(); // now
						
			fRevert = true;
		}
		
		Iterator iterator = travelDivisionsList.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartsList travelPartsList = (TravelPartsList)iterator.next();
			
			if (fRevert)
				travelPartsList = travelPartsList.GetRevertedList();
			
			FindConnections(travelPartsList, departureLimit, arrivalLimit, resultsList);
		}
		
		// checking if the change times still hold the limits 
		// (when walk/taxi parts get their time limits assigned it might turn out there is a longer change 
		// due to the time reservation taken by the subsequent chunks with the other means of transportation).
		// but also searching for some alter changes when no changes matching constraints are available causes this
		// (and it may also produce some too short changes)
		DropTasksBreakingConstraints(resultsList);
		
		// need to revert solutions if the reverting has been applied before
		if (fRevert)
		{
			ArrayList revertedResults = new ArrayList(resultsList.size());
			
			iterator = resultsList.iterator();
			
			while (iterator.hasNext())
			{
				ConnectionsFinderTask task = (ConnectionsFinderTask)iterator.next();
				
				ConnectionsFinderTask revertedTask = new ConnectionsFinderTask();
				
				revertedTask.connection = task.connection;
				revertedTask.dayOfWeek = task.dayOfWeek;
				
				int chunksCount = task.GetConnectionChunksCount();
				
				for (int i = chunksCount - 1; i >= 0; --i)
				{
					TravelPartConnectionChunk chunk = task.GetChunk(i);
					
					revertedTask.AddConnectionChunk(chunk.GetRevertedConnectionChunk());
				}
				
				revertedResults.add(revertedTask);
			}
			
			resultsList = revertedResults;
		}
				
		// removing similar solutions (with the same routes and time of start and end, but with different times between the start and the end)
		RemoveSimilarSolutions(resultsList);
				
		return resultsList;
	}
	
	/**
	 * Removes from the given list of solutions similar ones (with the same routes and time of start and end, but with different times between the start and the end).
	 * @author Piotrek
	 * @param solutionsList A list of solutions to be cleaned of similar solutions.
	 */
	private void RemoveSimilarSolutions(ArrayList solutionsList)
	{
		if (solutionsList == null)
			throw new NullPointerException("Solutions list must not be null in TravelDivisionsConnectionsFinder.RemoveSimilarSolutions().");
		
		ArrayList base = new ArrayList();
		
		Iterator iterator = solutionsList.iterator();
		
		while (iterator.hasNext())
		{
			ConnectionsFinderTask task = (ConnectionsFinderTask)iterator.next();
			
			// checking if there is already a similar one in the base
			
			boolean fSimilarExists = false;
			
			Iterator baseIterator = base.iterator();
			
			while (baseIterator.hasNext())
			{
				ConnectionsFinderTask baseTask = (ConnectionsFinderTask)baseIterator.next();
				
				if (task.IsSimilar(baseTask))
				{
					boolean fReplace = false;
					
					// if there is, then leave in the base the better one of them
					int taskAverageDuration = task.GetAverageChangeDuration();
					int baseTaskAverageDuration = baseTask.GetAverageChangeDuration();
					
					if (taskAverageDuration < baseTaskAverageDuration)
						fReplace = true;
					
					if (taskAverageDuration == baseTaskAverageDuration)
					{
						int taskMaxDuration = task.GetMaxChangeDuration();
						int baseTaskMaxDuration = baseTask.GetMaxChangeDuration();
						
						if (taskMaxDuration < baseTaskMaxDuration)
							fReplace = true;
					}
					
					// replacing
					if (fReplace)
					{
						base.remove(baseTask);
						
						base.add(task);
					}
					
					fSimilarExists = true;
					
					break;
				}
			}
						
			// if not, then add it to the base
			if (!fSimilarExists)
				base.add(task);
		}
		
		// leaving only the base solutions
		solutionsList.clear();
		
		solutionsList.addAll(base);
	}
	
	/**
	 * Indicates whether the search is performed with the arrival time limit (then everything is reverted).
	 * @author Piotrek
	 */
	private boolean fRevert = false;
	
	/**
	 * The travel constraint to be applied.
	 * @author Piotrek
	 */
	private TravelConstraints travelConstraint = null;
	
	/**
	 * Finds out connections for the given division of the travel matching the given time limits and other constraints.
	 * @author Piotrek
	 * @param travelPartsList A division of the travel to search for the connections for. 
	 * @param departureLimit A departure time limit.
	 * @param arrivalLimit An arrival time limit. Set it to null if there is no arrival limit.
	 * @param resultsList A list to store all the results in.
	 */
	private void FindConnections(TravelPartsList travelPartsList, Date departureLimit, Date arrivalLimit, ArrayList resultsList)
	{
		if (travelPartsList == null)
			throw new NullPointerException("Travel parts list must not be null in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (departureLimit == null)
			throw new NullPointerException("Departure time limit must not be null in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (!fRevert && arrivalLimit != null && !departureLimit.before(arrivalLimit))
			throw new IllegalArgumentException("The departure date limit must be earlier then the arrival date limit in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (fRevert && arrivalLimit != null && !departureLimit.after(arrivalLimit))
			throw new IllegalArgumentException("The departure date limit must be later then the arrival date limit since reverting in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (resultsList == null)
			throw new NullPointerException("Results list must not be null in TravelDivisionsConnectionsFinder.FindConnections().");
		
		if (travelPartsList.GetTravelPartsCount() == 0)
			return;
		
		// a list to keep all the routes which are under construction (updated each time a travel part is completed)
		// it is a list of tasks (in the end each of them will hold a list of connection chunks constituting a complete travel plan)
		ArrayList routes = new ArrayList();
				
		// searching for the connections
		boolean fHasInitiatedNonWalkTaxiPart = false;
			
		Iterator iterator = travelPartsList.GetTravelParts();
				
		while (iterator.hasNext())
		{
			TravelPart travelPart = (TravelPart)iterator.next();
			
			int meanOfTransportation = travelPart.GetMeanOfTransportation();
			
			// non stations-only-based means of transportation
			if (meanOfTransportation == MeanOfTransportation.WALK || meanOfTransportation == MeanOfTransportation.TAXI)
			{
				TravelPartConnectionChunk connectionChunk = new TravelPartConnectionChunk(meanOfTransportation, travelPart.GetOrigin(), travelPart.GetDestination());
				
				connectionChunk.SetOriginStation(travelPart.GetOriginStation());
				connectionChunk.SetDestinationStation(travelPart.GetDestinationStation());
				
				connectionChunk.SetEndDate(null); // to be set at the later time...
				
				ArrayList tasksToBeRemoved = new ArrayList();
				
				Iterator routesIterator = routes.iterator();
				
				while (routesIterator.hasNext())
				{
					ConnectionsFinderTask task = (ConnectionsFinderTask)routesIterator.next();
					
					TravelPartConnectionChunk chunkCopy = (TravelPartConnectionChunk)connectionChunk.clone();
					
					chunkCopy.SetStartDate(task.GetLastConnectionChunk().GetEndDate());
					
					task.AddConnectionChunk(chunkCopy);
					
					task.currentStation = travelPart.GetDestinationStation();
					task.startStation = travelPart.GetDestinationStation();
					
					task.connection = null;
					task.dayOfWeek = 0;
					
					task.startStationDepartureDate = null;
					
					if (task.GetChangesCount() > travelConstraint.maxChangesCount)
						tasksToBeRemoved.add(task);
				}
				
				if (routes.size() == tasksToBeRemoved.size() && routes.size() > 0) // all so far found solutions broke the max changes count constraint after update of them, hence there are no solutions to this travel division
					return;
				
				// removing those that broke the max changes constraint, it won't be all
				routes.removeAll(tasksToBeRemoved);
								
				if (routes.size() == 0) // there were no tasks so far (it is the first part in the whole travel)
				{
					ConnectionsFinderTask task = new ConnectionsFinderTask();
					
					connectionChunk.SetStartDate(null); // it will be computed later
					
					task.AddConnectionChunk(connectionChunk);
					
					task.currentStation = travelPart.GetDestinationStation();
					task.startStation = travelPart.GetDestinationStation();
					
					task.connection = null;
					task.dayOfWeek = 0;
					
					task.startStationDepartureDate = null;
					
					if (task.GetChangesCount() <= travelConstraint.maxChangesCount)
						routes.add(task);
				}
			}
			else // stations-only-based means of transportation
			{
				tasksManager.Clear();
				
				if (fHasInitiatedNonWalkTaxiPart) // continuation of the previous travel parts
				{
					// continuing the so far found connections for the previous travel parts
					Iterator routesIterator = routes.iterator();
					
					while (routesIterator.hasNext())
					{
						ConnectionsFinderTask task = (ConnectionsFinderTask)routesIterator.next();
					
						// searching for the possible changes
						Station changeStation = travelPart.GetOriginStation();//task.GetLastConnectionChunk().GetDestinationStation(); // cannot be null
						
						TravelPartConnectionChunk lastWithEndDate = task.GetLastConnectionChunkWithEndDate();
						
						if (lastWithEndDate == null) // skipping if there was none
							continue;
						
						Date changeStartTime = (Date)lastWithEndDate.GetEndDate().clone(); // there must be an end date chunk since there was an non walk/taxi travel part
						
						int sign = 1;
						
						if (fRevert) // reverting
							sign = -1;
						
						// checking if there is a need to reserve some time for the previous WALK/TAXI travel part
						if (task.GetLastConnectionChunk() != task.GetLastConnectionChunkWithEndDate())
						{							
							TravelPartConnectionChunk lastChunk = task.GetLastConnectionChunk();
							
							if (lastChunk.GetMeanOfTransportation() == MeanOfTransportation.WALK)
								TravelTime.ShiftDate(changeStartTime, sign * Math.max(GetWalkReservationTime(lastChunk), travelConstraint.minChangeTime));
							
							if (lastChunk.GetMeanOfTransportation() == MeanOfTransportation.TAXI)
								TravelTime.ShiftDate(changeStartTime, sign * (GetTaxiDriveReservationTime(lastChunk) + 2 * travelConstraint.minChangeTime));
						}
						
						int extraTimeBeforeDeparture = 0;
						
						// checking if there is a need to reserve some extra time for being earlier (airports and harbours)
						if (meanOfTransportation == MeanOfTransportation.AIRPLANE)
							extraTimeBeforeDeparture = travelConstraint.timeBeforeFlight;
						
						if (meanOfTransportation == MeanOfTransportation.SHIP)
							extraTimeBeforeDeparture = travelConstraint.timeBeforeCruise;
						
						if (extraTimeBeforeDeparture > 0)
							TravelTime.ShiftDate(changeStartTime, sign * extraTimeBeforeDeparture);
					
						// getting a list of possible changes to the new mean of transportation
						ArrayList continuationConnections = GetChangeConnections(travelPart.GetMeanOfTransportation(), changeStation, travelPart.GetDestinationStation(), changeStartTime, arrivalLimit, new Boolean(false));
						
						if (continuationConnections.size() == 0) // there are no changes within the time constraints and/or a direct change
							continuationConnections = GetDeparturingConnections(travelPart.GetMeanOfTransportation(), changeStation, travelPart.GetDestinationStation(), changeStartTime, arrivalLimit, MAX_ALTER_CHANGE_CONNECTIONS_COUNT);
						
						// adding them to the copies of the task and subsequently the result tasks to the tasks manager
						Iterator connectionsIterator = continuationConnections.iterator();
						
						while (connectionsIterator.hasNext())
						{
							ConnectionsFinderTask changeTask = (ConnectionsFinderTask)connectionsIterator.next();
														
							ConnectionsFinderTask taskCopy = (ConnectionsFinderTask)task.clone();
							
							// updating (marking the next connection to consider)															
							taskCopy.SetCurrentWorkingData(changeTask);
							
							// adding to the tasks manager
							if (taskCopy.GetChangesCount() <= travelConstraint.maxChangesCount)
							{
								taskCopy.ClearVisitedStations();
								taskCopy.AddVisitedStation(changeStation);
								
								// checking if it is a direct connection
								taskCopy.fDirect = IsDirectConnection(taskCopy.connection.GetStationIndex(changeStation), taskCopy.connection.GetStationIndex(travelPart.GetDestinationStation()));
								
								taskCopy.extraTimeBeforeDeparture = extraTimeBeforeDeparture;
								
								tasksManager.AddTask(taskCopy);
							}
						}
					}
				}
				else // initiating the possible connections list 
				{
					// searching for the initial connections
					
					Date searchStartTime = (Date)departureLimit.clone();
					
					int extraTimeBeforeDeparture = 0;
					
					// checking if to reserve some time for the previous WALK/TAXI part or if it already has an end date
					if (routes.size() == 1) // there was a WALK or TAXI travel part before this one 
					{
						int timeShift = 0; // shift in minutes
						
						ConnectionsFinderTask previousTask = (ConnectionsFinderTask)routes.get(0);
						
						TravelPartConnectionChunk lastChunk = previousTask.GetLastConnectionChunk();
						
						if (lastChunk.GetEndDate() != null) // there is an end date
							searchStartTime = lastChunk.GetEndDate();
						else // there is not, reserving some extra time
						{
							int sign = 1;
							
							if (fRevert)
								sign = -1;
							
							// walk/taxi reservation
							if (lastChunk.GetMeanOfTransportation() == MeanOfTransportation.WALK)
								timeShift = sign * Math.max(GetWalkReservationTime(lastChunk), travelConstraint.minChangeTime);
							
							if (lastChunk.GetMeanOfTransportation() == MeanOfTransportation.TAXI)
								timeShift = sign * (GetTaxiDriveReservationTime(lastChunk) + travelConstraint.minChangeTime);
												
							// time before flight/cruise						
							if (meanOfTransportation == MeanOfTransportation.AIRPLANE)
								extraTimeBeforeDeparture = travelConstraint.timeBeforeFlight;
							
							if (meanOfTransportation == MeanOfTransportation.SHIP)
								extraTimeBeforeDeparture = travelConstraint.timeBeforeCruise;
							
							if (extraTimeBeforeDeparture > 0)
								timeShift += sign * extraTimeBeforeDeparture;
							
							TravelTime.ShiftDate(searchStartTime, timeShift);
						}
					}
					
					ArrayList startingConnections = GetDeparturingConnections(travelPart.GetMeanOfTransportation(), travelPart.GetOriginStation(), travelPart.GetDestinationStation(), searchStartTime, arrivalLimit, MAX_STARTING_CONNECTIONS_COUNT);
					
					fHasInitiatedNonWalkTaxiPart = (startingConnections.size() > 0);
					
					// adding them to the task manager
					Iterator connectionsIterator = startingConnections.iterator();
					
					while (connectionsIterator.hasNext())
					{
						ConnectionsFinderTask startTask = (ConnectionsFinderTask)connectionsIterator.next();
						
						// checking if it is a direct connection
						startTask.fDirect = IsDirectConnection(startTask.connection.GetStationIndex(travelPart.GetOriginStation()), startTask.connection.GetStationIndex(travelPart.GetDestinationStation()));
						
						if (routes.size() == 0) // there were no taxi/walk travel parts before
						{
							startTask.extraTimeBeforeDeparture = extraTimeBeforeDeparture;
							
							tasksManager.AddTask(startTask);
							
							startTask.AddVisitedStation(travelPart.GetOriginStation());
						}
						else // there were, need to update them
						{
							Iterator routesIterator = routes.iterator();
							
							while (routesIterator.hasNext())
							{
								// updating their copies (marking a connection to consider)
								ConnectionsFinderTask routesTask = (ConnectionsFinderTask)routesIterator.next();
								
								ConnectionsFinderTask routesTaskCopy = (ConnectionsFinderTask)routesTask.clone(); 
																
								routesTaskCopy.SetCurrentWorkingData(startTask);
															
								if (routesTaskCopy.GetChangesCount() <= travelConstraint.maxChangesCount)
								{
									// it is a new travel part
									routesTaskCopy.ClearVisitedStations();
									routesTaskCopy.AddVisitedStation(startTask.currentStation);
									
									routesTaskCopy.extraTimeBeforeDeparture = extraTimeBeforeDeparture;
								
									tasksManager.AddTask(routesTaskCopy);
								}
							}
						}
					}
				}
				
				// solving all the prepared tasks
				ArrayList completeTravelPartSolutions = new ArrayList(); // it will hold all solutions to this travel part
				
				while (tasksManager.GetTasksCount() > 0)
				{
					ConnectionsFinderTask task = tasksManager.GetTask();
					
					// continuing the current task's connection (going one station forward)
					int currentStationIndex = task.connection.GetStationIndex(task.currentStation);
										
					if ((!fRevert && currentStationIndex < task.connection.GetStations().size() - 1) 
			            || (fRevert && currentStationIndex > 0)) // if there is the next station at all
					{
						int startStationIndex = task.connection.GetStationIndex(task.startStation);
						
						int startStationDepartureTime = task.connection.GetDepartureTime(startStationIndex);
						
						// going to the next station of the connection
						Station nextStation;
						
						if (!fRevert)
							nextStation = (Station)task.connection.GetStations().get(++currentStationIndex);
						else
							nextStation = (Station)task.connection.GetStations().get(--currentStationIndex);
						
						if (!task.HasVisitedStation(nextStation)) // only if haven't already visited this station
						{
							task.currentStation = nextStation; // marking the current station
							
							task.AddVisitedStation(nextStation);
							
							// checking if it is not the destination
							if (nextStation.IsEqual(travelPart.GetDestinationStation())) // reached the goal
							{
								// adding a new connection chunk to the task
								TravelPartConnectionChunk connectionChunk = new TravelPartConnectionChunk(task.connection.GetMeanOfTransportation(), task.startStation.GetLocation(), task.currentStation.GetLocation());
								
								connectionChunk.SetRealConnection(task.connection);
														
								connectionChunk.SetOriginStation(task.startStation);
								connectionChunk.SetOriginStationIndex(startStationIndex);
								
								connectionChunk.SetStartDate(task.startStationDepartureDate);
														
								connectionChunk.SetDestinationStation(task.currentStation);
								connectionChunk.SetDestinationStationIndex(currentStationIndex);
								
								connectionChunk.extraTimeBeforeDeparture = task.extraTimeBeforeDeparture;
								
								int endStationArrivalTime = task.connection.GetArrivalTime(currentStationIndex);
								
								Date endTime = (Date)task.startStationDepartureDate.clone();
								
								TravelTime.ShiftDate(endTime, endStationArrivalTime - startStationDepartureTime); // it is ok if reverting as the difference will be negative
								
								if (fRevert && endTime.before(arrivalLimit)) // it breakes the limitation
									continue;
										
								connectionChunk.SetEndDate(endTime);
								
								task.AddConnectionChunk(connectionChunk);
								
								// adding the task to the list of complete solutions of the given travel part
								completeTravelPartSolutions.add(task);
								
								// take the next task
								continue;
							}
														
							// it is not the destination station - going further
										
							// searching for the possible changes, if there is a place for another one
							if (task.GetChangesCount() < travelConstraint.maxChangesCount) 
							{	
								// getting the change time
								int currentStationArrivalTime = task.connection.GetArrivalTime(currentStationIndex);
													
								Date changeStartTime = (Date)task.startStationDepartureDate.clone();
								
								TravelTime.ShiftDate(changeStartTime, currentStationArrivalTime - startStationDepartureTime);
								
								// searching for the change candidates
								Boolean fTooLongChangesExist = new Boolean(false); 
								
								ArrayList changeConnections = GetChangeConnections(travelPart.GetMeanOfTransportation(), task.currentStation, travelPart.GetDestinationStation(), changeStartTime, arrivalLimit, fTooLongChangesExist);
								
								// if there are no possibilities then check for ones with longer change time
								if (changeConnections.size() == 0 && fTooLongChangesExist.booleanValue())
								{								
									// need to modify the change time (shifting it by the min change time)
									int sign = 1;
									
									if (fRevert) // reverting
										sign = -1;
									
									TravelTime.ShiftDate(changeStartTime, sign * travelConstraint.minChangeTime);
									
									changeConnections = GetDeparturingConnections(travelPart.GetMeanOfTransportation(), task.currentStation, travelPart.GetDestinationStation(), changeStartTime, arrivalLimit, MAX_ALTER_CHANGE_CONNECTIONS_COUNT);
								}
														
								// going through them to check wich are possible (by the means of change rules)
								Iterator connectionsIterator = changeConnections.iterator();
								
								while (connectionsIterator.hasNext())
								{
									ConnectionsFinderTask changeTask = (ConnectionsFinderTask)connectionsIterator.next();
									
									// skipping the current task's connection
									if (changeTask.connection.IsEqual(task.connection))
										continue;
									
									// checking if it is a direct connection 
									changeTask.fDirect =  IsDirectConnection(currentStationIndex, changeTask.connection.GetStationIndex(travelPart.GetDestinationStation()));
										
									// skipping the change, if we know it is a bad idea to take it (rules)
									if (changesJudger.ShouldDropChange(task, changeTask, travelPart, fRevert))
										continue;

									// preparing the change
									ConnectionsFinderTask taskCopy = (ConnectionsFinderTask)task.clone();
									
									// adding a connection chunk to the copy task
									TravelPartConnectionChunk connectionChunk = new TravelPartConnectionChunk(taskCopy.connection.GetMeanOfTransportation(), taskCopy.startStation.GetLocation(), taskCopy.currentStation.GetLocation());
									
									connectionChunk.SetRealConnection(taskCopy.connection);
															
									connectionChunk.SetOriginStation(taskCopy.startStation);
									connectionChunk.SetOriginStationIndex(startStationIndex);
									
									connectionChunk.SetStartDate(taskCopy.startStationDepartureDate);
															
									connectionChunk.SetDestinationStation(taskCopy.currentStation);
									connectionChunk.SetDestinationStationIndex(currentStationIndex);
											
									connectionChunk.SetEndDate(changeStartTime);
									
									connectionChunk.extraTimeBeforeDeparture = taskCopy.extraTimeBeforeDeparture;
									
									taskCopy.AddConnectionChunk(connectionChunk);
															
									// marking the change connection to be taken											
									taskCopy.SetCurrentWorkingData(changeTask);
									
									taskCopy.extraTimeBeforeDeparture = 0;
									
									// adding new task to the tasks manager
									if (taskCopy.GetChangesCount() <= travelConstraint.maxChangesCount)
										tasksManager.AddTask(taskCopy);
								}
							}
							
							// adding the current task back to the manager for further exploration
							tasksManager.AddTask(task);
						}
					}
				}
												
				// replacing the routes list with those that solve the current travel part
				routes = completeTravelPartSolutions;
				
				if (routes.size() == 0) // there are no possibilities for the given travel division
					return;
			}
		}
		
		// finding out the dates for the connection chunks with the unset dates
		SolveUnsetDates(routes, departureLimit, arrivalLimit);
		
		// updating the taxi connection chunks (shortening their duration to have time for a change - it has been reserved before)
		UpdateTaxiConnectionChunksDates(routes);
					
		// adding found solutions to the result list
		resultsList.addAll(routes);
	}
	
	/**
	 * Shortens the duration of the taxi connection chunks so there will be time for a change.
	 * @author Piotrek
	 * @param routes A list of tasks to be updated.
	 */
	private void UpdateTaxiConnectionChunksDates(ArrayList routes)
	{
		if (routes == null)
			throw new NullPointerException("Routes must nit be null in TravelDivisionsConnectionsFinder.UpdateTaxiConnectionChunksDates().");
		
		Iterator iterator = routes.iterator();
		
		while (iterator.hasNext())
		{
			ConnectionsFinderTask task = (ConnectionsFinderTask)iterator.next();
			
			Iterator chunksIterator = task.GetConnectionChunks();
			
			boolean fFirst = true;
			
			while (chunksIterator.hasNext())
			{
				TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)chunksIterator.next();
				
				if (chunk.GetMeanOfTransportation() != MeanOfTransportation.TAXI)
					continue;
				
				boolean fLast = !chunksIterator.hasNext();
				
				// updating the dates
				
				if  (!fLast)
				{
					// shift back the end date
					Date testEndDate = (Date)chunk.GetEndDate().clone();
					
					int sign = -1;
					
					if (fRevert)
						sign = 1;
					
					TravelTime.ShiftDate(testEndDate, sign * travelConstraint.minChangeTime);
					
					if (fRevert && testEndDate.before(chunk.GetStartDate()))
						chunk.SetEndDate(testEndDate);
					
					if (!fRevert && testEndDate.after(chunk.GetStartDate()))
						chunk.SetEndDate(testEndDate);
				}
								
				if (!fFirst)
				{
					// shift the start date
					Date testStartDate = (Date)chunk.GetStartDate().clone();
					
					int sign = 1;
					
					if (fRevert)
						sign = -1;
					
					TravelTime.ShiftDate(testStartDate, sign * travelConstraint.minChangeTime);
					
					if (fRevert && testStartDate.after(chunk.GetEndDate()))
						chunk.SetStartDate(testStartDate);
					
					if (!fRevert && testStartDate.before(chunk.GetEndDate()))
						chunk.SetStartDate(testStartDate);
				}
				
				fFirst = false;
			}
		}
	}
		
	/**
	 * A judger of changes.
	 * @author Piotrek
	 */
	private ChangesJudger changesJudger = null;
	
	/**
	 * Sets the judger of changes.
	 * @author Piotrek
	 * @param changesJudger A judger of changes to be used.
	 */
	public void SetChangesJudger(ChangesJudger changesJudger)
	{
		if (changesJudger == null)
			throw new NullPointerException("ChangesJudger must not be null in TravelDivisionsConnectionsFinder.SetChangesJudger().");
		
		this.changesJudger = changesJudger;
	}
	
	/**
	 * A tasks manager.
	 * @author Piotrek
	 */
	private ConnectionsFinderTasksManager tasksManager = null;
	
	/**
	 * Sets the tasks manager.
	 * @author Piotrek
	 * @param tasksManager A tasks manager to be used.
	 */
	public void SetTasksManager(ConnectionsFinderTasksManager tasksManager)
	{
		if (tasksManager == null)
			throw new NullPointerException("Tasks manager must not be null in TravelDivisionsConnectionsFinder.SetTasksManager().");
		
		this.tasksManager = tasksManager;
	}
	
	/**
	 * Removes from the given list of tasks those that break the constraints. If it would cause a removal of all the tasks, then none are removed.
	 * @author Piotrek
	 * @param tasksList A list of tasks to be checked.
	 */
	private void DropTasksBreakingConstraints(ArrayList tasksList)
	{
		ArrayList tasksToBeRemoved = new ArrayList();
		
		int tasksCount = tasksList.size();
		
		// going from the last to the first task
		for (int i = tasksCount - 1; i >= 0; --i)
		{
			ConnectionsFinderTask task = (ConnectionsFinderTask)tasksList.get(i);
			
			boolean fRemove = false;
			
			int chunksCount = task.GetConnectionChunksCount();
			
			TravelPartConnectionChunk previousChunk = null; 
				
			// going through the connection chunks
			for (int j = 0; j < chunksCount; ++j)
			{
				TravelPartConnectionChunk chunk = task.GetChunk(j);
				
				if (previousChunk != null) // if there was a pravious chunk (there have to be one to have a change)
				{
					int timeDifference = 0;
					
					int meanOfTransportation = chunk.GetMeanOfTransportation();
										
					// change is when there was something else then a walk, now it's a walk and there is something else after the walk
					if (meanOfTransportation == MeanOfTransportation.WALK && j < chunksCount - 1)
					{
						// to compute the change duration it is enough to take the duration of walk as it has dates set to the end and start of its neighbouring chunks
						if (!fRevert)
							timeDifference = TravelTime.GetDatesDifference(chunk.GetStartDate(), chunk.GetEndDate());
						else
							timeDifference = TravelTime.GetDatesDifference(chunk.GetEndDate(), chunk.GetStartDate());
													
						// if the change time is too long
						if (timeDifference > travelConstraint.maxChangeTime)
							fRemove = true;
						else
						if (timeDifference < travelConstraint.minChangeTime)
							fRemove = true;
					}
					else
					// change is also when there is a time difference for two chunks both of which
					// are of other mean of transportation then a walk
					if (previousChunk.GetMeanOfTransportation() != MeanOfTransportation.WALK && meanOfTransportation != MeanOfTransportation.WALK)
					{
						if (!fRevert)
							timeDifference = TravelTime.GetDatesDifference(previousChunk.GetEndDate(), chunk.GetStartDate());
						else
							timeDifference = TravelTime.GetDatesDifference(chunk.GetStartDate(), previousChunk.GetEndDate());
						
						timeDifference -= chunk.extraTimeBeforeDeparture;
						
						if (timeDifference > travelConstraint.maxChangeTime)
							fRemove = true;
						else
						if (timeDifference < travelConstraint.minChangeTime)
							fRemove = true;
					}
				}
				
				previousChunk = chunk;
				
				if (fRemove)
					break;
			}
			
			if (fRemove)
				tasksToBeRemoved.add(task);
		}
		
		// removing only if it won't remove all the tasks
		if (tasksList.size() > tasksToBeRemoved.size())
			tasksList.removeAll(tasksToBeRemoved);
	}
	
	/**
	 * Sets the start and end dates for all the connection chunks with unset dates.
	 * @author Piotrek
	 * @param tasksList A list of tasks for which to set dates for its connection tasks.
	 * @param departureLimit A departure date limit.
	 * @param arrivalLimit An arrival date limit. Set it to null if there is no limit.
	 */
	private void SolveUnsetDates(ArrayList tasksList, Date departureLimit, Date arrivalLimit)
	{
		if (tasksList == null)
			throw new NullPointerException("Tasks list must not be null in TravelDivisionsConnectionsFinder.SolveUnsetDates().");
		
		if (departureLimit == null)
			throw new NullPointerException("Departure time limit must not be null in TravelDivisionsConnectionsFinder.SolveUnsetDates().");
		
		if (!fRevert && arrivalLimit != null && !departureLimit.before(arrivalLimit))
			throw new IllegalArgumentException("The departure date limit must be earlier then the arrival date limit in TravelDivisionsConnectionsFinder.SolveUnsetDates().");
		
		if (fRevert && arrivalLimit != null && !departureLimit.after(arrivalLimit))
			throw new IllegalArgumentException("The departure date limit must be later then the arrival date limit since reverting in TravelDivisionsConnectionsFinder.SolveUnsetDates().");
		
		Iterator iterator = tasksList.iterator();
		
		while (iterator.hasNext())
		{
			ConnectionsFinderTask task = (ConnectionsFinderTask)iterator.next();
			
			if (task.HasUnsetDatesChunks())
			{
				Iterator chunksIterator = task.GetUnsetDatesChunksIndecies();
				
				while (chunksIterator.hasNext()) // iterator porvides indecies from the last to the first one 
				{
					int chunkIndex = ((Integer)chunksIterator.next()).intValue();
					
					TravelPartConnectionChunk chunk = task.GetChunk(chunkIndex);
										
					// only end date is not set
					if (chunk.GetStartDate() != null && chunk.GetEndDate() == null)
					{
						// it is not the last chunk
						if (task.GetConnectionChunksCount() > chunkIndex + 1)
						{
							TravelPartConnectionChunk nextChunk = task.GetChunk(chunkIndex + 1); 
							
							Date endDate = (Date)nextChunk.GetStartDate().clone();
							
							if (nextChunk.extraTimeBeforeDeparture > 0)
							{
								int sign = -1;
								
								if (fRevert)
									sign = 1;
								
								TravelTime.ShiftDate(endDate, sign * nextChunk.extraTimeBeforeDeparture);
							}
							
							chunk.SetEndDate(endDate);
						}
						else // it is the last chunk
						{
							Date endDate = (Date)chunk.GetStartDate().clone();
							
							if (arrivalLimit != null)
								if (endDate.before(arrivalLimit))
									throw new RuntimeException("The last connection chunk starts before the arrival time limit what is bad since reverting om TravelDivisionsConnectionsFinder.SolveUnsetDates().");
							
							// adding some extra time for a walk or a taxi drive
							int meanOfTransportation = chunk.GetMeanOfTransportation();
							
							int sign = 1;
							
							if (fRevert)
								sign = -1;
							
							if (meanOfTransportation == MeanOfTransportation.WALK)
								TravelTime.ShiftDate(endDate, sign * GetWalkReservationTime(chunk));
							else
							if (meanOfTransportation == MeanOfTransportation.TAXI)							
								TravelTime.ShiftDate(endDate, sign * (GetTaxiDriveReservationTime(chunk) + travelConstraint.minChangeTime));
							else
								throw new RuntimeException("Encountered a connection chunk with the unset date with the mean of transportation other then WALK or TAXI.");
							
							// shortening it if it is too long
							if (arrivalLimit != null)
								if (endDate.before(arrivalLimit))
									endDate = (Date)arrivalLimit.clone();
																
							chunk.SetEndDate(endDate);
						}
					}
					else
					// only begin date is not set
					if (chunk.GetStartDate() == null && chunk.GetEndDate() != null)
					{
						// it is not possible
						throw new RuntimeException("Encountered a connection chunk with only the start date unset.");
					}
					else
					// both dates are not set
					if (chunk.GetStartDate() == null && chunk.GetEndDate() == null)
					{
						if (task.GetConnectionChunksCount() > chunkIndex + 1) // if it is not the last chunk
						{
							TravelPartConnectionChunk nextChunk = task.GetChunk(chunkIndex + 1); 
														
							Date startDate = (Date)nextChunk.GetStartDate().clone();
							
							if (nextChunk.extraTimeBeforeDeparture > 0)
							{
								int sign = -1;
								
								if (fRevert)
									sign = 1;
								
								TravelTime.ShiftDate(startDate, sign * nextChunk.extraTimeBeforeDeparture);
							}
							
							Date endDate = (Date)startDate.clone();
							
							// making the start date earlier / later 
							int meanOfTransportation = chunk.GetMeanOfTransportation();
							
							int sign = -1;
							
							if (fRevert)
								sign = 1;
							
							if (meanOfTransportation == MeanOfTransportation.WALK)
								TravelTime.ShiftDate(startDate, sign * GetWalkReservationTime(chunk));
							else
							if (meanOfTransportation == MeanOfTransportation.TAXI)
								TravelTime.ShiftDate(startDate, sign * (GetTaxiDriveReservationTime(chunk) + travelConstraint.minChangeTime));
							else
								throw new RuntimeException("Encountered a connection chunk with the unset dates with the mean of transportation other then WALK or TAXI.");
							
							// shortening it if it is too long
							if (!fRevert)
							{
								if (startDate.before(departureLimit))
									startDate = (Date)departureLimit.clone();
							}
							else
							{
								if (startDate.after(departureLimit))
									startDate = (Date)departureLimit.clone();
							}
															
							chunk.SetStartDate(startDate);
							chunk.SetEndDate(endDate);
						}
						else
						{
							// it is when it was the only chunk in the whole travel
							Date endDate = (Date)departureLimit.clone();
													
							// adding some extra time for a walk or a taxi drive
							int meanOfTransportation = chunk.GetMeanOfTransportation();
							
							int sign = 1;
							
							if (fRevert) // reverting
								sign = -1;
							
							if (meanOfTransportation == MeanOfTransportation.WALK)
								TravelTime.ShiftDate(endDate, sign * GetWalkReservationTime(chunk));
							else
							if (meanOfTransportation == MeanOfTransportation.TAXI)
								TravelTime.ShiftDate(endDate, sign * GetTaxiDriveReservationTime(chunk));
							else
								throw new RuntimeException("Encountered a connection chunk with the unset dates with the mean of transportation other then WALK or TAXI.");
							
							// shortening it if it is too long
							if (arrivalLimit != null)
								if (endDate.before(arrivalLimit))
									endDate = (Date)arrivalLimit.clone();
								
							chunk.SetStartDate((Date)departureLimit.clone());
							chunk.SetEndDate(endDate);
						}
					}
				}
			}
		}
	}
	
	/**
	 * Defines a minimal duration in minutes of the walk travel part of the maximal possible length.
	 * @author Piotrek
	 */
	public static final int MAX_LENGTH_WALK_RESERVED_TIME = 5;
	
	/**
	 * Gets an amount of time in minutes needed to be reserved for the given walk connection chunk.
	 * @author Piotrek
	 * @param chunk A connection chunk for which the time is being reserved.
	 * @return An amount of time in minutes to be reserved.
	 */
	private static int GetWalkReservationTime(TravelPartConnectionChunk chunk)
	{
		double distance = Location.GetDistance(chunk.GetOrigin(), chunk.GetDestination());
				
		return (int)(MAX_LENGTH_WALK_RESERVED_TIME * (distance / MeansOfTransportationProvider.MAX_WALK_DISTANCE));
	}
	
	/**
	 * Defines a minimal duration in minutes of the taxi travel part of the maximal possible length.
	 * @author Piotrek
	 */
	public static final int MAX_LENGTH_TAXI_RESERVED_TIME = 15;
	
	/**
	 * Gets an amount of time in minutes needed to be reserved for the given taxi connection chunk for the drive only.
	 * @author Piotrek
	 * @param chunk A connection chunk for which the time is being reserved.
	 * @return An amount of time in minutes to be reserved.
	 */
	private static int GetTaxiDriveReservationTime(TravelPartConnectionChunk chunk)
	{
		double distance = Location.GetDistance(chunk.GetOrigin(), chunk.GetDestination());
				
		return (int)(MAX_LENGTH_TAXI_RESERVED_TIME * (distance / MeansOfTransportationProvider.MAX_TAXI_DISTANCE));
	}
	
	/**
	 * Gets a list of at most connectionsCountLimit possible connections leaving from the given station not before the given time limit with the minimal time spans between the time limit and the departure time.
	 * @author Piotrek
	 * @param meanOfTransportation A mean of transportation the connections are to be taken by.
	 * @param station An origin station the connections are to leave from.
	 * @param destinationStation A destination station. If there are direct connections the will be, each of them will be on the list even if that would mean breaking the count limit.
	 * @param departureLimit A departure date limit. Connections must not depart from the given station before / after (depends on reverting) this date.
	 * @param arrivalLimit An arrival time limit. Connections must not depart from the given station before this date. Set it to null if not reverting.
	 * @param connectionsCountLimit A maximal number of connections to be returned.
	 * @return A list of all possible connections matching all the given constraints. Each of the connections is added as a task. If there are no possibilities an empty list is returned.
	 */
	private ArrayList GetDeparturingConnections(int meanOfTransportation, Station station, Station destinationStation, Date departureLimit, Date arrivalLimit, int connectionsCountLimit)
	{
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("Unexpected mean of transportation in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (station == null)
			throw new NullPointerException("Station must not be null in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (destinationStation == null)
			throw new NullPointerException("Destination station must not be null in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (station.GetMeanOfTransportation() != meanOfTransportation)
			throw new IllegalArgumentException("Station supports some other mean of transportation then specified in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (departureLimit == null)
			throw new NullPointerException("Departure limit must not be null in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (fRevert && arrivalLimit == null)
			throw new NullPointerException("Arrival limit must not be null in the TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (connectionsCountLimit < 1)
			throw new IllegalArgumentException("Connections count limit must be at least one in TravelDivisionsConnectionsFinder.GetStartingConnections().");
		
		if (!TestOptions.ShouldLimitAlternativeChangesConnections()) // TIP: this 'if' might be removed in production version
			connectionsCountLimit = Integer.MAX_VALUE - 10;
		
		// a working list of qualified connections
		ArrayList candidateConncetions = new ArrayList();
		
		// a tag for this list - the timeDifference field is used to sort the candidates		
		class CandidateConnectionsListTag
		{
			public Connection connection = null;
			
			public int timeDifference = 0;
			
			public char dayOfWeek = 0;
			
			public Station nextStation = null;
			
			public boolean fDirect = false;
		}
		
		// checking what day of week the departure limit is
		Calendar calendar = Calendar.getInstance();
		
		calendar.setTime(departureLimit);
		
		char departureDayOfWeek = Connection.CalendarToConnectionDayOfWeek(calendar.get(Calendar.DAY_OF_WEEK));
			
		// converting the departure time limit to the minutes since midnight format (need time on that day of week)
		
		int departureDayTime = calendar.get(Calendar.HOUR_OF_DAY) * 60 + calendar.get(Calendar.MINUTE);
		
		int departureTimeSinceSunday = TravelTime.GetTimeSinceSunday(departureDayOfWeek, departureDayTime);
			
		// choosing the connections
		ArrayList connections = infoProvider.GetConnectionsForStation(station, meanOfTransportation);
		
		Iterator iterator = connections.iterator();
		
		while (iterator.hasNext())
		{
			Connection connection = (Connection)iterator.next();
			
			int stationIndex = connection.GetStationIndex(station);
			
			if (stationIndex != -1) // it cannot be -1 provided the info provider is working properly...
			{
				// check if the departure station is not the last station of this connection
				if (fRevert && stationIndex == 0)
					continue;
							
				if (!fRevert && stationIndex == connection.GetStationsCount() - 1)
					continue;
								
				int nextStationIndex;
				
				if (fRevert)
					nextStationIndex = stationIndex - 1;
				else
					nextStationIndex = stationIndex + 1;
				
				Station nextStation = (Station)connection.GetStations().get(nextStationIndex);
				
				// checking if it is a direct connection
				boolean fDirect =  IsDirectConnection(stationIndex, connection.GetStationIndex(destinationStation));
				
				// there is the next station, checking departure time
				int departureTimeInMinutes = connection.GetDepartureTime(stationIndex);
				
				int connectionDaysCount = departureTimeInMinutes / (24 * 60);
				int connectionDeparatureDayTime = departureTimeInMinutes % (24 * 60);
				
				char day = Connection.MONDAY;
				
				for (int i = 0; i < 7; ++i) // checking all the repetitions of this connection in the week
				{
					if ((connection.GetDayOfWeekMask() | day) > 0)
					{
						char connectionDepartureDayOfWeek = Connection.ShiftDayOfWeek(day, connectionDaysCount);
						
						int connectionDepartureTimeSinceSunday = TravelTime.GetTimeSinceSunday(connectionDepartureDayOfWeek, connectionDeparatureDayTime);
						
						int timeDifference = connectionDepartureTimeSinceSunday - departureTimeSinceSunday; 
						
						if (!fRevert && timeDifference < 0) // it is the next week, actually
							timeDifference += 7 * 24 * 60;
						
						if (fRevert && timeDifference > 0) // it is the previous week, actually
							timeDifference -= 7 * 24 * 60;
							
						if (fRevert && WillBeBeforeDate(departureLimit, timeDifference, arrivalLimit))
						{
							// loop step - take the next day of week
							day = Connection.ShiftDayOfWeek(day, 1);
						
							continue;
						}
						
						if (fRevert)
							timeDifference *= -1; // since now the sign doesn't matter (absolute value does)
						
						// trying to insert it into the working list (we need top connectionsCountLimit 
						// connections with the smallest difference times & having different next stations)
						int limit = candidateConncetions.size();						
						
						boolean fNextStationPresent = false;
						
						// first checking if it is better then any of the already added candidates
						for (int k = 0; k < limit; ++k)
						{
							CandidateConnectionsListTag tag = (CandidateConnectionsListTag)candidateConncetions.get(k);
							
							if (nextStation.IsEqual(tag.nextStation))
							{
								// if the current one is better, replacing the old one (indirect one must not replace a direct one)
								if (tag.timeDifference > timeDifference && (!tag.fDirect || (tag.fDirect && fDirect)))
								{
									tag.connection = connection;
									tag.dayOfWeek = connectionDepartureDayOfWeek;
									
									tag.timeDifference = timeDifference;
									
									tag.fDirect = fDirect;
								}
								
								fNextStationPresent = true;
								
								// breaking anyway
								break;
							}
						}
						
						// it hasn't been added so far and there is still a place for another candidate 
						// or it is a direct connection (which hasn't been added so far)
						if (!fNextStationPresent && (candidateConncetions.size() < connectionsCountLimit || fDirect))
						{
							CandidateConnectionsListTag newTag = new CandidateConnectionsListTag();
							
							newTag.connection = connection;
							newTag.dayOfWeek = connectionDepartureDayOfWeek;
							
							newTag.timeDifference = timeDifference;
							
							newTag.nextStation = nextStation;
							
							candidateConncetions.add(newTag);
						}
						else
						// it is not a direct connection and it hasn't been added so far, inserting it
						if (!fNextStationPresent && candidateConncetions.size() >= connectionsCountLimit)
						{	
							for (int k = 0; k < candidateConncetions.size(); ++k)
							{
								CandidateConnectionsListTag tag = (CandidateConnectionsListTag)candidateConncetions.get(k);
								
								// if the current one is better, replacing the old one
								if (tag.timeDifference > timeDifference)
								{
									CandidateConnectionsListTag newTag = new CandidateConnectionsListTag();
									
									newTag.connection = connection;
									newTag.dayOfWeek = connectionDepartureDayOfWeek;
									
									newTag.timeDifference = timeDifference;
									
									newTag.nextStation = nextStation;
									
									candidateConncetions.add(k, newTag);
									
									// removing all those that are breaking the count limit and are not direct connections
									for (int m = candidateConncetions.size() - 1; m >= connectionsCountLimit; --m)
									{
										tag = (CandidateConnectionsListTag)candidateConncetions.get(m);
										
										if (!tag.fDirect)
											candidateConncetions.remove(m);
									}
									
									break;
								}
							}
						}
					}
					
					// loop step - take the next day of week
					day = Connection.ShiftDayOfWeek(day, 1);
				}
			}
		}
		
		// creating the result list
		ArrayList possibleConnections = new ArrayList(candidateConncetions.size());
		
		iterator = candidateConncetions.iterator();
		
		while (iterator.hasNext())
		{
			CandidateConnectionsListTag tag = (CandidateConnectionsListTag)iterator.next();
			
			ConnectionsFinderTask task = new ConnectionsFinderTask();
						
			task.connection = tag.connection;
			task.dayOfWeek = tag.dayOfWeek;
			
			task.currentStation = station;
			task.startStation = station;
			
			Date startStationDepartureDate = (Date)departureLimit.clone();
			
			int sign = 1;
			
			if (fRevert) // reverting
				sign = -1;
			
			TravelTime.ShiftDate(startStationDepartureDate, sign * tag.timeDifference);
			
			task.startStationDepartureDate = startStationDepartureDate; 
			
			possibleConnections.add(task);
		}
				
		return possibleConnections;
	}
	
	/**
	 * Defines a maximal number of connections taken into account when starting the search for the connections leaving from the first station of the whole travel.
	 * @author Piotrek 
	 */
	public static final int MAX_STARTING_CONNECTIONS_COUNT = 5;
	
	/**
	 * Defines a maximal number of change connections when searching for some other changes if there where none
	 * matching the travel constraints, but there are still some with a change time longer then the max change time.
	 * Top MAX_ALTER_CHANGE_CONNECTIONS_COUNT of them will be taken (with the shortest change times).
	 * @author Piotrek
	 */
	public static final int MAX_ALTER_CHANGE_CONNECTIONS_COUNT = 4;
	
	/**
	 * Checks if the given date shifted by the given number of minutes will be before the given date limit.
	 * @author Piotrek
	 * @param date A date to be shifted. The given parameter will not be modified.
	 * @param shift A number of minutes to shift the date by. It must not be positive.
	 * @param dateLimit A date that is limiting the result date from below.
	 * @return A number of minutes to be added to the shift paramter. Zero means the date is latter then the given time limit.
	 */
	private boolean WillBeBeforeDate(Date date, int shift, Date dateLimit)
	{
		if (date == null)
			throw new NullPointerException("Date must not be null in TravelDivisionsConnectionsFinder.WillBeBeforeDate().");
		
		if (shift > 0)
			throw new IllegalArgumentException("Shift minutes count must not be positive in TravelDivisionsConnectionsFinder.WillBeBeforeDate().");
		
		if (dateLimit == null)
			throw new NullPointerException("Date limit must not be null in TravelDivisionsConnectionsFinder.WillBeBeforeDate().");
		
		Date clone = (Date)date.clone();
		
		TravelTime.ShiftDate(clone, shift);
		
		return clone.before(dateLimit);
	}
	
	/**
	 * Gets a list of all connections leaving from the given station not before the given change time and holding the change constraints. This method does not test if the returned connections are in line with the changes rules (like do not change to the local train if very far away from the destination).
	 * @author Piotrek
	 * @param meanOfTransportation A mean of transportation the connections are to be taken by.
	 * @param changeStation A station the connections are to leave from.
	 * @param destinationStation A destination station. The best direct change that does not hold the change time constraints will be added to the list of changes unless some other direct change holding the constrains will be added.
	 * @param changeStartTime A time since when the change connections may leave.
	 * @param arrivalLimit An arrival limit. Set it only if reverting.
	 * @param fTooLongChangesExist After a call to this method this will hold a boolean value indicating whether there were changes with the change time longer then the one allowed by the constraints.
	 * @return A list of all possible connections matching all the given constraints. Each of the connections is added as a task. If there are no possibilities an empty list is returned.
	 */
	private ArrayList GetChangeConnections(int meanOfTransportation, Station changeStation, Station destinationStation, Date changeStartTime, Date arrivalLimit, Boolean fTooLongChangesExist)
	{
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("Unexpected mean of transportation in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (changeStation == null)
			throw new NullPointerException("Change station must not be null in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (destinationStation == null)
			throw new NullPointerException("Destination station must not be null in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (changeStation.GetMeanOfTransportation() != meanOfTransportation)
			throw new IllegalArgumentException("Change station supports some other mean of transportation then specified in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (changeStartTime == null)
			throw new NullPointerException("Change start time must not be null in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (fRevert && arrivalLimit == null)
			throw new NullPointerException("Arrival limit must not be null in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		if (fTooLongChangesExist == null)
			throw new NullPointerException("Too long changes flag must not be null in the TravelDivisionsConnectionsFinder.GetChangeConnections().");
		
		fTooLongChangesExist = Boolean.FALSE;
		
		// a list of qualified connections
		ArrayList possibleConnections = new ArrayList();
				
		// checking what day of week the departure limit is
		Calendar calendar = Calendar.getInstance();
		
		calendar.setTime(changeStartTime);
		
		char changeStartTimeDayOfWeek = Connection.CalendarToConnectionDayOfWeek(calendar.get(Calendar.DAY_OF_WEEK));
			
		// converting the change start time to the minutes since midnight format (need time on that day of week)
		
		int changeStartTimeDayTime = calendar.get(Calendar.HOUR_OF_DAY) * 60 + calendar.get(Calendar.MINUTE);
		
		int changeStartTimeTimeSinceSunday = TravelTime.GetTimeSinceSunday(changeStartTimeDayOfWeek, changeStartTimeDayTime);
		
		// keeping the best direct connection the doesn't hold the time constraint to be added to the list of changes
		ConnectionsFinderTask bestDirectTask = null;
		int bestDirectTaskTimeDifference = 0;
		
		// choosing the connections
		ArrayList connections = infoProvider.GetConnectionsForStation(changeStation, meanOfTransportation);
		
		Iterator iterator = connections.iterator();
		
		boolean fAddedDirectConnection = false;
		
		while (iterator.hasNext())
		{
			Connection connection = (Connection)iterator.next();
			
			int stationIndex = connection.GetStationIndex(changeStation);
			
			if (stationIndex != -1) // it cannot be -1 provided the info provider is working properly...
			{
				int departureTimeInMinutes = connection.GetDepartureTime(stationIndex);
				
				int connectionDaysCount = departureTimeInMinutes / (24 * 60);
				int connectionDeparatureDayTime = departureTimeInMinutes % (24 * 60);
				
				char day = Connection.MONDAY;
				
				for (int i = 0; i < 7; ++i) // checking all the repetitions of this connection in the week
				{
					if ((connection.GetDayOfWeekMask() | day) > 0)
					{
						char connectionDepartureDayOfWeek = Connection.ShiftDayOfWeek(day, connectionDaysCount);
						
						int connectionDepartureTimeSinceSunday = TravelTime.GetTimeSinceSunday(connectionDepartureDayOfWeek, connectionDeparatureDayTime);
						
						int timeDifference = connectionDepartureTimeSinceSunday - changeStartTimeTimeSinceSunday; 
						
						if (!fRevert && timeDifference < 0) // it is the next week, actually
							timeDifference += 7 * 24 * 60;
						
						if (fRevert && timeDifference > 0) // it is the previous week, actually
							timeDifference -= 7 * 24 * 60;
							
						if (fRevert && WillBeBeforeDate(changeStartTime, timeDifference, arrivalLimit))
						{
							// loop step - take the next day of week
							day = Connection.ShiftDayOfWeek(day, 1);
						
							continue;
						}
						
						if (fRevert)
							timeDifference *= -1; // since now the sign mostly doesn't matter (absolute value does)
						
						// checking if it is a direct connection
						boolean fDirect =  IsDirectConnection(stationIndex, connection.GetStationIndex(destinationStation));
						
						// checking the change time constrains
						
						if (fDirect || (timeDifference >= travelConstraint.minChangeTime && timeDifference <= travelConstraint.maxChangeTime))
						{
							// adding it to the list
							ConnectionsFinderTask task = new ConnectionsFinderTask();
							
							task.connection = connection;
							task.dayOfWeek = connectionDepartureDayOfWeek;
							
							task.currentStation = changeStation;
							task.startStation = changeStation;
							
							Date startStationDepartureDate = (Date)changeStartTime.clone();
							
							int sign = 1;
							
							if (fRevert) // reverting
								sign = -1;
							
							TravelTime.ShiftDate(startStationDepartureDate, sign * timeDifference);
							
							task.startStationDepartureDate = startStationDepartureDate;
							
							if (timeDifference >= travelConstraint.minChangeTime && timeDifference <= travelConstraint.maxChangeTime)
							{
								possibleConnections.add(task);
								
								if (fDirect)
									fAddedDirectConnection = true;
							}
							else // direct change not matching the time constraints
							{
								// but take it only if it was a too long change
								if (bestDirectTask == null || (timeDifference < bestDirectTaskTimeDifference && timeDifference >= travelConstraint.minChangeTime))
								{
									bestDirectTask = task;
									
									bestDirectTaskTimeDifference = timeDifference;
								}
							}							
						}
						
						if (timeDifference > travelConstraint.maxChangeTime)
							fTooLongChangesExist = Boolean.TRUE;
					}
					
					// loop step - take the next day of week
					day = Connection.ShiftDayOfWeek(day, 1);
				}
			}
		}
		
		if (!fAddedDirectConnection && bestDirectTask != null)
			possibleConnections.add(bestDirectTask);
				
		return possibleConnections;
	}
	
	/**
	 * Checks whether a connection is a driect connection basing on the indecies of the current and destination stations'.
	 * @author Piotrek
	 * @param currentStationIndex An index of the cuurent station within a connection.
	 * @param destinationStationIndex An index of the destination station within a connection.
	 * @return A boolean value indicating whether it is a direct connection.
	 */
	private boolean IsDirectConnection(int currentStationIndex, int destinationStationIndex)
	{
		if (currentStationIndex < 0)
			throw new IllegalArgumentException("Current stationindex must not be negative in TravelConnectionsFinder.IsDirectCOnnection().");
		
		if (destinationStationIndex != -1)
		{
			if (fRevert && destinationStationIndex <= currentStationIndex)
				return true;
			
			if (!fRevert && destinationStationIndex >= currentStationIndex)
				return true;
		}
		
		return false;
	}
}
