/*
 * Created on 2005-04-30
 */
package scheduler;

import java.util.ArrayList;
import java.util.Iterator;
import scheduler.modulesinterfaces.FeedbackTravelPartsProvider;
import scheduler.modulesinterfaces.LocationsTravelInfoProvider;
import scheduler.modulesinterfaces.MeansOfTransportationProvider;
import world.Location;
import world.MeanOfTransportation;
import world.Station;

/**
 * A generator of travel parts for the given origin and destination.
 * @author Piotrek
 */
public final class TravelPartsGenerator
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public TravelPartsGenerator()
	{
		// does nothing
	}
	
	/**
	 * A provider of the travel information on the world.
	 * @author Piotrek
	 */
	private LocationsTravelInfoProvider infoProvider = null;
	
	/**
	 * Sets the provider of the travel information on the world.
	 * @author Piotrek
	 * @param infoProvider A provider to be set.
	 */
	public void SetTravelInfoProvider(LocationsTravelInfoProvider infoProvider)
	{
		if (infoProvider == null)
			throw new NullPointerException("The travel info provider must nomt be null in TravelPartsGenerator.SetTravelInfoProvider().");
		
		this.infoProvider = infoProvider;
	}
	
	/**
	 * Gets a list of possible divisions of the travel into the travel parts.
	 * @author Piotrek
	 * @param origin An origin of the travel.
	 * @param destination A destination of the travel.
	 * @param travelConstraints The travel constraints.
	 * @return A list of lists of travel parts. If there are no possible divisions, the empty list is returned. @see TravelPartsList
	 */
	public ArrayList GenerateTravelParts(Location origin, Location destination, TravelConstraints travelConstraints)
	{
		if (origin == null)
			throw new NullPointerException("Origin must not be null in TravelPartsGenerator.GenerateTravelParts().");
		
		if (destination == null)
			throw new NullPointerException("Destination must not be null in TravelPartsGenerator.GenerateTravelParts().");
		
		if (travelConstraints == null)
			throw new NullPointerException("Travel constraints must not be null in TravelPartsGenerator.GenerateTravelParts().");
		
		if (infoProvider == null)
			throw new RuntimeException("Set the info provider first in TravelPartsGenerator.GenerateTravelParts().");
		
		if (meansOfTransportationProvider == null)
			throw new RuntimeException("Set the means of transportation provider first in TravelPartsGenerator.GenerateTravelParts().");
		
		if (feedbackTravelPartsProvider == null)
			throw new RuntimeException("Set the feedback travel parts provider first in TravelPartsGenerator.GenerateTravelParts().");
		
		this.travelConstraints = travelConstraints;
		
		ArrayList divisions = GenerateParts(origin, destination, null, null, 1);
		
		RemoveDuplicatedDivisions(divisions);
		
		return divisions;
	}
	
	/**
	 * The travel constraints.
	 * @author Piotrek
	 */
	private TravelConstraints travelConstraints = null;
	
	/**
	 * Removes all duplicated divisions. Leaves only one of indentical divisions.
	 * @author Piotrek
	 * @param divisions A list of divisions to be cleared of duplicated divisions.
	 */
	private void RemoveDuplicatedDivisions(ArrayList divisions)
	{
		if (divisions.size() < 2)
			return;
		
		for (int i = divisions.size() - 2; i >= 0; --i)
		{
			TravelPartsList firstOne = (TravelPartsList)divisions.get(i);
			
			for (int j = divisions.size() - 1; j > i; --j)
			{
				TravelPartsList secondOne = (TravelPartsList)divisions.get(j);
				
				if (secondOne.IsEqual(firstOne))
					divisions.remove(j);
			}
		}
	}
	
	/**
	 * Gets a boolean value indicating whether the two given locations are close enough to each other to be accepted as a space between two travel parts.
	 * @author Piotrek
	 * @param x The first of the locations to be tested.
	 * @param y The second of the locations to be tested.
	 * @return A boolean value indicating whether the two given locations are close enough to each other to be accepted as a space between two travel parts.
	 */
	public static boolean AreCloseEnough(Location x, Location y)
	{
		if (x == null || y == null)
			throw new NullPointerException("Both locations must not be null in TravelPartsGenerator.AreCloseEnough().");
		
		if (Location.GetDistance(x, y) <= MAX_TRAVEL_PARTS_SPACE)
			return true;
		
		return false;
	}
	
	/**
	 * Defines a maximal space between two travel parts in the division.
	 * Do not change its value, otherwise the second step of the scheduling algorithm might not work properly.
	 * @author Piotrek
	 */
	public static final double MAX_TRAVEL_PARTS_SPACE = 0.0;
	
	/**
	 * Defines the deepest level at which the partitioning algorithm performs to search. Below it, the algorithm stops.
	 * @author Piotrek
	 */
	public static final int MAX_DEPTH = 4;
	
	/**
	 * Gets the max parts count with the current MAX_DEPTH value.
	 * @author Piotrek
	 * @return The max parts count.
	 */
	public static int GetMaxPartsCount()
	{
		int partsCount = 1;
		
		for (int i = 0; i < MAX_DEPTH - 1; ++i)
			partsCount = 2 * partsCount + 1;
		
		return partsCount;
	}
	
	/**
	 * Gets a list of all possible divisions into parts taken by some mean of transportation of a travel from the given origin to the given destination.
	 * @author Piotrek
	 * @param origin The origin of the travel to be divided.
	 * @param destination The destination of the travel to be devided.
	 * @param originStation The origin station. Null if the origin location is not a station. 
	 * @param destinationStation The destination station. Null if the destination location is not a station.
	 * @param depth The one-based depth of the recursive call.
	 * @return A list of lists of travel parts. If there are no possibilities, the empty list is returned.
	 */
	private ArrayList GenerateParts(Location origin, Location destination, Station originStation, Station destinationStation, int depth)
	{
		ArrayList divisions = new ArrayList();
		
		// if the depth is too big then STOP...
		if (depth > MAX_DEPTH)
			return divisions;
				
		// some statistics to be gathered
		//double divisionsLengthsSum = 0;
		
		boolean fGotCloseRoute = false;
		boolean fGotOneSideOnlyCloseRoute = false;
		
		int lastNotCloseRouteIndex = -1;
		int notCloseRoutesCount = 0;
				
		// a distance from the origin to the destination 
		double startingDistance = Location.GetDistance(origin, destination);
		
		// getting the means of transportation for the main part
		ArrayList meansOfTransportation = meansOfTransportationProvider.GetMeansOfTransportation(origin, destination);
		
		RemoveForbiddenMeansOfTransportation(meansOfTransportation);
						
		// for each mean of transportation solving out the minor parts
		Iterator iterator = meansOfTransportation.iterator();
		
		while (iterator.hasNext()) // going through all the possible means of transportations
		{
			int meanOfTransportation = ((Integer)iterator.next()).intValue();
					
			ArrayList originStations = GetStations(meanOfTransportation, origin, (int)(1.5 * startingDistance));
			ArrayList destinationStations = GetStations(meanOfTransportation, destination, (int)(1.5 * startingDistance));
						
			// TIP: the next two points might be done in parallel
			
			Integer pairWithAll = new Integer(-1);
			
			// finding out divisions for the origin and each of the origin stations			
			ArrayList originMinorParts = new ArrayList(originStations.size());
			
			ArrayList usedOriginStations = new ArrayList(originStations.size()); 
			
			ArrayList usedOriginStationsPairs = new ArrayList(originStations.size()); // defines with which destination stations they will be combined
			
			Iterator originStationsIterator = originStations.iterator();
			
			while (originStationsIterator.hasNext())
			{
				Station station = (Station)originStationsIterator.next();
								
				// ignoring loops
				if (station.IsEqual(originStation) || station.IsEqual(destinationStation))
					continue;
				
				//if (Location.GetDistance(origin, station.GetLocation()) >= startingDistance)
				//	continue;
				
				usedOriginStations.add(station);
				usedOriginStationsPairs.add(pairWithAll);
				
				originMinorParts.add(GenerateParts(origin, station.GetLocation(), originStation, station, depth + 1));
			}
			
			// finding out divisions for each of the destination stations and the destination
			ArrayList destinationMinorParts = new ArrayList(destinationStations.size());
			
			ArrayList usedDestinationStations = new ArrayList(destinationStations.size());
			
			ArrayList usedDestinationStationsPairs = new ArrayList(originStations.size()); // defines with which origin stations they will be combined
			
			Iterator destinationStationsIterator = destinationStations.iterator();
			
			while (destinationStationsIterator.hasNext())
			{
				Station station = (Station)destinationStationsIterator.next();
				
				// ignoring loops
				if (station.IsEqual(originStation) || station.IsEqual(destinationStation))
					continue;
				
				//if (Location.GetDistance(station.GetLocation(), destination) >= startingDistance)
				//	continue;
				
				usedDestinationStations.add(station);
				usedDestinationStationsPairs.add(pairWithAll);
				
				destinationMinorParts.add(GenerateParts(station.GetLocation(), destination, station, destinationStation, depth + 1));
			}
			
			// adding feedback
			if (depth <= MAX_FEEDBACK_KNOWLEDGE_DEPTH)
			{
				int pairNumber = 1;
				
				ArrayList feedbackTravelParts = feedbackTravelPartsProvider.GetFeedbackTravelParts(origin, destination, meanOfTransportation);
				
				Iterator feedbackIterator = feedbackTravelParts.iterator();
				
				while (feedbackIterator.hasNext())
				{
					FeedbackStationsPair feedbackStationsPair = (FeedbackStationsPair)feedbackIterator.next();
					
					// ignoring loops
					if (feedbackStationsPair.originStation.IsEqual(originStation) || feedbackStationsPair.originStation.IsEqual(destinationStation))
						continue;
					
					if (feedbackStationsPair.destinationStation.IsEqual(originStation) || feedbackStationsPair.destinationStation.IsEqual(destinationStation))
						continue;
					
					// checking if there is already such a pair among the normal stations
					int originIndex = usedOriginStations.indexOf(feedbackStationsPair.originStation);
					int destinationIndex = usedDestinationStations.indexOf(feedbackStationsPair.destinationStation);
					
					if (originIndex != -1 && destinationIndex != -1)
					{
						Integer originMarker = (Integer)usedOriginStationsPairs.get(originIndex);
						Integer destinantionMarker = (Integer)usedDestinationStationsPairs.get(destinationIndex);
						
						// there is already such a pair, so ignore this one
						if (originMarker.intValue() != destinantionMarker.intValue())
							continue;
					}
					
					// adding division part
					Integer pairMarker = new Integer(pairNumber);
					
					usedOriginStations.add(feedbackStationsPair.originStation);
					usedOriginStationsPairs.add(pairMarker);
					
					usedDestinationStations.add(feedbackStationsPair.destinationStation);
					usedDestinationStationsPairs.add(pairMarker);
					
					originMinorParts.add(GenerateParts(origin, feedbackStationsPair.originStation.GetLocation(), originStation, feedbackStationsPair.originStation, depth + 1));
					destinationMinorParts.add(GenerateParts(feedbackStationsPair.destinationStation.GetLocation(), destination, feedbackStationsPair.destinationStation, destinationStation, depth + 1));
					
					++pairNumber;
				}
			}
			
			// we don't need these anymore
			originStations = null;
			destinationStations = null;
		
			// building a final list of all possible divisions
						
			// skipping the current mean of transportation if there are no possibilites
			if (usedOriginStations.size() == 0 || usedDestinationStations.size() == 0)
				continue;
			
			// building
			for (int i = 0; i < usedOriginStations.size(); ++i)
			{
				Station usedOriginStation = (Station)usedOriginStations.get(i);
				
				ArrayList originMinorDivisions = (ArrayList)originMinorParts.get(i);
				
				Integer usedOriginStationPair = (Integer)usedOriginStationsPairs.get(i);
				
				// make sure there is at least one TravelPartsList to be copied at later time
				if (originMinorDivisions.size() == 0)
					originMinorDivisions.add(new TravelPartsList());
				
				// adding main part to each of them (need to make copies of the original list)
				int size = usedDestinationStations.size();
								
				for (int j = 0; j < size; ++j)
				{
					Station usedDestinationStation = (Station)usedDestinationStations.get(j);
										
					// ignore the destination station if it's the same as the origin one
					if (usedOriginStation.IsEqual(usedDestinationStation))
						continue;
					
					Integer usedDesitnationStationPair = (Integer)usedDestinationStationsPairs.get(j);
					
					if (!usedOriginStationPair.equals(usedDesitnationStationPair))
						continue;
					
					// testing if adding such a main part is logical accordingly to the rules
					// (stations may have some extra location data - continent and island name)
					// do it only if the stations are not added by the feedback rules
					if (usedOriginStationPair.equals(pairWithAll))
					{
						ArrayList possibleMeansOfTransportation = meansOfTransportationProvider.GetMeansOfTransportation(usedOriginStation.GetLocation(), usedDestinationStation.GetLocation());
												 
						if (!possibleMeansOfTransportation.contains(new Integer(meanOfTransportation)))
							continue;
					}
										
					// creating the main part
					TravelPart mainPart = new TravelPart(meanOfTransportation, usedOriginStation.GetLocation(), usedDestinationStation.GetLocation());
					
					mainPart.SetOriginStation(usedOriginStation);
					mainPart.SetDestinationStation(usedDestinationStation);
					
					ArrayList divisionsWithMainPart = new ArrayList(originMinorDivisions.size() * size);
					
					// adding the main part to the origin divisions which are close enough
					Iterator originParts = originMinorDivisions.iterator();
										
					while (originParts.hasNext())
					{
						TravelPartsList originDivision = (TravelPartsList)originParts.next();
						
						if (originDivision.GetTravelPartsCount() > 0)
							if (!AreCloseEnough(originDivision.GetLastTravelPart().GetDestination(), usedOriginStation.GetLocation()))
								continue;
														
						TravelPartsList copy = (TravelPartsList)originDivision.clone();
												
						copy.AddTravelPart(mainPart);
												
						// max changes count prunning
						if (copy.GetChangesCount() > travelConstraints.maxChangesCount)
							continue;
						
						// is it feedback???
						if (!usedOriginStationPair.equals(pairWithAll))
							copy.fContainsFeedback = true;
						
						divisionsWithMainPart.add(copy);
					}
					
					// skipping if haven't added any main part
					if (divisionsWithMainPart.size() == 0)
						continue;
					
					// adding the destination divisions
					ArrayList destinationMinorDivisions = (ArrayList)destinationMinorParts.get(j);
					
					// make sure there is at least one TravelPartsList to be copied at later time
					if (destinationMinorDivisions.size() == 0)
						destinationMinorDivisions.add(new TravelPartsList());
					
					Iterator destinationParts = destinationMinorDivisions.iterator();
					
					while (destinationParts.hasNext())
					{
						TravelPartsList destinationList = (TravelPartsList)destinationParts.next();
						
						Iterator divisionsWithMainPartIterator = divisionsWithMainPart.iterator();
						
						while (divisionsWithMainPartIterator.hasNext())
						{
							TravelPartsList divisionWithMainPart = ((TravelPartsList)divisionsWithMainPartIterator.next());
							
							// checking if they are close enough
							if (destinationList.GetTravelPartsCount() > 0)
								if (!AreCloseEnough(divisionWithMainPart.GetLastTravelPart().GetDestination(), destinationList.GetFirstTravelPart().GetOrigin()))
									continue;
														
							// creating a copy to be extended
							TravelPartsList copy = (TravelPartsList)(divisionWithMainPart.clone());
														
							copy.AddTravelParts(destinationList);
							
							// max changes count prunning
							if (copy.GetChangesCount() > travelConstraints.maxChangesCount)
								continue;
							
							// the final creation of a whole route with some optimisations
							double originDistance = Location.GetDistance(origin, copy.GetFirstTravelPart().GetOrigin());
							double destinationDistance = Location.GetDistance(copy.GetLastTravelPart().GetDestination(), destination);
							
							// check if it makes the distance smaller
							if (startingDistance > originDistance + destinationDistance)
							{
								double maxUnplannedDistance = Math.max(originDistance, destinationDistance);
								double minUnplannedDistance = Math.min(originDistance, destinationDistance);
								
								boolean fCloseOnBothSides = maxUnplannedDistance <= MAX_UNPLANNED_DISTANCE;
								boolean fCloseOnOneSide = minUnplannedDistance <= MAX_UNPLANNED_DISTANCE;
								
								if (!fGotCloseRoute || (fGotCloseRoute && fCloseOnBothSides))
								{
									/*double averageLength = divisionsLengthsSum / divisions.size();
									
									double copyLength = copy.GetLength();
									
									// ignore too long connections
									if (copyLength > 2 * startingDistance && divisions.size() > 5)
										continue;
							
									// don't add it if it is much longer then some already acquired connections
									if (divisions.size() <= 3 || copyLength <= averageLength * 2)
									{*/
										divisions.add(copy);
										
										//divisionsLengthsSum += copyLength;
										
										if (fCloseOnBothSides)
											fGotCloseRoute = true;
										else
										{
											lastNotCloseRouteIndex = divisions.size() - 1;
											
											++notCloseRoutesCount;
											
											if (fCloseOnOneSide)
												fGotOneSideOnlyCloseRoute = true;
										}
									//}
								}
							}
						}
					}
				}
			}
		}
		
		// removing some superfluous divisions
				
		// removing divisions which start/end far away from origin/destination if there are ones which don't
		if (fGotCloseRoute || (!fGotCloseRoute && fGotOneSideOnlyCloseRoute))
		{
			for (int i = lastNotCloseRouteIndex; i >= 0 && notCloseRoutesCount > 0; --i)
			{
				TravelPartsList division = (TravelPartsList)divisions.get(i);
				
				double originDistance = Location.GetDistance(origin, division.GetFirstTravelPart().GetOrigin());
				double destinationDistance = Location.GetDistance(division.GetLastTravelPart().GetDestination(), destination);
				
				double maxUnplannedDistance = Math.max(originDistance, destinationDistance);
				double minUnplannedDistance = Math.min(originDistance, destinationDistance);
				
				boolean fCloseOnBothSides = maxUnplannedDistance <= MAX_UNPLANNED_DISTANCE;
				boolean fCloseOnOneSide = minUnplannedDistance <= MAX_UNPLANNED_DISTANCE;
				
				if (fGotCloseRoute)
				{
					if (!fCloseOnBothSides)
					{
						divisions.remove(i);
					
						--notCloseRoutesCount;
					}
				}
				else // there are one side only close ones
				{
					if (!fCloseOnOneSide)
						divisions.remove(i);
					
					if (!fCloseOnBothSides)
						--notCloseRoutesCount;
				}
			}
		}
		
		// removing any routes with the unplanned distance much longer then the other ones, 
		// if there are no routes close enough on both ends
		if (!fGotCloseRoute)
		{
			if (fGotOneSideOnlyCloseRoute)
				MinimiseUnplannedDistancesPerSide(divisions, origin, destination);
			else
				MinimiseUnplannedDistances(divisions, origin, destination);
		}
				
		// removing divisions much longer then the rest of them (preserving the feedback ones)
		RemoveFarTooLongDivisions(divisions);
		
		// removing divisions with the prohibited means of transportation
		RemoveProhibitedMeansOfTransportation(divisions);
		
		return divisions;
	}
	
	/**
	 * Removes the forbidden means of transportation by user unless it would means a removal of all of them.
	 * @author Piotrek
	 * @param meansOfTransportation A list to be cleared of forbidden means of transportation.
	 */
	private void RemoveForbiddenMeansOfTransportation(ArrayList meansOfTransportation)
	{
		// removing from the list those that we know the user won't be happy with...
		
		ArrayList toBeRemoved = new ArrayList(meansOfTransportation.size());
		
		Iterator iterator = meansOfTransportation.iterator();
		
		while (iterator.hasNext())
		{
			Integer meanOfTransportation = (Integer)iterator.next();
			
			switch (meanOfTransportation.intValue())
			{
				case MeanOfTransportation.AIRPLANE:
					
					if (!travelConstraints.isAirplaneAllowed)
						toBeRemoved.add(meanOfTransportation);
				
					break;
					
				case MeanOfTransportation.TRAIN:
					
					if (!travelConstraints.isTrainAllowed)
						toBeRemoved.add(meanOfTransportation);
				
					break;
					
				case MeanOfTransportation.BUS:
					
					if (!travelConstraints.isBusAllowed)
						toBeRemoved.add(meanOfTransportation);
				
					break;
					
				case MeanOfTransportation.SHIP:
					
					if (!travelConstraints.isShipAllowed)
						toBeRemoved.add(meanOfTransportation);
				
					break;
					
				case MeanOfTransportation.TAXI:
					
					if (!travelConstraints.isTaxiAllowed)
						toBeRemoved.add(meanOfTransportation);
				
					break;
			}
		}
		
		// removing only if there are some other left
		if (toBeRemoved.size() < meansOfTransportation.size())
			meansOfTransportation.removeAll(toBeRemoved);
	}
		
	/**
	 * Removes the divisions that use the forbiden means of transportation, unless it would mean a removal of all divisions.
	 * @author Piotrek
	 * @param divisions A list of travel divisions to be cleared.
	 */
	private void RemoveProhibitedMeansOfTransportation(ArrayList divisions)
	{
		ArrayList divisionsToBeRemoved = new ArrayList();
		
		Iterator iterator = divisions.iterator();
		
		while (iterator.hasNext())
		{
			TravelPartsList travelPartsList = (TravelPartsList)iterator.next();
			
			boolean fContainProhibitedMeanOfTransportation = false;
			
			Iterator parts = travelPartsList.GetTravelParts();
			
			while (parts.hasNext())
			{
				TravelPart travelPart = (TravelPart)parts.next();
				
				switch (travelPart.GetMeanOfTransportation())
				{
					case MeanOfTransportation.AIRPLANE:
						
						if (!travelConstraints.isAirplaneAllowed)
							fContainProhibitedMeanOfTransportation = true;
						
						break;
						
					case MeanOfTransportation.BUS:
						
						if (!travelConstraints.isBusAllowed)
							fContainProhibitedMeanOfTransportation = true;
						
						break;
						
					case MeanOfTransportation.TRAIN:
						
						if (!travelConstraints.isTrainAllowed)
							fContainProhibitedMeanOfTransportation = true;
						
						break;
						
					case MeanOfTransportation.SHIP:
						
						if (!travelConstraints.isShipAllowed)
							fContainProhibitedMeanOfTransportation = true;
						
						break;
						
					case MeanOfTransportation.TAXI:
						
						if (!travelConstraints.isTaxiAllowed)
							fContainProhibitedMeanOfTransportation = true;
						
						break;
						
					case MeanOfTransportation.WALK:
						
						// this one cannot be prohibited
						
						break;
						
					default:
						
						throw new RuntimeException("Unexpected mean of transportation in TravelPartsGenerator.RemoveProhibitedMeansOfTransportation().");
				}
				
				if (fContainProhibitedMeanOfTransportation)
					break;
			}
			
			if (fContainProhibitedMeanOfTransportation)
				divisionsToBeRemoved.add(travelPartsList);
		}
		
		if (divisionsToBeRemoved.size() < divisions.size())
			divisions.removeAll(divisionsToBeRemoved);
	}
	
	/**
	 * Gets a list of stations for the given location and mean of transportation.
	 * @author Piotrek
	 * @param meanOfTransportation A mean of transportation.
	 * @param location An origin location.
	 * @param range A search radius.
	 * @see MeanOfTransportation#AIRPLANE
	 * @see MeanOfTransportation#BUS
	 * @see MeanOfTransportation#TRAIN
	 * @see MeanOfTransportation#SHIP
	 * @return A list of stations or an empty list if the range was 0.
	 */
	private ArrayList GetStations(int meanOfTransportation, Location location, int range)
	{
		if (location == null)
			throw new NullPointerException("Location must not mbe null in TravelPartsGenerator.GetStations().");
				
		if (range < 0)
			throw new IllegalArgumentException("Range must not be negative in TravelPartsGenerator.GetStations().");
				
		if (range == 0)
		{
			ArrayList emptyList = new ArrayList(1);
			
			return emptyList;
		}
				
		switch (meanOfTransportation)
		{
			case MeanOfTransportation.AIRPLANE:
				
				return GetAirports(location, range);
				
			case MeanOfTransportation.BUS:
				
				return GetBusStops(location, range);
				
			case MeanOfTransportation.TRAIN:
				
				return GetTrainStations(location, range);
				
			case MeanOfTransportation.SHIP:
				
				return GetHarbours(location, range);
				
			case MeanOfTransportation.WALK:
				
				ArrayList stations = new ArrayList(1);
				stations.add(new Station(location, "Virtual Walk Station", MeanOfTransportation.WALK));
				
				return stations;
				
			case MeanOfTransportation.TAXI:
				
				stations = new ArrayList(1);
				stations.add(new Station(location, "Virtual Taxi Station", MeanOfTransportation.TAXI));
				
				return stations;
				
			default:
				
				throw new IllegalArgumentException("Unexpected mean of transportation in TravelPartsGenerator.GenerateParts().");
		}
	}
	
	/**
	 * Removes routes with the unplanned distances much longer then the of other ones.
	 * @author Piotrek
	 * @param divisions A set of travel divisions to be cleared.
	 * @param origin An origin of the travel.
	 * @param destination A destiantion of the travel.
	 */
	private void MinimiseUnplannedDistances(ArrayList divisions, Location origin, Location destination)
	{
		if (divisions == null)
			throw new NullPointerException("Divisions must not be null in TravelPartsGenerator.MinimiseUnplannedDistances().");
		
		if (origin == null)
			throw new NullPointerException("Origin must not be null in TravelPartsGenerator.MinimiseUnplannedDistances().");
		
		if (destination == null)
			throw new NullPointerException("Destination must not be null in TravelPartsGenerator.MinimiseUnplannedDistances().");
		
		// minimaising the sum of unplanned distances on both ends
		
		// finding out statistics
		double minSum = -1;
		
		for (int i = divisions.size() - 1; i >= 0; --i)
		{
			TravelPartsList division = (TravelPartsList)divisions.get(i);
			
			double originDistance = Location.GetDistance(origin, division.GetFirstTravelPart().GetOrigin());
			double destinationDistance = Location.GetDistance(division.GetLastTravelPart().GetDestination(), destination);
			
			double sum = originDistance + destinationDistance;
			
			if (sum > 2 * minSum && minSum != -1)
				divisions.remove(i);			
			
			if (sum < minSum || minSum == -1)
				minSum = sum;
		}
		
		for (int i = divisions.size() - 1; i >= 0; --i)
		{
			TravelPartsList division = (TravelPartsList)divisions.get(i);
			
			double originDistance = Location.GetDistance(origin, division.GetFirstTravelPart().GetOrigin());
			double destinationDistance = Location.GetDistance(division.GetLastTravelPart().GetDestination(), destination);
			
			double sum = originDistance + destinationDistance;
			
			if (sum > 2 * minSum)
				divisions.remove(i);
		}
	}
	
	/**
	 * Removes routes with the unplanned distances much longer then the of other ones. Treats the origin and the destination separately.
	 * @author Piotrek
	 * @param divisions A set of travel divisions to be cleared.
	 * @param origin An origin of the travel.
	 * @param destination A destiantion of the travel.
	 */
	private void MinimiseUnplannedDistancesPerSide(ArrayList divisions, Location origin, Location destination)
	{
		if (divisions == null)
			throw new NullPointerException("Divisions must not be null in TravelPartsGenerator.MinimiseUnplannedDistancesPerSide().");
		
		if (origin == null)
			throw new NullPointerException("Origin must not be null in TravelPartsGenerator.MinimiseUnplannedDistancesPerSide().");
		
		if (destination == null)
			throw new NullPointerException("Destination must not be null in TravelPartsGenerator.MinimiseUnplannedDistancesPerSide().");
		
		// finding out the statistics		
		double minOriginUnplannedDistance = -1;
		double minDestinationUnplannedDistance = -1;
				
		for (int i = divisions.size() - 1; i >= 0; --i)
		{
			TravelPartsList division = (TravelPartsList)divisions.get(i);
			
			double originDistance = Location.GetDistance(origin, division.GetFirstTravelPart().GetOrigin());
			double destinationDistance = Location.GetDistance(division.GetLastTravelPart().GetDestination(), destination);
			
			if (originDistance > MAX_UNPLANNED_DISTANCE)
			{				
				if (originDistance > 2 * minOriginUnplannedDistance && minOriginUnplannedDistance != -1)
					divisions.remove(i);
				
				if (originDistance < minOriginUnplannedDistance || minOriginUnplannedDistance == -1)
					minOriginUnplannedDistance = originDistance;
			}
			else
			{	
				if (destinationDistance > 2 * minDestinationUnplannedDistance && minDestinationUnplannedDistance != -1)
					divisions.remove(i);
				
				if (destinationDistance < minDestinationUnplannedDistance || minDestinationUnplannedDistance == -1)
					minDestinationUnplannedDistance = destinationDistance;
			}
		}
		
		// filtering out divisions
		for (int i = divisions.size() - 1; i >= 0; --i)
		{
			TravelPartsList division = (TravelPartsList)divisions.get(i);
			
			double originDistance = Location.GetDistance(origin, division.GetFirstTravelPart().GetOrigin());
						
			if (originDistance > MAX_UNPLANNED_DISTANCE)
			{
				if (originDistance > 2 * minOriginUnplannedDistance)
					divisions.remove(i);
			}
			else
			{
				double destinationDistance = Location.GetDistance(division.GetLastTravelPart().GetDestination(), destination);
				
				if (destinationDistance > 2 * minDestinationUnplannedDistance)
					divisions.remove(i);
			}
		}
	}
	
	/**
	 * Removes travel divisions much longer then the rest of them.
	 * @author Piotrek
	 * @param divisions A list of travel divisions to be cleared of much longer ones.
	 */
	private void RemoveFarTooLongDivisions(ArrayList divisions)
	{
		// counting the average length
		
		double sum = 0;
		
		int count = 0;
		
		Iterator iterator = divisions.iterator();
				
		while (iterator.hasNext())
		{
			TravelPartsList division = (TravelPartsList)iterator.next();
			
			if (!division.fContainsFeedback)
			{
				sum += division.GetLength();
				
				++count;
			}
		}
		
		double averageLength = sum / count;
		
		// removing too long ones
		for (int i = divisions.size() - 1; i >= 0; --i)
		{
			TravelPartsList division = (TravelPartsList)divisions.get(i);
			
			// preserve the feedback ones...
			if (!division.fContainsFeedback && division.GetLength() > averageLength * 1.5 && divisions.size() > 6)
				divisions.remove(i);
		}
	}
	
	/**
	 * Defines a maximal distance route might be away from origin or destination to be considered a close one.
	 * @author Piotrek
	 */
	private static final double MAX_UNPLANNED_DISTANCE = 11;
	
	/**
	 * Gets a list of at most given number of stations being the closest ones from the given list to the given location.
	 * @author Piotrek
	 * @param stations A list of stations to choose from.
	 * @param location A location to count the distances to.
	 * @param countLimit A max number of stations to return.
	 * @return A list of selected stations.
	 */
	private ArrayList GetClosestStations(ArrayList stations, Location location, int countLimit)
	{
		ArrayList resultList = new ArrayList(countLimit);
		
		Iterator iterator = stations.iterator();
		
		while (iterator.hasNext())
		{
			Station station = (Station)iterator.next();
			
			double distance = Location.GetDistance(location, station.GetLocation());
			
			int size = resultList.size();
			
			// if the result list is empty, we need to add it automatically
			if (size == 0)
				resultList.add(station);
			
			// if the result list is not empty
			for (int i = 0; i < Math.min(size, countLimit); ++i)
			{
				Station otherStation = (Station)resultList.get(i);
				
				double otherDistance = Location.GetDistance(location, otherStation.GetLocation());
				
				if (distance < otherDistance)
					resultList.add(i, station);
				else
					if (i == size - 1 && i < countLimit)// this one is farer then all the other on the list, but there is still place for it
						resultList.add(station);
			}
		}
		
		if (resultList.size() > countLimit)
		{
			for (int i = resultList.size() - 1; i >= countLimit; --i)
				resultList.remove(i);
		}
		
		return resultList;
	}
	
	/**
	 * Gets the airports close to the given location, but no more then some limit value.
	 * @author Piotrek
	 * @param location A location around which to search for the airports.
	 * @param range A range to search for the airports in.
	 * @return A list of airports close to the given location.
	 */
	private ArrayList GetAirports(Location location, int range)
	{
		if (location == null)
			throw new IllegalArgumentException("Location must not be null in TravelPartsGenerator.GetAirports().");
		
		if (range < 1)
			throw new IllegalArgumentException("Range must be a positive number in TravelPartsGenerator.GetAirports().");
		
		ArrayList airportsInRange = null;
		
		for (int tempRange = Math.min(range, (int)AIRPORTS_START_SEARCH_RANGE); tempRange <= range; tempRange = Math.min(range, tempRange * 2))
		{
			airportsInRange = infoProvider.GetAirportsInRange(location.x - tempRange, location.y - tempRange, location.x + tempRange, location.y + tempRange);
			
			if (airportsInRange.size() > MAX_AIRPORTS_IN_RANGE_NUMBER)
				return GetClosestStations(airportsInRange, location, MAX_AIRPORTS_IN_RANGE_NUMBER);
			
			if (airportsInRange.size() == MAX_AIRPORTS_IN_RANGE_NUMBER)
				return airportsInRange;
			
			if (tempRange == range)
				break;
		}
		
		return airportsInRange;
	}
	
	/**
	 * The max number of airports to be returned when searching for airports in some range.
	 * @author Piotrek
	 */
	public static final int MAX_AIRPORTS_IN_RANGE_NUMBER = 2;//4;
	
	/**
	 * The max number of bus stops to be returned when searching for bus stops in some range.
	 * @author Piotrek
	 */
	public static final int MAX_BUS_STOPS_IN_RANGE_NUMBER = 3;
	
	/**
	 * The max number of train stations to be returned when searching for train stations in some range.
	 * @author Piotrek
	 */
	public static final int MAX_TRAIN_STATIONS_IN_RANGE_NUMBER = 3;//4;
	
	/**
	 * The max number of harbours to be returned when searching for harbours in some range.
	 * @author Piotrek
	 */
	public static final int MAX_HARBOURS_IN_RANGE_NUMBER = 2;//4;
	
	/**
	 * Gets the bus stops close to the given location, but no more then some limit value.
	 * @author Piotrek
	 * @param location A location around which to search for the bus stops.
	 * @param range A range to search for the bus stops in.
	 * @return A list of bus stops close to the given location.
	 */
	private ArrayList GetBusStops(Location location, int range)
	{
		if (location == null)
			throw new IllegalArgumentException("Location must not be null in TravelPartsGenerator.GetBusStops().");
		
		if (range < 1)
			throw new IllegalArgumentException("Range must be a positive number in TravelPartsGenerator.GetBusStops().");
		
		ArrayList busStopsInRange = null;
		
		for (int tempRange = Math.min(range, (int)BUS_STOPS_START_SEARCH_RANGE); tempRange <= range; tempRange = Math.min(range, tempRange * 2))
		{
			busStopsInRange = infoProvider.GetBusStopsInRange(location.x - tempRange, location.y - tempRange, location.x + tempRange, location.y + tempRange);
			
			if (busStopsInRange.size() > MAX_BUS_STOPS_IN_RANGE_NUMBER)
				return GetClosestStations(busStopsInRange, location, MAX_BUS_STOPS_IN_RANGE_NUMBER);
			
			if (busStopsInRange.size() == MAX_BUS_STOPS_IN_RANGE_NUMBER)
				return busStopsInRange;
			
			if (tempRange == range)
				break;
		}
		
		return busStopsInRange;
	}
	
	/**
	 * A range to start the search for the bus stops in, if the greater range value is specified.
	 * @author Piotrek
	 */
	public static final double BUS_STOPS_START_SEARCH_RANGE = 1.0;
	
	/**
	 * Gets the train stations close to the given location, but no more then some limit value.
	 * @author Piotrek
	 * @param location A location around which to search for the train stations.
	 * @param range A range to search for the train station in.
	 * @return A list of train stations close to the given location.
	 */
	private ArrayList GetTrainStations(Location location, int range)
	{
		if (location == null)
			throw new IllegalArgumentException("Location must not be null in TravelPartsGenerator.GetTrainStations().");
		
		if (range < 1)
			throw new IllegalArgumentException("Range must be a positive number in TravelPartsGenerator.GetTrainStations().");
		
		ArrayList trainStationsInRange = null;
		
		for (int tempRange = Math.min(range, (int)TRAIN_STATIONS_START_SEARCH_RANGE); tempRange <= range; tempRange = Math.min(range, tempRange * 2))
		{
			trainStationsInRange = infoProvider.GetTrainStationsInRange(location.x - tempRange, location.y - tempRange, location.x + tempRange, location.y + tempRange);
			
			if (trainStationsInRange.size() > MAX_TRAIN_STATIONS_IN_RANGE_NUMBER)
				return GetClosestStations(trainStationsInRange, location, MAX_TRAIN_STATIONS_IN_RANGE_NUMBER);
			
			if (trainStationsInRange.size() == MAX_TRAIN_STATIONS_IN_RANGE_NUMBER)
				return trainStationsInRange;
			
			if (tempRange == range)
				break;
		}
		
		return trainStationsInRange;
	}
	
	/**
	 * A range to start the search for the train stations in, if the greater range value is specified.
	 * @author Piotrek
	 */
	public static final double TRAIN_STATIONS_START_SEARCH_RANGE = 5.0;
	
	/**
	 * A range to start the search for the train stations in, if the greater range value is specified.
	 * @author Piotrek
	 */
	public static final double AIRPORTS_START_SEARCH_RANGE = 10.0;
	
	/**
	 * Gets the harbours close to the given location, but no more then some limit value.
	 * @author Piotrek
	 * @param location A location around which to search for the harbours.
	 * @param range A range to search for the harbours in.
	 * @return A list of harbours close to the given location.
	 */
	private ArrayList GetHarbours(Location location, int range)
	{
		if (location == null)
			throw new IllegalArgumentException("Location must not be null in TravelPartsGenerator.GetHarbours().");
		
		if (range < 1)
			throw new IllegalArgumentException("Range must be a positive number in TravelPartsGenerator.GetHarbours().");
		
		ArrayList harboursInRange = null;
		
		for (int tempRange = Math.min(range, (int)HARBOURS_START_SEARCH_RANGE); tempRange <= range; tempRange = Math.min(range, tempRange * 2))
		{
			harboursInRange = infoProvider.GetHarboursInRange(location.x - tempRange, location.y - tempRange, location.x + tempRange, location.y + tempRange);
			
			if (harboursInRange.size() > MAX_HARBOURS_IN_RANGE_NUMBER)
				return GetClosestStations(harboursInRange, location, MAX_HARBOURS_IN_RANGE_NUMBER);
			
			if (harboursInRange.size() == MAX_HARBOURS_IN_RANGE_NUMBER)
				return harboursInRange;
			
			if (tempRange == range)
				break;
		}
		
		return harboursInRange;
	}
	
	/**
	 * A range to start the search for the harbours in, if the greater range value is specified.
	 * @author Piotrek
	 */
	public static final double HARBOURS_START_SEARCH_RANGE = 50.0;
	
	/**
	 * A provider of means of transportation for the two given points.
	 * @author Piotrek
	 */
	private MeansOfTransportationProvider meansOfTransportationProvider = null;
	
	/**
	 * Sets the provider of the appropriate means of transportation for the two given locations.
	 * @author Piotrek
	 * @param meansOfTransportationProvider A provider of means of transportation to be used.
	 */
	public void SetMeansOfTransportationProvider(MeansOfTransportationProvider meansOfTransportationProvider)
	{
		if (meansOfTransportationProvider == null)
			throw new NullPointerException("MeansOfTransportationProvider must not be null in TravelPartsGenerator.SetMeansOfTransportationProvider().");
		
		this.meansOfTransportationProvider = meansOfTransportationProvider;
	}
	
	/**
	 * A provider of feedback travel parts.
	 * @author Piotrek
	 */
	private FeedbackTravelPartsProvider feedbackTravelPartsProvider = null;
	
	/**
	 * Sets the provider of the feedback travel parts.
	 * @author Piotrek
	 * @param feedbackTravelPartsProvider A provider of feedback travel parts to be used.
	 */
	public void SetFeedbackTravelPartsProvider(FeedbackTravelPartsProvider feedbackTravelPartsProvider)
	{
		if (feedbackTravelPartsProvider == null)
			throw new NullPointerException("FeedbackTravelPartsProvider must not be null in TravelPartsGenerator.SetFeedbackTravelPartsProvider().");
		
		this.feedbackTravelPartsProvider = feedbackTravelPartsProvider;
	}
	
	/**
	 * Defines the maximal depth at which the feedback rules are launched.
	 * @author Piotrek
	 */
	public static final int MAX_FEEDBACK_KNOWLEDGE_DEPTH = 2;
}
