/*
 * Created on 2005-04-30
 */
package scheduler;

import java.util.ArrayList;
import java.util.Iterator;
import world.Location;
import world.MeanOfTransportation;

/**
 * A list of parts of a single travel.
 * @author Piotrek
 */
public final class TravelPartsList implements Cloneable
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public TravelPartsList()
	{
		parts = new ArrayList();
	}
	
	/**
	 * A list of parts of the travel.
	 * @author Piotrek
	 */
	private ArrayList parts = null;
	
	/**
	 * Adds a new travel part to the list. The new part must start where the old one has ended. 
	 * @author Piotrek
	 * @param travelPart A travel part to be added.
	 */
	public void AddTravelPart(TravelPart travelPart)
	{
		if (travelPart == null)
			throw new NullPointerException("Travel part must not be null in the TravelPartsList.AddTravelPart().");
		
		if (parts.contains(travelPart))
			return;
		
		if (parts.size() > 0)
		{
			TravelPart lastPart = (TravelPart)parts.get(parts.size() - 1);
			
			if (lastPart.GetMeanOfTransportation() == travelPart.GetMeanOfTransportation())
				throw new IllegalArgumentException("The new travel part must have some other mean of transportation then the last one from the current list.");
			
			if (!TravelPartsGenerator.AreCloseEnough(lastPart.GetDestination(), travelPart.GetOrigin()))
				throw new IllegalArgumentException("The new travel part must starts close enough the end of the previous one in TravelPartsList.AddTravelPart().");
		}
				
		// checking if it means a change	
		int partsCount = parts.size();
		
		if (travelPart.GetMeanOfTransportation() != MeanOfTransportation.WALK && partsCount > 0)
		{
			TravelPart oldTravelPart = null;
			
			int i = partsCount - 1;
			
			while (i >= 0 && (oldTravelPart = (TravelPart)parts.get(i)).GetMeanOfTransportation() == MeanOfTransportation.WALK)
				--i;
			
			if (oldTravelPart.GetMeanOfTransportation() != MeanOfTransportation.WALK)
				++changesCount;
		}
		
		parts.add(travelPart);
	}
	
	/**
	 * Gets the iterator to the parts of the travel.
	 * @author Piotrek
	 * @return An iterator to the list of travel parts.
	 */
	public Iterator GetTravelParts()
	{
		return parts.iterator();
	}
	
	/**
	 * Creates a shallow copy of the list.
	 * @see java.lang.Object#clone()
	 * @author Piotrek
	 */
	protected Object clone()
	{
		TravelPartsList copy = new TravelPartsList();
		
		Iterator iterator = parts.iterator();
		
		while (iterator.hasNext())
		{
			copy.AddTravelPart((TravelPart)iterator.next());
		}
		
		copy.fContainsFeedback = fContainsFeedback;
		
		return copy;
	}
	
	/**
	 * Adds a new travel part to the list. The new part must start where the old one has ended. 
	 * @author Piotrek
	 * @param travelPartsList A list of travel parts to be added.
	 */
	public void AddTravelParts(TravelPartsList travelPartsList)
	{
		if (travelPartsList == null)
			throw new NullPointerException("Travel parts list must not be null in the TravelPartsList.AddTravelParts().");
		
		Iterator iterator = travelPartsList.GetTravelParts();
		
		while (iterator.hasNext())
		{
			TravelPart travelPart = (TravelPart)iterator.next();
		
			AddTravelPart(travelPart);
		}
		
		if (travelPartsList.fContainsFeedback)
			fContainsFeedback = true;
	}
	
	/**
	 * Gets the first travel part from the list.
	 * @author Piotrek
	 * @return The first travel part from the list or null if the list is empty.
	 */
	public TravelPart GetFirstTravelPart()
	{
		if (parts.size() > 0)
			return (TravelPart)parts.get(0);
		
		return null;
	}
	
	/**
	 * Gets the last travel part from the list.
	 * @author Piotrek
	 * @return The last travel part from the list or null if the list is empty.
	 */
	public TravelPart GetLastTravelPart()
	{
		if (parts.size() > 0)
			return (TravelPart)parts.get(parts.size() - 1);
		
		return null;
	}
	
	/**
	 * Gets the number of the travel parts on the list.
	 * @author Piotrek
	 * @return A number of the travel parts on the list.
	 */
	public int GetTravelPartsCount()
	{
		return parts.size();
	}
	
	/**
	 * Gets the total length of travel parts.
	 * @author Piotrek
	 * @return A total length of travel parts.
	 */
	public double GetLength()
	{
		double sum = 0;
		
		TravelPart previousTravelPart = null;
		
		Iterator iterator = parts.iterator();
		
		while (iterator.hasNext())
		{
			TravelPart travelPart = (TravelPart)iterator.next();
			
			if (previousTravelPart != null)
				sum += Location.GetDistance(previousTravelPart.GetDestination(), travelPart.GetOrigin());
			
			sum += Location.GetDistance(travelPart.GetOrigin(), travelPart.GetDestination());
			
			previousTravelPart = travelPart;
		}
		
		return sum;
	}
	
	/**
	 * Gets a boolean value indicating whether this and the given one lists are equal.
	 * @author Piotrek
	 * @param other
	 * @return A boolean value indicating whether this and the given one lists are equal.
	 */
	public boolean IsEqual(TravelPartsList other)
	{
		if (other.GetTravelPartsCount() != parts.size())
			return false;
		
		Iterator myParts = parts.iterator();
		Iterator otherParts = other.GetTravelParts();
		
		while (myParts.hasNext())
		{
			TravelPart myTravelPart = (TravelPart)myParts.next();
			TravelPart otherTravelPart = (TravelPart)otherParts.next();
			
			if (myTravelPart.GetMeanOfTransportation() != otherTravelPart.GetMeanOfTransportation())
				return false;
			
			if (!myTravelPart.GetOrigin().IsEqual(otherTravelPart.GetOrigin()))
				return false;
			
			if (!myTravelPart.GetDestination().IsEqual(otherTravelPart.GetDestination()))
				return false;
			
			if (myTravelPart.GetOriginStation() != null && otherTravelPart.GetOriginStation() != null)
			{
				if (!myTravelPart.GetOriginStation().IsEqual(otherTravelPart.GetOriginStation()))
					return false;
			}
			else
			{
				if (myTravelPart.GetOriginStation() != otherTravelPart.GetOriginStation())
					return false;
			}
			
			if (myTravelPart.GetDestinationStation() != null && otherTravelPart.GetDestinationStation() != null)
			{
				if (!myTravelPart.GetDestinationStation().IsEqual(otherTravelPart.GetDestinationStation()))
					return false;
			}
			else
			{
				if (myTravelPart.GetDestinationStation() != otherTravelPart.GetDestinationStation())
					return false;
			}
		}
		
		return true;
	}
	
	/**
	 * @author Piotrek
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object other)
	{
		if (!(other instanceof TravelPartsList))
			return false;
		
		return IsEqual((TravelPartsList)other);
	}
	
	/**
	 * Gets the changes count in the current list.
	 * @author Piotrek
	 * @return The changes count.
	 */
	public int GetChangesCount()
	{
		return changesCount;
	}
	
	/**
	 * The changes count implied by the travel parts on the list.
	 * @author Piotrek
	 */
	private int changesCount = 0;
	
	/**
	 * Gets the reverted list.
	 * @author Piotrek
	 * @return The reverted list.
	 */
	public TravelPartsList GetRevertedList()
	{
		TravelPartsList revertedList = new TravelPartsList();
		
		for (int i = parts.size() - 1; i >= 0; --i)
		{
			TravelPart travelPart = (TravelPart)parts.get(i);
			
			revertedList.AddTravelPart(travelPart.GetRevertedPart());
		}
		
		return revertedList;
	}
	
	/**
	 * Indicates whether this travel parts list contains expert knowledge feedback. It is not set by this class, except for the cloning and adding of other travel parts list.
	 * @author Piotrek
	 */ 
	public boolean fContainsFeedback = false;
}
