/*
 * Created on 2005-06-15
 */
package scheduler;

import java.util.ArrayList;
import java.util.Iterator;
import scheduler.modulesinterfaces.TravelSolutionsJudger;

/**
 * A class used to sort a given list of travel solutions (ConnectionsFinderTasks).
 * @author Piotrek
 */
public final class TravelSolutionsSorter
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param travelSolutionsJudger The travel solutions judger.
	 */
	public TravelSolutionsSorter(TravelSolutionsJudger travelSolutionsJudger)
	{
		this.travelSolutionsJudger = travelSolutionsJudger;
	}
	
	/**
	 * The CRN manager used to judge the travel solutions when sorting on similarity to the history.
	 * @author Piotrek
	 */
	TravelSolutionsJudger travelSolutionsJudger = null;
	
	/**
	 * Sorts the given list of travel solutions on the given property.
	 * @author Piotrek
	 * @param travelSolutions A list of travel solutions to be sorted.
	 * @param sortType A property of a travel solution to sort on.
	 * @see #SORT_ON_AVERAGE_CHANGE_DURATION
	 * @see #SORT_ON_CHANGES_COUNT
	 * @see #SORT_ON_MAX_CHANGE_DURATION
	 * @see #SORT_ON_SIMILARITIES_TO_HISTORY
	 * @see #SORT_ON_TOTAL_DISTANCE
	 * @see #SORT_ON_TOTAL_DURATION
	 * @return A sorted list of travel solutions.
	 */
	public ArrayList SortTravelSolutions(ArrayList travelSolutions, int sortType)
	{
		if (travelSolutions == null)
			throw new NullPointerException("Travel solutions list must not be null in the TravelSolutionsSorter.Sort().");
		
		// boxing
		ArrayList taggedList = GetTaggedList(travelSolutions);
		
		// assigning the values of the propety the sorting is performed on
		// this will throw an exception if sort type is invalid
		AssignPropertyValues(taggedList, sortType);
		
		QuickSort(taggedList, 0, taggedList.size() - 1);
		
		return GetUntaggedList(taggedList);
	}
	
	/**
	 * Gets a list of travel tasks in the same order as in the given list of SortTags.
	 * @author Piotrek
	 * @param taggedList A list of SortTags.
	 * @return An list of travel task in the same order as the given list of SortTags.
	 */
	private ArrayList GetUntaggedList(ArrayList taggedList)
	{
		if (taggedList == null)
			throw new NullPointerException("The tagged list must not be null in TravelSolutionsSorter.GetIntaggedList().");
		
		ArrayList resultList = new ArrayList(taggedList.size());
		
		Iterator iterator = taggedList.iterator();
		
		while (iterator.hasNext())
		{
			SortTag tag = (SortTag)iterator.next();
						
			resultList.add(tag.task);
		}
		
		return resultList;
	}
	
	/**
	 * Swaps two given element on the given list.
	 * @author Piotrek
	 * @param array A list on which the given elements are swapped.
	 * @param i Index of the first of elements to be swapped.
	 * @param j Index of the second of elements to be swapped
	 */
    private void Swap(ArrayList array, int i, int j)
    {
        Object tmp = array.get(i);
        
        array.set(i, array.get(j));
        
        array.set(j, tmp);
    }

    /**
     * Partitions the given range of elements in the given array using the SortTag in the middle as the pivot.
     * @author Piotrek
     * @param array An array part of which is to be partitioned.
     * @param begin The first index of the range to be partitioned.
     * @param end The last index of the range to be partitioned.
     * @return An index of the pivot element after the partitioning.
     */
    private int Partition(ArrayList array, int begin, int end)
    {
        int index = (begin + end) / 2;
        
        SortTag middleOne = (SortTag)array.get(index);
        
        Swap(array, index, end);
        
        index = begin;
        
        for (int i = begin; i < end; ++i) 
        {
        	SortTag tag = (SortTag)array.get(i);
        	
            if (tag.assignedValue <= middleOne.assignedValue)
            {
                Swap(array, index, i);
                
                ++index;
            }
        }
        
        Swap(array, index, end);
        
        return index;
    }

    /**
     * Sorts the given range of SortTags on the given list.
     * @author Piotrek
     * @param array A list to be sorted.
     * @param begin The first index to be sorted.
     * @param end The last index to be sorted.
     */
    private void QuickSort(ArrayList array, int begin, int end) 
    {
        if (end > begin) 
        {
            int index = Partition(array, begin, end);
            
            QuickSort(array, begin, index - 1);
            QuickSort(array, index + 1,  end);
        }
    }
	
	/**
	 * Assigns the values to the list of sort tags.
	 * @author Piotrek
	 * @param taggedList A list of sort tags to be evalueted.
	 * @param sortType A property of a travel solution to sort on.
	 * @see #SORT_ON_AVERAGE_CHANGE_DURATION
	 * @see #SORT_ON_CHANGES_COUNT
	 * @see #SORT_ON_MAX_CHANGE_DURATION
	 * @see #SORT_ON_SIMILARITIES_TO_HISTORY
	 * @see #SORT_ON_TOTAL_DISTANCE
	 * @see #SORT_ON_TOTAL_DURATION
	 */
	private void AssignPropertyValues(ArrayList taggedList, int sortType)
	{
		if (taggedList == null)
			throw new NullPointerException("Tagged list must not be null in the TravelSolutionsSorter.AssignPropertyValues().");
		
		Iterator iterator = taggedList.iterator();
		
		while (iterator.hasNext())
		{
			SortTag tag = (SortTag)iterator.next();
			
			switch (sortType)
			{
				case SORT_ON_AVERAGE_CHANGE_DURATION:
					
					tag.assignedValue = tag.task.GetAverageChangeDuration();
					
					break;
					
				case SORT_ON_CHANGES_COUNT:
					
					tag.assignedValue = tag.task.GetChangesCount();
					
					break;
					
				case SORT_ON_MAX_CHANGE_DURATION:
					
					tag.assignedValue = tag.task.GetMaxChangeDuration();
					
					break;
					
				case SORT_ON_SIMILARITIES_TO_HISTORY:
					// minus the judgement since the sorting makes the order from the smallest to the greatest value
					tag.assignedValue = -travelSolutionsJudger.JudgeTravelSolution(tag.task);
					
					tag.task.historyBasedJudgement = tag.assignedValue;
					
					break;
					
				case SORT_ON_TOTAL_DISTANCE:
					
					tag.assignedValue = tag.task.GetTotalDistance();
					
					break;
					
				case SORT_ON_TOTAL_DURATION:
					
					tag.assignedValue = tag.task.GetTotalDuration();
					
					break;
					
				default:
					
					throw new RuntimeException("Unexpected sorting type in TravelSolutionsSorter.AssignPropertyValues().");
			}
		}
	}
	
	/**
	 * Gets a list of the given travel solutions boxed within the sort tags. It does not assign the values of the properties.
	 * @author Piotrek
	 * @param travelSolutions A list of travel solutions to be boxed.
	 * @return A list of boxed travel solutions.
	 */
	private ArrayList GetTaggedList(ArrayList travelSolutions)
	{
		if (travelSolutions == null)
			throw new NullPointerException("Travel solutions list must not be null in TravelSolutionsSorter.GetTaggedList().");
		
		ArrayList boxedSolutions = new ArrayList(travelSolutions.size());
		
		Iterator iterator = travelSolutions.iterator();
		
		while (iterator.hasNext())
		{
			ConnectionsFinderTask task = (ConnectionsFinderTask)iterator.next();
			
			SortTag tag = new SortTag();
			
			tag.task = task;
			
			boxedSolutions.add(tag);
		}
		
		return boxedSolutions;
	}
	
	/**
	 * Used to indicate that sorting should be on the total duration of the travel.
	 * @author Piotrek
	 */
	public static final int SORT_ON_TOTAL_DURATION = 0;
	
	/**
	 * Used to indicate that sorting should be on the average duration of the change.
	 * @author Piotrek
	 */
	public static final int SORT_ON_AVERAGE_CHANGE_DURATION = 1;
	
	/**
	 * Used to indicate that sorting should be on the maximal duration of the change.
	 * @author Piotrek
	 */
	public static final int SORT_ON_MAX_CHANGE_DURATION = 2;
	
	/**
	 * Used to indicate that sorting should be on the number of changes.
	 * @author Piotrek
	 */
	public static final int SORT_ON_CHANGES_COUNT = 3;
	
	/**
	 * Used to indicate that sorting should be on the similarity level to the historical cases from the CRN.
	 * @author Piotrek
	 */
	public static final int SORT_ON_SIMILARITIES_TO_HISTORY = 4;
	
	/**
	 * Used to indicate that sorting should be on the total distance of the travel.
	 * @author Piotrek
	 */
	public static final int SORT_ON_TOTAL_DISTANCE = 5;
	
	/**
	 * A tag used while sorting. It keeps a reference to the travel solution and an object holding a value of the property the sorting is performed on.
	 * @author Piotrek
	 */
	private class SortTag
	{
		/**
		 * A travel solution.
		 */
		ConnectionsFinderTask task = null;
		
		/**
		 * A value of the property the sorting is performed on for this travel solution.
		 * It is a double or an integer casted to double depanding on the sort type.
		 */
		double assignedValue = 0.0;
	}
}
