/*
 * Created on 2005-08-30
 */
package scheduler.modules;

import java.util.ArrayList;
import scheduler.ConnectionsFinderTask;
import scheduler.TestOptions;
import scheduler.TravelConstraints;
import scheduler.TravelPart;
import scheduler.modulesinterfaces.ChangesJudger;
import world.Location;
import world.Station;

/**
 * A class judging if it is wise no to take a given change.
 * @author Piotrek
 */
public final class ChangesJudgerStub implements ChangesJudger
{
	/**
	 * Checks whether it is wise to do not take the given change.
	 * @author Piotrek
	 * @param task A task describing the current connection.
	 * @param changeTask A task describing the change.
	 * @param travelPart A travel part that is being solved. It provides the destination.
	 * @param fRevert Specifies whether the planning is perfomred with the arrival date limit.
	 * @return A boolean value indicating whether to drop the given change.
	 */
	public boolean ShouldDropChange(ConnectionsFinderTask task, ConnectionsFinderTask changeTask, TravelPart travelPart, boolean fRevert)
	{
		if (task == null)
			throw new NullPointerException("Task must not be null in ChangesJudgerStub.ShouldDropChange().");
		
		if (changeTask == null)
			throw new NullPointerException("Change task must not be null in ChangesJudgerStub.ShouldDropChange().");
		
		if (travelPart == null)
			throw new NullPointerException("Travel part must not be null in ChangesJudgerStub.ShouldDropChange().");
		
		if (travelConstraints == null)
			throw new RuntimeException("Set the travel constrains first in ChangesJudgerStub.ShouldDropChange().");
		
		// using rules to decide if to take this change into consideration, or not
		
		int changeActualStationIndex = changeTask.connection.GetStationIndex(changeTask.currentStation);
		int currentActualStationIndex = task.connection.GetStationIndex(task.currentStation);
		
		if (TestOptions.ShouldUseRulesDroppingChanges()) // TIP: this 'if' might be removed in production version
		{
			// rule for direct connections (change only if the change connection reaches destination station faster 
			// and it is also a direct connection - comparing only the travel durations as the comparison of real times of
			// arrivals would be too expensive; taking also min change time into acount)
			if (task.fDirect)
			{
				// drop it if it is not a direct connection
				if (!changeTask.fDirect)
					return true;
				
				// checking if the change connection arrives sooner at the destination station
				
				int changeDestinationStationIndex = changeTask.connection.GetStationIndex(travelPart.GetDestinationStation());
				int currentDestinationStationIndex = task.connection.GetStationIndex(travelPart.GetDestinationStation());
							
				int changeTravelTime = changeTask.connection.GetArrivalTime(changeDestinationStationIndex) - changeTask.connection.GetDepartureTime(changeActualStationIndex);
				int currentTravelTime = task.connection.GetArrivalTime(currentDestinationStationIndex) - task.connection.GetDepartureTime(currentActualStationIndex);
				
				if (fRevert) // reverting - it will be enough to remove the negative sign
				{
					changeTravelTime *= -1;
					currentTravelTime *= -1;
				}
				
				// droping it if it definitely gains nothing
				if (changeTravelTime >= currentTravelTime + travelConstraints.minChangeTime)
					return true;
			}
			
			// rule of the next station (change only if the next station is different then in the current connection 
			// or the change connection arrives there faster - here also comparing only travel durations to save time)
	
			// if there are next stations in both cases
			if ((!fRevert && currentActualStationIndex + 1 < task.connection.GetStationsCount() && changeActualStationIndex + 1 < changeTask.connection.GetStationsCount())
				|| (fRevert && currentActualStationIndex - 1 >= 0 && changeActualStationIndex - 1 >= 0))
			{
				int nextStationDelta = 1;
				
				if (fRevert)
					nextStationDelta = -1;
				
				Station currentNextStation = (Station)task.connection.GetStations().get(currentActualStationIndex + nextStationDelta);
				Station changeNextStation = (Station)changeTask.connection.GetStations().get(changeActualStationIndex + nextStationDelta);
				
				// if it is the same station
				if (currentNextStation.IsEqual(changeNextStation))
				{				
					int changeTravelTime = changeTask.connection.GetArrivalTime(changeActualStationIndex + nextStationDelta) - changeTask.connection.GetDepartureTime(changeActualStationIndex);
					int currentTravelTime = task.connection.GetArrivalTime(currentActualStationIndex + nextStationDelta) - task.connection.GetDepartureTime(currentActualStationIndex);
					
					if (fRevert) // reverting - it will be enough to remove the negative sign
					{
						changeTravelTime *= -1;
						currentTravelTime *= -1;
					}
					
					// droping it if it definitely gains nothing
					if (changeTravelTime >= currentTravelTime + travelConstraints.minChangeTime)
						return true;
				}
			}
			
			// distance based rule (drop change only if the next, the middle and the end stations do have longer distances to
			// the destination station then the current one)
			
			double currentDistance = Location.GetDistance(task.currentStation.GetLocation(), travelPart.GetDestination());
			
			ArrayList changeStations = changeTask.connection.GetStations();
			
			int changeCurrentStationIndex = changeTask.connection.GetStationIndex(changeTask.currentStation);
			
			// at first three points based checking
			double changeEndStationDistance = -1;
			
			int changeEndStationIndex = -1;
			
			// it will be the current station in the worst case
			if (!fRevert)
				changeEndStationIndex = changeStations.size() - 1;
			else
				changeEndStationIndex = 0;
			
			if (changeCurrentStationIndex != changeEndStationIndex) // if the end station is not the current station
				changeEndStationDistance = Location.GetDistance(((Station)changeStations.get(changeEndStationIndex)).GetLocation(), travelPart.GetDestination());
			
			// the next station distance		
			int changeNextStationIndex = -1; 
			
			double changeNextStationDistance = -1;
			
			if (!fRevert && changeCurrentStationIndex < changeStations.size() - 1) // if there is the next station at all
				changeNextStationIndex = changeCurrentStationIndex + 1;
	
			
			if (fRevert && changeCurrentStationIndex > 0) // if there is the next station at all (reverting)
				changeNextStationIndex = changeCurrentStationIndex - 1;
			
			if (changeNextStationIndex >= 0) // if there was the next station
				changeNextStationDistance = Location.GetDistance(((Station)changeStations.get(changeNextStationIndex)).GetLocation(), travelPart.GetDestination());
			
			// the middle station distance
			double changeMiddleStationDistance = -1;
			
			int changeMiddleStationIndex = -1;
			
			if (changeNextStationIndex >= 0) // there was a next station (in the worst case the middle station will be the next station)
				changeMiddleStationIndex = (changeNextStationIndex + changeEndStationIndex) / 2;
			
			if (changeMiddleStationIndex >= 0)
				changeMiddleStationDistance = Location.GetDistance(((Station)changeStations.get(changeMiddleStationIndex)).GetLocation(), travelPart.GetDestination());
					
			// checking distances of three test points
			boolean fAllAreFarer = true;
			
			if (changeNextStationDistance >= 0 && changeNextStationDistance <= currentDistance)
				fAllAreFarer = false;
			
			if (changeMiddleStationDistance >= 0 && changeMiddleStationDistance <= currentDistance)
				fAllAreFarer = false;
			
			if (changeEndStationDistance >= 0 && changeEndStationDistance <= currentDistance)
				fAllAreFarer = false;
			
			if (fAllAreFarer && changeNextStationIndex >= 0) 
			{
				// it appears it should be dropped... need to check all the rest of stations to make sure
				int delta = 1;
				
				if (fRevert)
					delta = -1;
				
				int index = changeCurrentStationIndex + 2 * delta; // the station after the next station
							
				while (index >= 0 && index < changeStations.size())
				{
					if (index != changeMiddleStationIndex && index != changeEndStationIndex)
					{
						double changeStationDistance = Location.GetDistance(((Station)changeStations.get(index)).GetLocation(), travelPart.GetDestination());
						
						if (changeStationDistance <= currentDistance)
						{
							fAllAreFarer = false;
							
							break;
						}
					}
					
					index += delta;
				}
			}
			
			if (fAllAreFarer) // all the stations have greater distances then the current one, so drop it
				return true;
		}
				
		return false;
	}
	
	/**
	 * @author Piotrek
	 * @see ChangesJudger#SetTravelConstraints(scheduler.TravelConstraints)
	 */
	public void SetTravelConstraints(TravelConstraints constraints)
	{
		if (constraints == null)
			throw new NullPointerException("Constraints must not be null in ChangesJudgerStub.SetTravelConstraints().");
		
		travelConstraints = constraints;
	}
	
	/**
	 * The current travel travelConstraints.
	 * @author Piotrek
	 */
	private TravelConstraints travelConstraints = null;
}
