/*
 * Created on 2005-09-04
 */
package scheduler.modules;

import java.util.ArrayList;
import java.util.Iterator;
import jess.Fact;
import jess.JessException;
import jess.Rete;
import jess.Token;
import jess.ValueVector;
import scheduler.TestOptions;
import scheduler.modulesinterfaces.MeansOfTransportationProvider;
import world.Location;
import world.MeanOfTransportation;

/**
 * A JESS based provider of potentail means of transportation for the first step of planning algorithm.
 * @author Piotrek
 */
public final class MeansOfTransportationJessChooser implements MeansOfTransportationProvider
{
	/**
	 * Constructor.
	 * @author Piotrek
	 */
	public MeansOfTransportationJessChooser()
	{
		try
		{
			// loading rules
			rete.executeCommand("(batch \"./Data/JESS means of transportation rules.txt\")");
			
			// these facts will be always asserted even after resets
			rete.executeCommand("(deffacts distancesLimitations" +
								"(airplaneMinDistance " + MeansOfTransportationProvider.MIN_FLIGHT_DISTANCE + ")" +
								"(trainMinDistance " + MeansOfTransportationProvider.MIN_TRAIN_DISTANCE + ")" +
								"(busMinDistance " + MeansOfTransportationProvider.MIN_BUS_DISTANCE + ")" +
								"(shipMinDistance " + MeansOfTransportationProvider.MIN_CRUISE_DISTANCE + ")" +
								"(walkMaxDistance " + MeansOfTransportationProvider.MAX_WALK_DISTANCE + ")" +
								"(taxiMinDistance " + MeansOfTransportationProvider.MIN_TAXI_DISTANCE + ")" +
								"(taxiMaxDistance " + MeansOfTransportationProvider.MAX_TAXI_DISTANCE + ")" +
					            ")");
			
			// this will define the query for the proper means of transportation
			rete.executeCommand("(defquery searchProperMeans" +
								"\"Finds properMean facts\"" +
								"(properMean ?X))");
		}
		
		catch (JessException xcp)
		{
			// this will cause the algorithm to stop
			throw new RuntimeException("Line: " + xcp.getLineNumber() + "\nContext: " + xcp.getContext() + "\nData: " + xcp.getData() + "\nProgram text: " + xcp.getProgramText() + "\nMessage: " + xcp.getMessage());
		}
	}
	
	/**
	 * The Java Expert Shell System used to support rules genereting the proper means of transportation.
	 * @author Piotrek
	 */
	private Rete rete = new Rete();
		
	/**
	 * @see scheduler.modulesinterfaces.MeansOfTransportationProvider#GetMeansOfTransportation(world.Location, world.Location)
	 * @author Piotrek
	 */
	public ArrayList GetMeansOfTransportation(Location origin, Location destination)
	{
		ArrayList meansOfTransportation = new ArrayList();

		if (TestOptions.ShouldUseRulesChoosingMeansOfTransportation()) // TIP: this 'if' and 'else' might be removed in production version
		{
			try
			{
				rete.reset();
				
				// load data into the rete
	
				// distance
				rete.executeCommand("(assert (distance " + Location.GetDistance(origin, destination) + "))");
				
				// continents
				if (origin.continentName.compareTo("") != 0 && !TestOptions.ShouldIgnoreContinentIslandInformation())
					rete.executeCommand("(assert (continent1 \"" + origin.continentName + "\"))");
				else
					rete.executeCommand("(assert (continent1 NotAvailable))");
				
				if (destination.continentName.compareTo("") != 0 && !TestOptions.ShouldIgnoreContinentIslandInformation())
					rete.executeCommand("(assert (continent2 \"" + destination.continentName + "\"))");
				else
					rete.executeCommand("(assert (continent2 NotAvailable))");
				
				// islands
				if (origin.islandName.compareTo("") != 0 && !TestOptions.ShouldIgnoreContinentIslandInformation())
					rete.executeCommand("(assert (island1 \"" + origin.islandName + "\"))");
				else
					rete.executeCommand("(assert (island1 NotAvailable))");
				
				if (destination.islandName.compareTo("") != 0 && !TestOptions.ShouldIgnoreContinentIslandInformation())
					rete.executeCommand("(assert (island2 \"" + destination.islandName + "\"))");
				else
					rete.executeCommand("(assert (island2 NotAvailable))");
	
				// running the rules
				rete.executeCommand("(run)");
				
				// issue query for each mean of transportation
				
				// Run the query, store the result
			    rete.store("RESULT", rete.runQuery("searchProperMeans",
			                        new ValueVector()));
			    
			    rete.executeCommand("(store RESULT (run-query searchProperMeans))");
	
			    // Fetch the result (an Iterator).
			    Iterator resultsIterator = (Iterator) rete.fetch("RESULT").externalAddressValue(null);
			    
			    ArrayList properMeansOfTransportation = new ArrayList();
	
			    // Pick each element of the Iterator apart and store the
			    // interesting part in the ArrayListproperMeansOfTransportationv.
			    while (resultsIterator.hasNext())
			    {
					Token token = (Token)resultsIterator.next();
	
			    	// We want the second fact in the token - the first is the query trigger
			    	Fact fact = token.fact(1);
	
			    	// The first and only slot of this fact is the __data multislot.
			    	ValueVector multiSlot = fact.get(0).listValue(null);
	
			    	// The second element of this slot is the datum we're interested in.
			    	properMeansOfTransportation.add(multiSlot.get(0).stringValue(null));
			    }
			    
			    for (Iterator answers = properMeansOfTransportation.iterator(); answers.hasNext();)
			    {
			    	String meanOfTransportation = (String)answers.next();
			    	
			    	if (meanOfTransportation.compareTo("Airplane") == 0)
			    		meansOfTransportation.add(new Integer(MeanOfTransportation.AIRPLANE));
			    	else
			    	if (meanOfTransportation.compareTo("Bus") == 0)
				    	meansOfTransportation.add(new Integer(MeanOfTransportation.BUS));
			    	else
			    	if (meanOfTransportation.compareTo("Train") == 0)
			    		meansOfTransportation.add(new Integer(MeanOfTransportation.TRAIN));
			    	else
		    		if (meanOfTransportation.compareTo("Ship") == 0)
			    		meansOfTransportation.add(new Integer(MeanOfTransportation.SHIP));
		    		else
			    	if (meanOfTransportation.compareTo("Walk") == 0)
			    		meansOfTransportation.add(new Integer(MeanOfTransportation.WALK));
			    	else
			    	if (meanOfTransportation.compareTo("Taxi") == 0)
			    		meansOfTransportation.add(new Integer(MeanOfTransportation.TAXI));
			    	else
			    		throw new RuntimeException("Unexpected mean of transportation received from the JESS system.");
			    }
			}
			
			catch (JessException xcp)
			{
				// this will cause the algorithm to stop
				throw new RuntimeException("There was a JESS exception in the JESS based means of transportation provider: " + xcp.getMessage());
			}
		}
		else
		{
			double distance = Location.GetDistance(origin, destination);
			
			// simple walk rule
			if (distance <= MAX_WALK_DISTANCE)
				meansOfTransportation.add(new Integer(MeanOfTransportation.WALK));
				
			// taxi simple rule		
			if (distance >= MIN_TAXI_DISTANCE && distance <= MAX_TAXI_DISTANCE && distance > MAX_WALK_DISTANCE)
				meansOfTransportation.add(new Integer(MeanOfTransportation.TAXI));
			
			meansOfTransportation.add(new Integer(MeanOfTransportation.BUS));
			meansOfTransportation.add(new Integer(MeanOfTransportation.TRAIN));
			meansOfTransportation.add(new Integer(MeanOfTransportation.SHIP));
			meansOfTransportation.add(new Integer(MeanOfTransportation.AIRPLANE));
		}
		
		return meansOfTransportation;
	}
}
