/*
 * Created on 2005-06-16
 */
package scheduler.modules;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.SimpleTimeZone;
import javax.swing.JOptionPane;
import scheduler.ConnectionsFinderTask;
import scheduler.TravelPartConnectionChunk;
import scheduler.modulesinterfaces.TravelSolutionsJudger;
import world.Location;
import world.MeanOfTransportation;
import world.Station;
import crn.CRN;
import crn.CRNFileParser;
import crn.Case;
import crn.IE;
import crn.IEImportance;
import crn.comparators.ChangesCountsComparator;
import crn.comparators.DepartureHoursOfDayComparator;
import crn.comparators.LocationsComparator;
import crn.comparators.RanksComparator;
import crn.comparators.TotalDurationsComparator;

/**
 * A class used to keep eye on user's history of travels (both accepted and rejected cases are stored). 
 * It also performs the finding out of similar cases from the past to the given one.
 * It uses the CRN to achieve that all. 
 * @author Piotrek
 */
public final class SchedulerCRNManager implements TravelSolutionsJudger
{
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param historyFileName A name of the file with the history to be loaded now and saved on exit. 
	 */
	public SchedulerCRNManager(String historyFileName)
	{
		if (historyFileName == null)
			throw new NullPointerException("History file name must not be null in SchedulerCRNManager.SchedulerCRNManager().");
		
		if (historyFileName.length() == 0)
			throw new IllegalArgumentException("History file name must not be empty in SchedulerCRNManager.SchedulerCRNManager().");
		
		try
		{
			CRNFileParser parser = new CRNFileParser();
			
			crn = parser.Parse(historyFileName);
		}
		
		catch (Exception xcp)
		{
			JOptionPane.showMessageDialog(null, "There was an error while loading the history: " + xcp.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
			
			crn = new CRN();
			
			// loading the weights since the parsing failed
			crn.SetIEWeight("UsedMeanOfTransportation", IEImportance.IMPORTANT);
			crn.SetIEWeight("DepartureHourOfDay", IEImportance.UNIMPORTANT);
			crn.SetIEWeight("Origin", IEImportance.NEUTRAL);
			crn.SetIEWeight("Destination", IEImportance.NEUTRAL);
			crn.SetIEWeight("FirstStation", IEImportance.IMPORTANT);
			crn.SetIEWeight("LastStation", IEImportance.IMPORTANT);
			crn.SetIEWeight("ChangesCount", IEImportance.NEUTRAL);
			crn.SetIEWeight("TotalDuration", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.WALK) + "Rank", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.TAXI) + "Rank", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.BUS) + "Rank", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.TRAIN) + "Rank", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.SHIP) + "Rank", IEImportance.NEUTRAL);
			crn.SetIEWeight(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.AIRPLANE) + "Rank", IEImportance.NEUTRAL);
		}
		
		// setting the specific comparators
		crn.AddSpecificComparator("DepartureHourOfDay", new DepartureHoursOfDayComparator());
		crn.AddSpecificComparator("Origin", new LocationsComparator());
		crn.AddSpecificComparator("Destination", new LocationsComparator());
		crn.AddSpecificComparator("ChangesCount", new ChangesCountsComparator());
		crn.AddSpecificComparator("TotalDuration", new TotalDurationsComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.WALK) + "Rank", new RanksComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.TAXI) + "Rank", new RanksComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.SHIP) + "Rank", new RanksComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.AIRPLANE) + "Rank", new RanksComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.BUS) + "Rank", new RanksComparator());
		crn.AddSpecificComparator(MeanOfTransportation.GetMeanOfTransportationString(MeanOfTransportation.TRAIN) + "Rank", new RanksComparator());
		
		this.historyFileName = historyFileName;
	}
	
	/**
	 * Gets the CRN used by this CRN manager.
	 * @author Piotrek
	 * @return The requested CRN.
	 */
	public CRN GetCRN()
	{
		return crn;
	}
	
	/**
	 * A name of the history file.
	 * @author Piotrek
	 */
	private String historyFileName = null;
	
	/**
	 * A CRN used to store the history.
	 * @author Piotrek
	 */
	private CRN crn = null;
	
	/**
	 * Saves the current history state to the file it has been loaded from.
	 * @author Piotrek
	 */
	public void SaveToFile()
	{
		try
		{
			crn.StoreToFile(historyFileName);
		}
		
		catch (IOException xcp)
		{
			JOptionPane.showMessageDialog(null, "There was an error while saving the history: " + xcp.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * @author Piotrek
	 * @see TravelSolutionsJudger#JudgeTravelSolution(ConnectionsFinderTask)
	 */
	public double JudgeTravelSolution(ConnectionsFinderTask travelSolution)
	{
		if (travelSolution == null)
			throw new NullPointerException("Travel solution must not be null in SchedulerCRNManager.JudgeTravelSolution().");
		
		Case travelSolutionCase = GetCase(travelSolution, true);
		
		Case theMostSimilarCase = crn.FindTheMostSimilarCase(travelSolutionCase);
		
		if (theMostSimilarCase == null) // no history data
			return 0.0;
		
		if (theMostSimilarCase.IsAccepted()) // similar to the accepted case
			return theMostSimilarCase.GetSimilarity();
		
		return -1 * theMostSimilarCase.GetSimilarity(); // similar to the rejected case
	}
	
	/**
	 * @author Piotrek
	 * @see TravelSolutionsJudger#AddTravelSolution(ConnectionsFinderTask, boolean)
	 */
	public void AddTravelSolution(ConnectionsFinderTask travelSolution, boolean fAccepted)
	{
		Case crnCase = GetCase(travelSolution, fAccepted);
		
		crn.AddCase(crnCase, IEImportance.NEUTRAL);
	}
	
	/**
	 * Converts the travel solution to the CRN case.
	 * @author Piotrek
	 * @param travelSolution A travel solution to be converted.
	 * @param fAccepted Indicates whether the returned case must be the accepted or rejected.
	 * @return A CRN case being a result of convertion (it is an accpeted case).
	 */
	private Case GetCase(ConnectionsFinderTask travelSolution, boolean fAccepted)
	{
		if (travelSolution == null)
			throw new NullPointerException("Travel solution must not be null in SchedulerCRNManager.GetCase().");
		
		Case crnCase = new Case("CRN Case", fAccepted, null);
		
		// building the case
		
		// used means of transportation
		ArrayList usedMeansOfTransportation = travelSolution.GetUsedMeansOfTransportation();
		 
		Iterator iterator = usedMeansOfTransportation.iterator();
		 
		while (iterator.hasNext())
			crnCase.AddIEReference(new IE("UsedMeanOfTransportation", MeanOfTransportation.GetMeanOfTransportationString(((Integer)iterator.next()).intValue())));
		
		// departure hour of day (without minutes)
		Calendar calendar = Calendar.getInstance(new SimpleTimeZone(0, "GMT"));
		
		calendar.setTime(travelSolution.GetFirstConnectionChunk().GetStartDate());
		
		crnCase.AddIEReference(new IE("DepartureHourOfDay", Integer.toString(calendar.get(Calendar.HOUR_OF_DAY))));
		
		// ranking of means of transportation (by the number of connection chunks)
		HashMap map = new HashMap();
		
		// gathering the stats
		iterator = travelSolution.GetConnectionChunks();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			Integer meanOfTransportation = new Integer(chunk.GetMeanOfTransportation());
			
			if (map.containsKey(meanOfTransportation)) // updating key value
			{
				Integer counter = (Integer)map.get(meanOfTransportation);
				
				counter = new Integer(counter.intValue() + 1);
				
				map.put(meanOfTransportation, counter);
			}
			else // creating a new key
			{
				Integer counter = new Integer(1);
				
				map.put(meanOfTransportation, counter);
			}
		}
		
		// sorting on frequency of use
		ArrayList ranking = new ArrayList(map.size());
				
		iterator = map.values().iterator();
		
		while (iterator.hasNext())
		{
			Integer counter = (Integer)iterator.next();
			
			boolean fAdded = false;
			
			for (int i = 0; i < ranking.size(); ++i)
			{
				Integer rankingCounter = (Integer)ranking.get(i);
				
				if (counter.intValue() > rankingCounter.intValue())
				{
					ranking.add(i, counter);
					
					fAdded = true;
					
					break;
				}
				
				if (counter.intValue() == rankingCounter.intValue())
				{
					// ignoring it (later all means of transportation with this count will be added)
					
					fAdded = true;
					
					break;
				}
			}
			
			if (!fAdded)
				ranking.add(counter);
		}
		
		// adding ranking to the CRN case
		int place = 1;
		
		iterator = ranking.iterator();
		
		while (iterator.hasNext())
		{
			Integer counter = (Integer)iterator.next();
			
			Iterator meansOfTransportation = usedMeansOfTransportation.iterator();
			
			int addedPlaces = 0;
			
			while (meansOfTransportation.hasNext())
			{
				Integer meanOfTransportation = (Integer)meansOfTransportation.next();
				
				Integer mapCounter = (Integer)map.get(meanOfTransportation);
				
				if (counter.intValue() == mapCounter.intValue())
				{
					crnCase.AddIEReference(new IE(MeanOfTransportation.GetMeanOfTransportationString(meanOfTransportation.intValue()) + "Rank", Integer.toString(place)));
					
					++addedPlaces;
				}
			}
			
			place += addedPlaces;
		}
		
		// origin
		Location origin = travelSolution.GetFirstConnectionChunk().GetOrigin();
		
		crnCase.AddIEReference(new IE("Origin", origin.x + ";" + origin.y));
		
		// destiantion
		Location destination = travelSolution.GetLastConnectionChunk().GetDestination();
		
		crnCase.AddIEReference(new IE("Destination", destination.x + ";" + destination.y));
		
		// finding out the first and last stations (non virtual ones)
		Station firstStation = null;
		Station lastStation = null;
		
		iterator = travelSolution.GetConnectionChunks();
		
		while (iterator.hasNext())
		{
			TravelPartConnectionChunk chunk = (TravelPartConnectionChunk)iterator.next();
			
			Station originStation = chunk.GetOriginStation();
			Station destinationStation = chunk.GetDestinationStation();
			
			if (originStation != null && originStation.name.compareTo("Virtual Taxi Station") != 0 && originStation.name.compareTo("Virtual Walk Station") != 0)
			{
				if (firstStation == null)
					firstStation = originStation;
			}
			
			if (destinationStation != null && destinationStation.name.compareTo("Virtual Taxi Station") != 0 && destinationStation.name.compareTo("Virtual Walk Station") != 0)
			{
				if (firstStation == null)
					firstStation = destinationStation;
			}
			
			if (originStation != null && originStation.name.compareTo("Virtual Taxi Station") != 0 && originStation.name.compareTo("Virtual Walk Station") != 0)
				lastStation = originStation;
			
			if (destinationStation != null && destinationStation.name.compareTo("Virtual Taxi Station") != 0 && destinationStation.name.compareTo("Virtual Walk Station") != 0)
				lastStation = destinationStation;
		}
		
		// first station
		if (firstStation != null)
			crnCase.AddIEReference(new IE("FirstStation", firstStation.name));
		
		// last station
		if (lastStation != null)
			crnCase.AddIEReference(new IE("LastStation", lastStation.name));
		
		// total duration
		crnCase.AddIEReference(new IE("TotalDuration", Integer.toString(travelSolution.GetTotalDuration())));
		
		// changes count
		crnCase.AddIEReference(new IE("ChangesCount", Integer.toString(travelSolution.GetChangesCount())));
		
		return crnCase;
	}
}
