/*
 * Created on 2005-03-30
 */
package world;

import gui.controls.WorldView;



/**
 * A location.
 * @author Piotrek
 */
public class Location
{
	/**
	 * The x coordinate of the object.
	 * @author Piotrek
	 */
	public int x;
	
	/**
	 * The y coordinate of the object.
	 * @author Piotrek
	 */
	public int y;
	
	/**
	 * Constructor.
	 * @author Piotrek
	 * @param x The x coordinate of object.
	 * @param y The y coordinate of object.
	 */
	public Location(int x, int y)
	{
		this.x = x;
		this.y = y;
	}
	
	/**
	 * A name of the continentName this location belongs to. An empty string, if no info is available.
	 * @author Piotrek
	 */
	public String continentName = "";
	
	/**
	 * A special value for the property of @see #islandName indicating this location lies on the continental part of the land.
	 * @author Piotrek
	 */
	public static final String CONTINENTAL_PART = "cp";
	
	/**
	 * A name of the island this point lies on or a special value of @see #CONTINENTAL_PART. An empty string, if no info is available.
	 * @author Piotrek 
	 */
	public String islandName = "";
	
	/**
	 * Compares this location to the other one. Two locations are equal if they have the same x and y and all the other values equal if they are set in both locations.
	 * @param other The location this one is to be compared to.
	 * @author Piotrek
	 * @return A value indicating whether both locations are equal.
	 */
	public boolean IsEqual(Location other)
	{
		if (x != other.x || y != other.y)
			return false;
		
		if (continentName.compareTo("") != 0 && other.continentName.compareTo("") != 0)
			if (continentName.compareTo(other.continentName) != 0)
				return false;

		if (islandName.compareTo("") != 0 && other.islandName.compareTo("") != 0)
			if (islandName.compareTo(other.islandName) != 0)
				return false;
		
		return true;
	}
	
	/**
	 * @author Piotrek
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object other)
	{
		if (!(other instanceof Location))
			return false;
		
		return IsEqual((Location)other);
	}
	
	/**
	 * Gets the distance between two locations. The world is a square, but the roundness of the world is simulated (East-West continuity). 
	 * @author Piotrek
	 * @param x One of the points.
	 * @param y The second of the points.
	 * @return The distance between the two given locations.
	 */
	public static double GetDistance(Location x, Location y)
	{
		if (x == null || y == null)
			throw new IllegalArgumentException("Both locations must not be null in Location.GetDistance().");
				
		// direct distance
		double directDistance = Math.sqrt((x.x - y.x) * (x.x - y.x) + (x.y - y.y) * (x.y - y.y));
		
		// East/West common edge distance
		double eastWestDistance = 0;
		
		if (x.x < y.x)
			eastWestDistance = Math.sqrt((x.x + WorldView.width - y.x) * (x.x + WorldView.width - y.x) + (x.y - y.y) * (x.y - y.y));
		else
			eastWestDistance = Math.sqrt((x.x - y.x - WorldView.width) * (x.x - y.x - WorldView.width) + (x.y - y.y) * (x.y - y.y));
		
		return Math.min(directDistance, eastWestDistance);
	}
		
	/**
	 * Gets a boolean value indicating whether this location lies in the given rectangle. The world is a square, but the roundness of the world is simulated (East-West continuity).
	 * The order of the given corners of the rectangle is irrelevant - in both cases they are treated the same; the rectangle with the diagonal (0, 0) - (1, 1) is the same as the one with the diagonal (1, 1) - (0, 0).
	 * @author Piotrek
	 * @param x1 The X coordinate of the first corner of the rectangle.
	 * @param y1 The Y coordinate of the first corner of the rectangle.
	 * @param x2 The X coordinate of the second corner of the rectangle.
	 * @param y2 The Y coordinate of the second corner of the rectangle.
	 * @return A boolean value indicating whether this location lies in the given rectangle.
	 */
	public boolean IsInRange(int x1, int y1, int x2, int y2)
	{	
		if (x1 < 0 && x2 < 0)
			throw new IllegalArgumentException("Both xs must not be negative in Location.IsInRange().");
		
		if (x1 > WorldView.width && x2 > WorldView.width)
			throw new IllegalArgumentException("Both xs must not be greater the the width of the world in Location.IsInRange().");
		
		if (Math.max(x1, x2) - WorldView.width >= WorldView.width)
			throw new IllegalArgumentException("One of the xs is too big in Location.IsInRange().");
		
		if (Math.min(x1, x2) + WorldView.width <= 0)
			throw new IllegalArgumentException("One of the xs is too small in Location.IsInRange().");
		
		// swapping the coordinates if required
		int temp;
		
		if (x1 > x2)
		{
			temp = x1;
			x1 = x2;
			x2 = temp;
		}
		
		if (y1 > y2)
		{
			temp = y1;
			y1 = y2;
			y2 = temp;
		}
		
		if (x1 >= 0 && x2 <= WorldView.width && x >= x1 && x <= x2 && y >= y1 && y <= y2)
			return true;
		
		if (x1 < 0 && x2 <= WorldView.width && !(x > x2 && x < x1 + WorldView.width) && y >= y1 && y <= y2)
			return true;
		
		if (x1 >= 0 && x2 > WorldView.width && !(x > x2 - WorldView.width && x < x1) && y >= y1 && y <= y2)
			return true;
		
		if (x1 < 0 && x2 > WorldView.width && y >= y1 && y <= y2)
			return true;
		
		return false;
	}
}
