/*
 * Created on 2005-04-09
 */
package world;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;

/**
 * A class providing API for the time manipulations and counting a time of journey. 
 * @author Piotrek
 */
public final class TravelTime
{
	/**
	 * Counts a time of travel between two points using the given mean of transportation.
	 * @author Piotrek
	 * @param x The starting point of travel.
	 * @param y The end poinf of th travel.
	 * @param meanOfTransportation A mean of transportation. This is required only to count the default times of travel.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @return A time of tavel in minutes.
	 */
	public static int GetTravelTime(Location x, Location y, int meanOfTransportation)
	{
		if (x == null || y == null)
			throw new NullPointerException("Location must not be null in TravelTime.GetTravelTime().");
		
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("This is not a valid mean of transportation in TravelTime.GetTravelTime().");
		
		double speed = 0;
		
		switch (meanOfTransportation)
		{
			case MeanOfTransportation.AIRPLANE:
				
				speed = AIRPLANE_SPEED;
			
				break;
			
			case MeanOfTransportation.BUS:
				
				speed = BUS_SPEED;
			
				break;
			
			case MeanOfTransportation.TRAIN:
				
				speed = TRAIN_SPEED;
			
				break;
			
			case MeanOfTransportation.SHIP:
				
				speed = SHIP_SPEED;
			
				break;
		}
		
		double distance = Location.GetDistance(x, y);
				
		return (int)(distance / speed);
	}
	
	/**
	 * A speed of an airplane in units per minute.
	 * @author Piotrek
	 */
	private static final double AIRPLANE_SPEED = 1.4;
	
	/**
	 * A speed of a train in units per minute.
	 * @author Piotrek
	 */
	private static final double TRAIN_SPEED = 0.5;
	
	/**
	 * A speed of a bus in units per minute.
	 * @author Piotrek
	 */
	private static final double BUS_SPEED = 0.3;
	
	/**
	 * A speed of a ship in units per minute.
	 * @author Piotrek
	 */
	private static final double SHIP_SPEED = 0.1;
	
	/**
	 * Provides a time of change for the given mean of transportation.
	 * @author Piotrek
	 * @param meanOfTransportation A mean of transportation. This is required only to count the default times of travel.
	 * @see world.MeanOfTransportation#AIRPLANE
	 * @see world.MeanOfTransportation#BUS
	 * @see world.MeanOfTransportation#TRAIN
	 * @see world.MeanOfTransportation#SHIP
	 * @return A time of change in minutes.
	 */
	public static int GetChangeTime(int meanOfTransportation)
	{
		if (!MeanOfTransportation.IsValidMeanOfTransportation(meanOfTransportation))
			throw new IllegalArgumentException("This is not a valid mean of transportation in TravelTime.GetChangeTime().");
		
		switch (meanOfTransportation)
		{
			case MeanOfTransportation.AIRPLANE:
				
				return 30;
			
			case MeanOfTransportation.BUS:
				
				return 5;
			
			case MeanOfTransportation.TRAIN:
				
				return 10;
			
			case MeanOfTransportation.SHIP:
				
				return 60;
		}
		
		return 0;
	}
	
	/**
	 * Converts the given time in minutes to the `+hh:mm' format.
	 * @author Piotrek
	 * @param time Duration in minutes.
	 * @return A formated time string.
	 */
	public static String GetShiftTime(int time)
	{
		if (time < 0)
			throw new IllegalArgumentException("Time must not be negative in TravelTime.GetShiftTime().");
		
		return "+" + GetTime(time);
	}
	
	/**
	 * Converts the given time in minutes since midnight to the `hh:mm' format.
	 * @author Piotrek
	 * @param time Time in minutes.
	 * @return A formated time string.
	 */
	public static String GetTime(int time)
	{
		if (time < 0)
			throw new IllegalArgumentException("The time in TravelTime.GetTime() must not be negative.");
		
		int hours = time / 60;
		int minutes = time % 60;
		
		if (minutes < 10 && hours >= 10)
			return hours + ":0" + minutes;
		
		if (minutes < 10 && hours < 10)
			return "0" + hours + ":0" + minutes;
		
		if (minutes >= 10 && hours < 10)
			return "0" + hours + ":" + minutes;
		
		return hours + ":" + minutes;
	}
	
	/**
	 * Converts the `+hh:mm' or `hh:mm' or `mm' to the duration time in minutes.
	 * In all cases it does not have to be exactly two digits for hours and minutes 
	 * (can be also one or more and none for hours only.) 
	 * @author Piotrek
	 * @param time A string to be converted.
	 * @return A duration in minutes.
	 * @throws ParseException Thrown if the given time string is not well formated.
	 */
	public static int ParseDuration(String time) throws ParseException
	{	
		if (time == null)
			throw new NullPointerException("Time must not be null in TravelTime.ParseDuration().");
		
		if (time.startsWith("+"))
			time = time.substring(1);
		
		return ParseTime(time);
	}
	
	/**
	 * Converts the `hh:mm' or `mm' to the time in minutes.
	 * In all cases it does not have to be exactly two digits for hours and minutes 
	 * (can be also one or more and none for hours only.) 
	 * @author Piotrek
	 * @param time A string to be converted.
	 * @return A time in minutes since midnight.
	 * @throws ParseException Thrown if the given time string is not well formated.
	 */
	public static int ParseTime(String time) throws ParseException
	{
		if (time == null)
			throw new NullPointerException("Time must not be null in TravelTime.ParseTime().");
				
		int duration = 0;
		
		String[] values = time.split(":");
		
		if (values.length > 2)
			throw new ParseException("There are too many `:' signs.", 0);
		
		if (values.length == 1) // just minutes
		{
			try
			{
				duration = Integer.parseInt(values[0]);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new ParseException("The minutes number is not a valid number.", 0);
			}
			
			if (duration < 1)
				throw new ParseException("The number of minutes must not be less then 1", 0);
		}
		else // both hours and minutes
		{
			int minutes = 0;
			
			try
			{
				minutes = Integer.parseInt(values[1]);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new ParseException("The minutes number is not a valid number.", 0);
			}
			
			if (minutes < 0)
				throw new ParseException("The number of minutes must not be negative.", 0);
			
			int hours = 0;
			
			try
			{
				hours = Integer.parseInt(values[0]);
			}
			
			catch (NumberFormatException xcp)
			{
				throw new ParseException("The hours number is not a valid number.", 0);
			}
			
			if (hours < 0)
				throw new ParseException("The number of hours must not be negative.", 0);
			
			duration = hours * 60 + minutes;
		}
		
		return duration;
	}
	
	/**
	 * Gets a list of times of runs for a connection with the given start time of one 
	 * of its runs and repetition time.
	 * @author Piotrek
	 * @param startTime A time one of the runs for the connection starts at (in minutes since midnight).
	 * @param repetitionTime A time between runs for the connection (in minutes).
	 * @param lowerLimit A lower limit of repetitions (in minutes since midnight).
	 * @param upperLimit An upper limit of repetitions (in minutes since midnight).
	 * @return A list of times (Integers, in minutes since midnight).
	 */
	public static ArrayList GetTimesPerDay(int startTime, int repetitionTime, int lowerLimit, int upperLimit)
	{
		if (startTime < 0 || startTime > 1439)
			throw new IllegalArgumentException("The start time in TravelTime.GetTimesPerDay() is invalid.");
		
		if (repetitionTime < 1 || repetitionTime > 1439)
			throw new IllegalArgumentException("The repetition time in TravelTime.GetTimesPerDay() is invalid.");
		
		if (lowerLimit < 0 || lowerLimit > 1439)
			throw new IllegalArgumentException("The lower time limit in TravelTime.GetTimesPerDay() is invalid.");
		
		if (upperLimit < 0 || upperLimit > 1439)
			throw new IllegalArgumentException("The upper time limit in TravelTime.GetTimesPerDay() is invalid.");
		
		if (lowerLimit >= upperLimit)
			throw new IllegalArgumentException("The lower time limit must be less then the upper time limit in TravelTime.GetTimesPerDay() is invalid.");
				
		ArrayList times = new ArrayList();
		
		startTime = startTime % repetitionTime;
		
		while (startTime < 1439)
		{
			if (startTime >= lowerLimit && startTime <= upperLimit)
				times.add(new Integer(startTime));
			
			startTime += repetitionTime;
			
			if (startTime > upperLimit)
				break;
		}
		
		return times;
	}
	
	/**
	 * Gets a time in minutes since Sunday 00:00.
	 * @author Piotrek
	 * @param dayOfWeek A day of week the reference point is on.
	 * @param dayTime A time in minutes since midnight on the given day of week.S
	 * @return A time in minutes since Sunday 00:00.
	 */
	public static int GetTimeSinceSunday(char dayOfWeek, int dayTime)
	{
		if (dayTime < 0 || dayTime >= 24 * 60)
			throw new IllegalArgumentException("The time of day is invalid in the TravelTime.GetTimeSinceSunday().");
		
		int shift = 0;
		
		switch (dayOfWeek)
		{
			case Connection.MONDAY:
				
				shift = 24 * 60;
			
				break;
				
			case Connection.TUESDAY:
				
				shift = 2 * 24 * 60;
			
				break;
				
			case Connection.WEDNESDAY:
				
				shift = 3 * 24 * 60;
			
				break;
				
			case Connection.THURSDAY:
				
				shift = 4 * 24 * 60;
			
				break;
				
			case Connection.FRIDAY:
				
				shift = 5 * 24 * 60;
			
				break;
				
			case Connection.SATURDAY:
				
				shift = 6 * 24 * 60;
			
				break;
				
			case Connection.SUNDAY:
				
				shift = 0;
			
				break;
				
			default:
				
				throw new IllegalArgumentException("Unexpected day of week in the TravelTime.GetTimeSinceSunday().");
		}
		
		return dayTime + shift;
	}
	
	/**
	 * Shifts the given date by the given number of minutes.
	 * @author Piotrek
	 * @param date A date to be shifted.
	 * @param minutes A number of minutes to shift the date by.
	 */
	public static void ShiftDate(Date date, int minutes)
	{
		Calendar calendar = Calendar.getInstance();
		
		calendar.setTime(date);
		
		calendar.add(Calendar.MINUTE, minutes);
		
		date.setTime(calendar.getTime().getTime());
	}
	
	/**
	 * Gets a difference of two dates in minutes.
	 * @author Piotrek
	 * @param dateA First of the dates to be compared.
	 * @param dateB Second of the dates to be compared. Must not be earlier then the date A.
	 * @return A number of minutes being a difference of dateB and dateA.
	 */
	public static int GetDatesDifference(Date dateA, Date dateB)
	{
		if (dateA == null)
			throw new NullPointerException("Date A must not be null in TravelTime.GetDatesDifference().");
		
		if (dateB == null)
			throw new NullPointerException("Date B must not be null in TravelTime.GetDatesDifference().");
		
		if (dateB.before(dateA))
			throw new IllegalArgumentException("Date B must not be earlier then the date A in TravelTime.GetDatesDifference().");
		
		// getting times in miliseconds since epoch
		long timeA = dateA.getTime();
		long timeB = dateB.getTime();
		
		// difference
		long difference = timeB - timeA + 30000; // 30000 = 0.5 sec - added to round the minutes amount
		
		return (int)(difference / 60000);
	}
	
	/**
	 * Converts the given number of minutes into (dd days)(hh:)mm format.
	 * @author Piotrek
	 * @param minutes
	 * @return The formatted string.
	 */
	public static String MinutesToString(int minutes)
	{
		if (minutes < 0)
			throw new IllegalArgumentException("Minutes must not be negative in the TravelTime.MinutesToString().");
		
		int minutesCount = minutes % 60;
		int hoursCount = (minutes / 60) % 24;
		int daysCount = (minutes / 60) / 24;
		
		String minutesString = Integer.toString(minutesCount);
		
		if (minutesString.length() == 1)
			minutesString = "0" + minutesString;
		
		String hoursString = Integer.toString(hoursCount);
		
		if (hoursString.length() == 1)
			hoursString = "0" + hoursString;
				
		return Integer.toString(daysCount) + " day(s) " + hoursString + ":" + minutesString;
	}
}
